﻿namespace NETing.ExperimentsConsole.Experiments.Abstract.Commands
{
    using System;
    using System.Collections.Generic;
    using System.Linq;

    using Index.Controller.Configurations;
    using Index.Controller.Configurations.ConcreteConfigurations;

    internal sealed class Configure : AbstractConsoleCommand
    {
        internal static readonly IPostingFileConfig[] configs = new IPostingFileConfig[]
        {
           new RawPostingFileConfig(),
           new VBytePostingFileConfig(),
           new NVBytePostingFileConfig(),
           new NPforPostingFileConfig(),
           new CNPforPostingFileConfig(),
           new DCNPforPostingFileConfig(),
           new DCNPforPostingFileConfig(64),
           new DCNPforPostingFileConfig(32),
           new DCNPforPostingFileConfig(16),
           new DCNPforPostingFileConfig(8),
           new PfxCNPforPostingFileConfig(),
           new PtsCNPforPostingFileConfig(),
           new PtsNPforPostingFileConfig()
        };

        public override string Name
        {
            get { return "Configure"; }
        }

        public override string ShortDesc
        {
            get { return "Configure the index."; }
        }

        public override void Execute(string arg)
        {
            for (int i = 0; i < configs.Length; i++)
            {
                Console.WriteLine(i + ": " + configs[i].Name);
            }

            Console.Write("Enter configuration number: ");
            IndexConsole.ConfigureIndexSystem(configs[int.Parse(Console.ReadLine() ?? "0")]);
        }

        internal static IPostingFileConfig ConfigByName(string name)
        {
            return configs.FirstOrDefault(postingFileConfig => name == postingFileConfig.Name);
        }

        internal static IEnumerable<IPostingFileConfig> ConfigsByNames(IEnumerable<string> names)
        {
            if (names.Any(name => name.Equals("all", StringComparison.InvariantCultureIgnoreCase)))
            {
                return configs;
            }

            return names.Select(ConfigByName);
        }
    }
}
