﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.IO;
using NETing.Index.Documents.DocumentModel.Scope;
using NETing.Index.Documents.Parser;
using NETing.Index.QueryLanguage;

namespace NETing.ExperimentsConsole.Experiments.Abstract.QueryTrace
{
    internal sealed class QueryTraceReader : IEnumerable<Query>, IDisposable
    {
        private readonly ScopeDirectory scopeDirectory;
        private readonly BinaryReader queryTraceFile;

        private int id = 0;

        public QueryTraceReader(
            ScopeDirectory scopeDirectory,
            BinaryReader queryTraceFile)
        {
            this.scopeDirectory = scopeDirectory;
            this.queryTraceFile = queryTraceFile;
        }

        public IEnumerator<Query> GetEnumerator()
        {
            while (queryTraceFile.ReadBoolean())
            {
                Query q = ReadQuery();
                q.Id = id++;
                yield return q;
            }
        }

        IEnumerator IEnumerable.GetEnumerator()
        {
            return GetEnumerator();
        }

        private Query ReadQuery()
        {
            var type = (QueryType)queryTraceFile.ReadByte();
            switch (type)
            {
                case QueryType.Simple: 
                    return ReadSimple();
                case QueryType.SimplePathType: 
                    return ReadSimplePathType();
                case QueryType.Intersection:
                    return ReadIntersection();
                case QueryType.IntersectionPathType:
                    return ReadIntersectionPathType();
            }

            return null;
        }

        private Query ReadSimple()
        {
            return new Query(new[] { ReadTerm() });
        }

        private Query ReadSimplePathType()
        {
            return new Query(new[] { ReadTerm() }, ReadPathTypes());
        }

        private Query ReadIntersection()
        {
            return new Query(ReadTerms());
        }

        private Query ReadIntersectionPathType()
        {
            return new Query(ReadTerms(), ReadPathTypes());
        }

        private Term[] ReadTerms()
        {
            int numTerms = queryTraceFile.ReadInt32();
            var terms = new Term[numTerms];
            for (int i = 0; i < numTerms; i++)
            {
                terms[i] = ReadTerm();    
            }

            return terms;
        }

        private Term ReadTerm()
        {
            return new Term(queryTraceFile.ReadChars(queryTraceFile.ReadInt32()));
        }

        private IEnumerable<PathType> ReadPathTypes()
        {
            int numPathTypes = queryTraceFile.ReadInt32();
            var pathTypes = new PathType[numPathTypes];

            for (int i = 0; i < numPathTypes; i++)
            {
                pathTypes[i] = scopeDirectory.PathTypes[queryTraceFile.ReadInt32()];
            }

            return pathTypes;
        }

        public void Dispose()
        {
            queryTraceFile.Dispose();
        }
    }
}