﻿namespace NETing.ExperimentsConsole.Experiments.Abstract.QueryTrace
{
    using System;
    using System.Collections.Generic;
    using System.IO;
    using Index.Documents.DocumentModel.Scope;
    using Index.Documents.Parser;
    using Index.QueryLanguage;
    using System.Linq;

    internal sealed class QueryTraceWriter : IDisposable
    {
        private readonly BinaryWriter traceFile;

        public QueryTraceWriter(BinaryWriter binaryWriter)
        {
            traceFile = binaryWriter;
        }

        public void AddQuery(Query query)
        {
            traceFile.Write(true);
            WriteQuery(query);
        }

        private void WriteQuery(Query query)
        {
            QueryType type = GetQueryType(query);
            traceFile.Write((byte)type);

            switch (type)
            {
                case QueryType.Simple:
                    WriteSimple(query);
                    break;
                case QueryType.SimplePathType:
                    WriteSimplePathType(query);
                    break;
                case QueryType.Intersection:
                    WriteIntersection(query);
                    break;
                case QueryType.IntersectionPathType:
                    WriteIntersectionPathType(query);
                    break;
            }
        }

        private void WriteSimple(Query query)
        {
            WriteTerm(query.Terms.Single());
        }

        private void WriteSimplePathType(Query query)
        {
            WriteTerm(query.Terms.Single());
            WritePathTypes(query.PathTypes);
        }

        private void WriteIntersection(Query query)
        {
            traceFile.Write(query.Terms.Length);
            foreach (var term in query.Terms)
            {
                WriteTerm(term);
            }
        }

        private void WriteIntersectionPathType(Query query)
        {
            traceFile.Write(query.Terms.Length);
            foreach (var term in query.Terms)
            {
                WriteTerm(term);
            }

            WritePathTypes(query.PathTypes);
        }

        private void WriteTerm(Term term)
        {
            traceFile.Write(term.Length);
            traceFile.Write(term.ArrayHolder.Array, term.Start, term.Length);
        }

        private void WritePathTypes(IEnumerable<PathType> pathTypes)
        {
            traceFile.Write(pathTypes.Count());
            foreach (PathType pathType in pathTypes)
            {
                traceFile.Write(pathType.PathId);
            }
        }

        private static QueryType GetQueryType(Query query)
        {
            if (query.Terms.Length == 1)
            {
                if ( query.PathTypes.Count() > 0)
                {
                    return QueryType.SimplePathType;
                }

                return QueryType.Simple;
            }

            if (query.PathTypes != null && query.PathTypes.Count() > 0)
            {
                return QueryType.IntersectionPathType;
            }

            return QueryType.Intersection;
        }

        public void Dispose()
        {
            traceFile.Write(false);
            traceFile.Dispose();
        }
    }
}
