﻿namespace NETing.ExperimentsConsole.Experiments
{
    using System;
    using System.Collections.Generic;
    using System.Linq;
    using System.Reflection;

    internal class ExperimentAttribute : Attribute
    {
        public string Name { get; set; }

        public IExperiment Instance { get; private set; }

        public static IEnumerable<ExperimentAttribute> GetInstalledExperiments()
        {
            Type[] types;
            try
            {
                types = typeof (ExperimentAttribute).Assembly.GetTypes();
            }
            catch (ReflectionTypeLoadException e)
            {
                types = e.Types.Where(t => t != null).ToArray();
            }

            foreach (Type experiment in 
                types.Where(
                       type => type.IsClass && !type.IsAbstract
                       && typeof(IExperiment).IsAssignableFrom(type)))
            {
                var attrib = (ExperimentAttribute) GetCustomAttribute(
                    experiment, typeof (ExperimentAttribute));
                try
                {
                    attrib.Instance = (IExperiment)experiment.GetConstructor(
                    new Type[0]).Invoke(new object[0]);
                }
                catch (Exception e)
                {
                    Console.WriteLine(e);
                }
                
                yield return attrib;
            }
        }
    }
}


