﻿namespace NETing.ExperimentsConsole.Experiments.IndexConsole.Commands.Build
{
    using System;
    using System.IO;
    
    using Abstract;
    using Index.Controller.Configurations;
    using Index.Documents.FileSource;
    using Index.FileProcessor;
    using Index.Controller.Services;
    using Index.Utils.Benchmark;

    internal class BuildIndex : AbstractConsoleCommand
    {
        public override string Name
        {
            get { return "BuildIndex"; }
        }

        public override string ShortDesc
        {
            get { return "Will build the index."; }
        }

        public override void Execute(string arg)
        {
            Console.WriteLine("------------------------------------------------------------------");
            Console.WriteLine("Will feed all XML files in one directory, please the path to the");
            Console.WriteLine("directory which contains the document collection.");
            Console.Write(">>> ");
            var sourceDir = new DirectoryInfo(Console.ReadLine() ?? "");
            if (!sourceDir.Exists)
            {
                Console.WriteLine("ERROR: Directory must exist, pleas try again.");
                return;
            }

            Console.WriteLine("Will per default feed all documents, enter artificial limit if any.");
            Console.Write(">>> ");
            var input = Console.ReadLine();
            var limit = string.IsNullOrEmpty(input) ? int.MaxValue : int.Parse(input);

            BuilderService builder = IndexConsole.Controller.PrepareBuild(IndexConsole.BuilderSreviceConfig);
            Console.WriteLine("Will feed no more than: " + limit + " documents");
            Console.WriteLine("From: " + sourceDir);
            Console.WriteLine("------------------------------------------------------------------");
            var configs = GetPostingFileConfigs();
            Console.WriteLine("------------------------------------------------------------------");

            builder.Start();
            long startTime = Timing.PerformanceCounter;
            var fileProcessor = new FileProcessor();
            int documents = 0;
            const int printPeriod = 10000;
            int period = 0;
            foreach (var file in new DirectoryCrawler(sourceDir, true))
            {
                builder.Add(fileProcessor.PrepareFile(file));
                if(++documents >= limit)
                {
                    break;
                }

                if (period++ >= printPeriod)
                {
                    period = 0;
                    long time = Timing.PerformanceCounter - startTime;
                    int dps = (int)(documents / (Timing.PerformanceCountToNanoSeconds(time) / 1e9));
                    Console.WriteLine(dps + " documents per second.");
                }
            }

            long feedTime = Timing.PerformanceCounter - startTime;
            Console.WriteLine(documents + " documents feeded in "
                + Timing.PerformanceCountToNanoSeconds(feedTime) / 1e9
                + " seconds.");
            
            builder.PrepareFinish();

            int docPerSec = (int)(documents / (Timing.PerformanceCountToNanoSeconds(feedTime) / 1e9));
            Console.WriteLine(docPerSec + " documents per second.");
            Console.WriteLine("------------------------------------------------------------------");
            Console.WriteLine("Will start writing documents.");
            var inverter = builder.Inverter;
            foreach (IPostingFileConfig config in configs)
            {
                Console.WriteLine("Writing index for system: " + config.Name);
                IndexConsole.ConfigureIndexSystem(
                    IndexConsole.Controller.ScopeDirectoryService.ScopeDirectory, config);
                builder = IndexConsole.Controller.PrepareBuild(IndexConsole.BuilderSreviceConfig);
                builder.Inverter = inverter;
                startTime = Timing.PerformanceCounter;
                builder.Finish();
                long finishTime = Timing.PerformanceCounter - startTime;
                Console.WriteLine("Index written to disk in: "
                    + Timing.PerformanceCountToNanoSeconds(finishTime) / 1e9
                    + " seconds.");
                docPerSec = (int)(documents / (Timing.PerformanceCountToNanoSeconds(finishTime + feedTime) / 1e9));
                Console.WriteLine(docPerSec + " documents per second. (total build time)");
                Console.WriteLine("------------------------------------------------------------------");
                Console.WriteLine("Saving datastructures....");
                builder.Write();
                Console.WriteLine("Finished!");
            }
        }
    }
}
