﻿namespace NETing.ExperimentsConsole.Experiments.IndexConsole.Commands.Query
{
    using System;
    using System.Collections;
    using System.Collections.Generic;
    using Abstract;
    using Index.Controller.Services;
    using Index.QueryLanguage;

    internal sealed class ExecuteQueries : AbstractConsoleCommand
    {
        private const int PageSize = 20;

        private bool running = true;

        public override string Name
        {
            get { return "Query";  }
        }

        public override string ShortDesc
        {
            get { return "Query execution interface"; }
        }

        private static void QueryLanguageSummary()
        {
            QueryParser.PrintSyntax(Console.Out);
        }

        public override void Execute(string arg)
        {
            PrintUsage();
            Console.WriteLine("Lading datastructures, this may take a while.");
            QueryService service = IndexConsole.Controller.PrepareQuery();
            Console.WriteLine("Ready!");
            while (running)
            {
                Step(service);
            }
        }

        private static void PrintUsage()
        {
            Console.WriteLine("Please feed queries. Queries are of the form: ");
            QueryLanguageSummary();
            Console.WriteLine("*** for shutdown");
        }

        private void Step(QueryService service)
        {
            Console.Write(">>> ");
            string query = Console.ReadLine() ?? "";
            if (string.IsNullOrEmpty(query))
            {
                Console.WriteLine("Invalid query: \"" + query + "\"");
                PrintUsage();
            }

            if (query.StartsWith("***"))
            {
                running = false;
            }
            else
            {
                Console.WriteLine("Results for query: " + query);
                ProcessResults(service.Query(query));
                Console.WriteLine("---------------------------------------");
            }
        }

        private void ProcessResults(IEnumerator<object> posting)
        {
            using (posting)
            {
                int page = 0;
                bool ignorePage = false;
                while (posting.MoveNext())
                {
                    if (!ignorePage && page++ > PageSize)
                    {
                        Console.WriteLine("Press enter for more results. Type all to disable paging.");
                        Console.Write(">>>>> ");
                        if ((Console.ReadLine() ?? "").Trim().StartsWith("all"))
                        {
                            ignorePage = true;
                        }

                        page = 0;
                    }

                    var enumerable = posting.Current as IEnumerable;
                    if (enumerable != null)
                    {
                        Console.WriteLine("------------------------------");
                        foreach (var item in enumerable)
                        {
                            Console.WriteLine(item);
                        }

                        Console.WriteLine("------------------------------");
                    }
                    else
                    {
                        Console.WriteLine(posting.Current);
                    }
                }
            }
        }
    }
}
