﻿namespace NETing.ExperimentsConsole.Experiments.IndexConsole.Commands.Query
{
    using System;
    using System.Collections.Generic;
    using Abstract;
    using Index.Controller.Services;
    using Index.Utils.Benchmark;

    internal sealed class QueryBenchMark : AbstractConsoleCommand
    {
        private static string[] instersectionQueries = new[]
           {
               "[the 911] # sentence",
               "(foreign debt)",
               "(yellow pages)",
               "(xml query processing)",
               "(german cake)",
               "(operating system free)",
               "(face book)",
               "(miami white pages)",
               "(kingston jamaica)",
               "(trondheim search engine)",
               "(encyclopedia free)",
               "(encyclopedia free web)",
               "(comic author english)",
               "(freedom preservation army)",
               "(dark side of the moon pink floyd)"
           };

        #region Overrides of AbstractConsoleCommand

        public override string Name
        {
            get { return "QueryBench"; }
        }

        public override string ShortDesc
        {
            get { return "Execute selceted query benchmarks."; }
        }

        public override void Execute(string arg)
        {
            Console.WriteLine("Will execute intersection queries.");
            QueryService service = IndexConsole.Controller.PrepareQuery();
            Console.WriteLine("Query service ready, press enter to continue.");
            Console.ReadLine();
            
            double totalBenchmarkTime = 0;
            for (int i = 0; i < 1; i++)
            {
                foreach (string query in instersectionQueries)
                {
                    long startTime = Timing.PerformanceCounter;
                    using (IEnumerator<object> enumer = service.Query(query))
                    {
                        int cnt = 0;
                        while (enumer.MoveNext())
                        {
                            cnt++;
                        }

                        long total = Timing.PerformanceCounter - startTime;
                        double seconds = Timing.PerformanceCountToNanoSeconds(total) / 1e9;
                        totalBenchmarkTime += seconds;
                        Console.WriteLine(query + ": " + cnt + " in " + seconds + " s");
                    }
                }
            }
            

            Console.WriteLine("Total benchmark time: " + totalBenchmarkTime);
        }

        #endregion
    }
}
