﻿namespace NETing.ExperimentsConsole.Experiments.QueryExecutionExperiment.Commands
{
    using System;

    using System.IO;
    using Index.QueryLanguage;

    internal sealed class ExecuteQueryTrace : AbstractQueryTraceCommand
    {
        public override string Name
        {
            get { return "ExecuteTrace"; }
        }

        public override string ShortDesc
        {
            get { return "Will execute a preprocessed query trace. Statistics will be collected and written to files.";  }
        }

        public override void Execute(string arg)
        {
            FileInfo queryTrace = GetFileInput("Enter full path to query trace:");
            int warmUp = GetIntegerInput("Enter number of query to use for warm up: ", 1);
            int experiments = GetIntegerInput("Enter number of experiments to run: ", 1);
            bool errorHandeling = GetBooleanInput("Catch errors? ", true);
            bool dense = GetBooleanInput("Dense query processing ", false);
            bool parallel = GetBooleanInput("Execute queries in parallel?", true);

            Query[] queries = LoadQueries(queryTrace);

            if (dense)
            {
                foreach (var query in queries)
                {
                    query.Dense = true;
                }
            }

            int concurrency = 0;
            if (parallel)
            {
                concurrency = GetIntegerInput("Enter concurrency level: ", 4);
            }

            for (int i = 0; i < warmUp; i++)
            {
                DoExecuteQueryTrace(queries, parallel, concurrency, errorHandeling);
            }

            Console.WriteLine("Warm up finished!");

            for (int i = 0; i < experiments; i++)
            {
                var results = DoExecuteQueryTrace(queries, parallel, concurrency, errorHandeling);
                Console.WriteLine("Num Queries: " + results.ExecutedQueries);
                Console.WriteLine("Total  time: " + results.TotalMillisecond / 1000.0 + " s");
                Console.WriteLine("all threads: " + results.TotalLatency / 1000.0 + " s");
                Console.WriteLine("Avg latency: " + results.AvgLatency + " ms");
                Console.WriteLine("Throughput : " + results.Throughput + " qps");
                Console.WriteLine("--------------------------------------------------");
            }
        }
    }
}
