﻿namespace NETing.ExperimentsConsole.Experiments.QueryExecutionExperiment.Commands
{
    using System;
    using System.IO;
    using System.Text;
    using Index.Controller.Configurations;
    using Index.QueryLanguage;
    using Utils;

    internal sealed class QueryTraceExperiment : AbstractQueryTraceCommand
    {
        public override string Name
        {
            get { return "QueryTraceExperiment"; }
        }

        public override string ShortDesc
        {
            get { return "Run a comparative experiment comparing several methods configurations."; }
        }

        public override void Execute(string arg)
        {
            string experimentName = GetStringInput("Enter experiment name: ", "unnamed");
            FileInfo queryTrace = GetFileInput("Enter full path to query trace: ");
            bool dense = GetBooleanInput("Dense processing? ", false);

            int warmUp = GetIntegerInput("Enter number of query to use for warm up runs: ", 1);
            int repetitions = GetIntegerInput("Enter number of repetions for each config: ", 1);

            bool parallel = GetBooleanInput("Execute queries in parallel? ", true);
            int concurrencyLevel = 0;
            if (parallel)
            {
                concurrencyLevel = GetIntegerInput("Enter concurrency level: ", 4);
            }

            var configs = GetPostingFileConfigs("Select configurations to run experiments on. ");
            
            DirectoryInfo outputDirectory = GetDirectoryInput("Enter directory to store results in: ");
            if (!outputDirectory.Exists)
            {
                outputDirectory.Create();    
            }

            bool pipeToConsole = GetBooleanInput("Duplicate results to console?: ", true);

            Query[] queries = LoadQueries(queryTrace);

            if (dense)
            {
                foreach (var query in queries)
                {
                    query.Dense = true;
                }
            }

            foreach (IPostingFileConfig config in configs)
            {
                IndexConsole.ConfigureIndexSystem(config);

                Console.WriteLine("Executing experiment for config: " + config.Name);
                var ouputFile = new FileInfo(outputDirectory.FullName + "/" + config.Name + "." + experimentName + ".log");
                TextWriter writer = new StreamWriter(ouputFile.OpenWrite());
                if (pipeToConsole)
                {
                    writer = new Tee(writer, Console.Out);
                }

                for (int i = 0; i < warmUp; i++)
                {
                    DoExecuteQueryTrace(queries, parallel, concurrencyLevel);
                }

                var logLineBuilder = new StringBuilder(1024);
                for (int i = 0; i < repetitions; i++)
                {
                    logLineBuilder.Clear();
                    var statistics = DoExecuteQueryTrace(queries, parallel, concurrencyLevel);
                    logLineBuilder
                        .Append(statistics.ExecutedQueries).Append('\t')
                        .Append(statistics.TotalMillisecond).Append('\t')
                        .Append(statistics.TotalLatency).Append('\t')
                        .Append(statistics.Throughput);

                    writer.WriteLine(logLineBuilder.ToString());
                }

                writer.Dispose();
            }
        }
    }
}
