﻿namespace NETing.ExperimentsConsole.Experiments.UtilityConsole.Commands
{
    using System;
    using System.Collections.Generic;
    using System.IO;
    using Abstract;
    using Index.Controller.Services;

    internal sealed class ClassifyQueryLog : AbstractConsoleCommand
    {
        public override string Name
        {
            get { return "ClassifyQueryLog"; }
        }

        public override string ShortDesc
        {
            get { return "Return a TREC formatted query trace in AND mode, write a file with result counts."; }
        }

        public override void Execute(string arg)
        {
            Console.WriteLine("Enter full path to query trace: ");
            string tracePath = Console.ReadLine() ?? "";
            Console.WriteLine("Enter name of output file: ");
            string output = Console.ReadLine();

            TextWriter outputWriter;
            if (string.IsNullOrEmpty(output))
            {
                Console.WriteLine("Will use console for output!");
                outputWriter = Console.Out;
            }
            else
            {
                outputWriter = new StreamWriter(output);
            }

            Console.WriteLine("Enter path expression for queries, if empty simple ");
            Console.WriteLine(" intersections will be calculated. ");
            Func<string[], string> queryfactory = GetQueryFactory(Console.ReadLine());

            using (outputWriter)
            {
                using (var traceReader = new StreamReader(tracePath))
                {
                    var queryService = IndexConsole.Controller.PrepareQuery();
                    while (!traceReader.EndOfStream)
                    {
                        string query = traceReader.ReadLine();
                        string[] queryParts = query.Split(':');
                        int results = CountResults(queryService, queryfactory(queryParts));
                        outputWriter.WriteLine(queryParts[0] + ": " + results);
                    }
                }
            }
        }

        private static Func<string[], string> GetQueryFactory(string pathExpression)
        {
            if (string.IsNullOrEmpty(pathExpression))
            {
                return TrecToIntersectionQuery;
            }

            return query => TrecToPathIntersectionQuery(query, pathExpression);
        }

        private static string TrecToIntersectionQuery(string[] trecQuery)
        {
            if (trecQuery[1].Contains(" "))
            {
                return "(" + trecQuery[1] + ")";
            }

            return trecQuery[1];
        }

        private static string TrecToPathIntersectionQuery(string[] trecQuery, string path)
        {
            if (trecQuery[1].Contains(" "))
            {
                return "[" + trecQuery[1] + "] # " + path;
            }

            return trecQuery[1];
        }

        private static int CountResults(QueryService service, string query)
        {
            int cnt = 0;
            using (IEnumerator<object> enumerator = service.Query(query))
            {
                while (enumerator.MoveNext())
                {
                    cnt++;
                }
            }

            return cnt;
        }
    }
}
