﻿namespace NETing.ExperimentsConsole.Experiments.UtilityConsole.Commands
{
    using System;
    using System.IO;

    internal sealed class CompileQueryTrace : AbstractQueryTraceCommand
    {
        public override string Name
        {
            get { return "CompileQueryTrace"; }
        }

        public override string ShortDesc
        {
            get { return "Compiles a query trace from a TREC style query log."; }
        }

        public override void Execute(string arg)
        {
            Console.WriteLine("Enter full path to trec query log: ");
            string trecPath = Console.ReadLine() ?? "";

            if (!File.Exists(trecPath))
            {
                Console.WriteLine("Cannot locate file: " + trecPath);
                Console.WriteLine("Please try again.");
                return;
            }

            Console.WriteLine("Enter name of output file: ");
            string output = Console.ReadLine() ?? "compileQueryTrace.out";

            Console.WriteLine("Enter path expression, leave out for simple intersection:");
            string pathExpression = Console.ReadLine() ?? "";

            if (string.IsNullOrWhiteSpace(pathExpression))
            {
                pathExpression = null;
            }

            int minTerms = GetIntegerInput("Enter minimum number of terms in query", 2);
            int maxTerms = GetIntegerInput("Enter maximum number of terms in query", 100);
            int minResults = GetIntegerInput("Enter minimum number of results", 1);
            int buckets = GetIntegerInput("Enter histogram buckets", 0);
            int bucketSize = GetIntegerInput("Enter histogram bucketSize", 0);

            DoCompileQueryTrace(
                new FileInfo(trecPath), new FileInfo(output), pathExpression, minTerms, maxTerms, minResults, buckets, bucketSize);
        }
    }
}
