﻿namespace NETing.ExperimentsConsole.Experiments.UtilityConsole.Commands
{
    using System;
    using System.Collections.Generic;
    using System.IO;
    using Utils;

    internal sealed class GenerateQueryTraces : AbstractQueryTraceCommand
    {
        public override string Name
        {
            get { return "GenerateQueryTraces"; }
        }

        public override string ShortDesc
        {
            get
            {
                return "Generates several query traces, the command creates one trace for each input file" +
                    " and one trace for each path expression if any.";
            }
        }

        public override void Execute(string arg)
        {
            IEnumerable<FileInfo> trecFiles = GetInputFiles("Enter path to trec query logs: ");
            DirectoryInfo outDirectory = GetDirectoryInput("Enter path to output directory: ");

            string paths = GetStringInput("Enter path expressions (separate by space), leave out for simple intersection:", null);
            bool pathQuery = paths != null;
            IEnumerable<string> pathExpressions = null;
            if (pathQuery)
            {
                pathExpressions = paths.Split(' ');
            }

            int minTerms = GetIntegerInput("Enter minimum number of terms in query", 2);
            int maxTerms = GetIntegerInput("Enter maximum number of terms in query", 100);
            int minResults = GetIntegerInput("Enter minimum number of results", 1);
            int buckets = GetIntegerInput("Enter histogram buckets", 0);
            int bucketSize = GetIntegerInput("Enter histogram bucket size", 0);

            bool doTee = GetBooleanInput("Pipe results to console", true);

            TextWriter logOutput = new StreamWriter(File.OpenWrite(outDirectory.FullName + "\\tracelog.log"));
            if (doTee)
            {
                logOutput = new Tee(logOutput, Console.Out);
            }

            foreach (FileInfo trecFile in trecFiles)
            {
                Console.WriteLine("----------------------------------------------------");
                Console.Write("Generating for: ");
                Console.WriteLine(trecFile);
                Console.WriteLine("intersection");
                var output = new FileInfo(outDirectory.FullName + "\\" + trecFile.Name + ".intersection.min" + minResults + "res");
                DoCompileQueryTrace(trecFile, output, null, minTerms, maxTerms, minResults, buckets, bucketSize, logOutput);
            
                if (pathQuery)
                {
                    foreach (string pathExpression in pathExpressions)
                    {
                        Console.WriteLine("Path: " + pathExpression);
                        Console.WriteLine("----------------------------------------------------");

                        output = new FileInfo(outDirectory.FullName + "//" + trecFile.Name + ".path." + pathExpression + ".min" + minResults + "res");
                        DoCompileQueryTrace(trecFile, output, pathExpression, minTerms, maxTerms, minResults, buckets, bucketSize, logOutput);
                    }
                }
            }

            logOutput.Dispose();
        }
    }
}
