﻿namespace NETing.ExperimentsConsole.Experiments.UtilityConsole.Commands
{
    using System;
    using System.Collections.Generic;
    using System.Threading.Tasks;
    
    using Abstract;
    
    using Index.Controller.Services;
    using Index.Documents.Parser;
    using Index.Utils.Benchmark;

    internal sealed class ReadIndex : AbstractConsoleCommand
    {
        private sealed class ReadExecutor
        {
            private readonly QueryService service;
            
            public ReadExecutor(QueryService service)
            {
                this.service = service;
            }

            public void ReadPostings(Term term)
            {
                using (IEnumerator<object> result = service.Query(term))
                {
                    while (result.MoveNext())
                    {
                    }
                }
            }
        }

        public override string Name
        {
            get { return "ReadIndex";  }
        }

        public override string ShortDesc
        {
            get { return "Read the entire index.";  }
        }

        public override void Execute(string arg)
        {
            Console.WriteLine("Preparing query.");
            QueryService service = IndexConsole.Controller.PrepareQuery();
            Console.WriteLine("Data structures loaded.");

            Console.Write("Read lists concurrently? (Y/N) (Y): ");
            string concurrentInput = (Console.ReadLine() ?? "").ToLowerInvariant();
            bool concurrent = false;
            if (string.IsNullOrEmpty(concurrentInput) || concurrentInput == "y")
            {
                concurrent = true;
                Console.WriteLine("Will read in parallel.");
            }
            else
            {
                Console.WriteLine("Will read in one thread.");
            }

            ReadExecutor executor = new ReadExecutor(service);
            long startTime = Timing.PerformanceCounter;
            if (concurrent)
            {
                Parallel.ForEach(IndexConsole.Controller.DictionaryService.GetAllTerms(),
                                 executor.ReadPostings);
            }
            else
            {
                foreach (Term term in IndexConsole.Controller.DictionaryService.GetAllTerms())
                {
                    executor.ReadPostings(term);
                }    
            }
            
            double totalSeconds = Timing.PerformanceCountToNanoSeconds(
                Timing.PerformanceCounter - startTime) / 1e9;
            Console.WriteLine("Read entire index in: " + totalSeconds + " seconds");
        }
    }
}
