﻿namespace NETing.Index.Documents.DocumentModel.Scope
{
    using NUnit.Framework;

    [TestFixture]
    public class DeweyBuilderTest
    {
        [Test]
        public void TestDeweyBuilder()
        {
            var builder = new DeweyBuilder();

            /* Tries the following document
             * 1: <a>                   0
             * 2:  <b>                  0.0
             * 3:  </b>                 0
             * 4:  <b>                  0.1
             * 5:   <c>                 0.1.0
             * 6:   </c>                0.1
             * 7:   <c>                 0.1.1
             * 8:   </c>                0.1
             * 9:   <c>                 0.1.2
             * 10:  </c>                0.1
             * 11:  <c>                 0.1.3
             * 12:   <d>                0.1.3.0
             * 13:    <e>               0.1.3.0.0
             * 14:    </e>              0.1.3.0
             * 15:    <e>               0.1.3.0.1
             * */

            Dewey dewey = new Dewey(new ushort[5]);
            builder.Enter(); // 1: <a>
            builder.Build(dewey);
            Assert.AreEqual(new Dewey(new ushort[] { 0 }), dewey);
            builder.Enter(); // 2:  <b>
            builder.Build(dewey);
            Assert.AreEqual(new Dewey(new ushort[] { 0, 0 }), dewey);
            builder.Exit(); // 3: </b>
            builder.Build(dewey);
            Assert.AreEqual(new Dewey(new ushort[] { 0 }), dewey);
            builder.Enter(); // 4: <b>
            builder.Build(dewey);
            Assert.AreEqual(new Dewey(new ushort[] { 0, 1 }), dewey);
            builder.Enter(); // 5: <c>
            builder.Build(dewey);
            Assert.AreEqual(new Dewey(new ushort[] { 0, 1, 0 }), dewey);
            builder.Exit();  // 6: </c>
            builder.Build(dewey);
            Assert.AreEqual(new Dewey(new ushort[] { 0, 1 }), dewey);
            builder.Enter(); // 7: <c>
            builder.Build(dewey);
            Assert.AreEqual(new Dewey(new ushort[] { 0, 1, 1 }), dewey);
            builder.Exit(); // 8: </c>
            builder.Build(dewey);
            Assert.AreEqual(new Dewey(new ushort[] { 0, 1 }), dewey);
            builder.Enter(); // 9: <c>
            builder.Build(dewey);
            Assert.AreEqual(new Dewey(new ushort[] { 0, 1, 2 }), dewey);
            builder.Exit(); // 10: </c>
            builder.Build(dewey);
            Assert.AreEqual(new Dewey(new ushort[] { 0, 1 }), dewey);
            builder.Enter(); // 11: <c>
            builder.Build(dewey);
            Assert.AreEqual(new Dewey(new ushort[] { 0, 1, 3 }), dewey);
            builder.Enter(); // 12: <d>
            builder.Build(dewey);
            Assert.AreEqual(new Dewey(new ushort[] { 0, 1, 3, 0 }), dewey);
            builder.Enter(); // 13: <e>
            builder.Build(dewey);
            Assert.AreEqual(new Dewey(new ushort[] { 0, 1, 3, 0, 0 }), dewey);
            builder.Exit(); // 14: </e>
            builder.Build(dewey);
            Assert.AreEqual(new Dewey(new ushort[] { 0, 1, 3, 0 }), dewey);
            builder.Enter(); // 15: <e>
            builder.Build(dewey);
            Assert.AreEqual(new Dewey(new ushort[] { 0, 1, 3, 0, 1 }), dewey);
        }
    }
}


