﻿namespace NETing.Index.Documents.DocumentModel.Scope
{
    using System.IO;
    using System.Linq;
    using System.Runtime.Serialization.Formatters.Binary;
    using NUnit.Framework;

    [TestFixture]
    public class ScopeDirectoryTest
    {
        [Test]
        public void TestScopeDirectoryCursor()
        {
            var directory = new ScopeDirectory();
            ScopeDirectory.Cursor cursor = directory.GetCursor();

            cursor.Enter("1");
            cursor.UpdateCurrentScope();
            Assert.AreEqual(cursor.CurrentScope.Dewey, new Dewey(new ushort[] { 0 }));
            Assert.AreEqual(cursor.CurrentScope.PathType.Depth, 1);
            Assert.AreEqual(cursor.CurrentScope.PathType.LastTag.Name, "1");
            Assert.AreEqual(cursor.CurrentScope.PathType.LastTag.TagId, 0);

            cursor.Enter("2");
            cursor.UpdateCurrentScope();
            Assert.AreEqual(cursor.CurrentScope.Dewey, new Dewey(new ushort[] { 0, 0 }));
            Assert.AreEqual(cursor.CurrentScope.PathType.Depth, 2);
            Assert.AreEqual(cursor.CurrentScope.PathType.LastTag.Name, "2");
            Assert.AreEqual(cursor.CurrentScope.PathType.LastTag.TagId, 1);
            Assert.That(
                cursor.CurrentScope.PathType.SequenceEqual(
                    new[]
                        {
                            directory.TagArray[0], directory.TagArray[1]
                        }));

            cursor.Enter("3");
            cursor.UpdateCurrentScope();
            Assert.AreEqual(cursor.CurrentScope.Dewey, new Dewey(new ushort[] { 0, 0, 0 }));
            Assert.AreEqual(cursor.CurrentScope.PathType.Depth, 3);
            Assert.AreEqual(cursor.CurrentScope.PathType.LastTag.Name, "3");
            Assert.AreEqual(cursor.CurrentScope.PathType.LastTag.TagId, 2);
            Assert.That(
                cursor.CurrentScope.PathType.SequenceEqual(
                    new[]
                        {
                            directory.TagArray[0], directory.TagArray[1], directory.TagArray[2]
                        }));

            cursor.Exit();
            cursor.UpdateCurrentScope();
            Assert.AreEqual(cursor.CurrentScope.Dewey, new Dewey(new ushort[] { 0, 0 }));
            Assert.AreEqual(cursor.CurrentScope.PathType.Depth, 2);
            Assert.AreEqual(cursor.CurrentScope.PathType.LastTag.Name, "2");
            Assert.AreEqual(cursor.CurrentScope.PathType.LastTag.TagId, 1);
            Assert.That(
                cursor.CurrentScope.PathType.SequenceEqual(
                    new[]
                        {
                            directory.TagArray[0], directory.TagArray[1]
                        }));

            cursor.Enter("4");
            cursor.UpdateCurrentScope();
            Assert.AreEqual(cursor.CurrentScope.Dewey, new Dewey(new ushort[] { 0, 0, 1 }));
            Assert.AreEqual(cursor.CurrentScope.PathType.Depth, 3);
            Assert.AreEqual(cursor.CurrentScope.PathType.LastTag.Name, "4");
            Assert.AreEqual(cursor.CurrentScope.PathType.LastTag.TagId, 3);
            Assert.That(
                cursor.CurrentScope.PathType.SequenceEqual(
                    new[]
                        {
                            directory.TagArray[0], directory.TagArray[1], directory.TagArray[3]
                        }));

            cursor.Exit();
            cursor.UpdateCurrentScope();
            Assert.AreEqual(cursor.CurrentScope.Dewey, new Dewey(new ushort[] { 0, 0 }));
            Assert.AreEqual(cursor.CurrentScope.PathType.Depth, 2);
            Assert.AreEqual(cursor.CurrentScope.PathType.LastTag.Name, "2");
            Assert.AreEqual(cursor.CurrentScope.PathType.LastTag.TagId, 1);
            Assert.That(
                cursor.CurrentScope.PathType.SequenceEqual(
                    new[]
                        {
                            directory.TagArray[0], directory.TagArray[1]
                        }));

            cursor.Enter("3");
            cursor.UpdateCurrentScope();
            Assert.AreEqual(cursor.CurrentScope.Dewey, new Dewey(new ushort[] { 0, 0, 2 }));
            Assert.AreEqual(cursor.CurrentScope.PathType.Depth, 3);
            Assert.AreEqual(cursor.CurrentScope.PathType.LastTag.Name, "3");
            Assert.AreEqual(cursor.CurrentScope.PathType.LastTag.TagId, 2);
            Assert.That(
                cursor.CurrentScope.PathType.SequenceEqual(
                    new[]
                        {
                            directory.TagArray[0], directory.TagArray[1], directory.TagArray[2]
                        }));
        }

        [Test]
        public void TestSerialization()
        {
            var directory = new ScopeDirectory();
            ScopeDirectory.Cursor cursor = directory.GetCursor();
            cursor.Enter("a");
            cursor.Enter("b");
            cursor.Enter("c");
            cursor.Exit();
            cursor.Enter("d");
            cursor.Exit();
            cursor.Exit();
            cursor.Exit();

            MemoryStream stream = new MemoryStream(1000000);
            BinaryFormatter formatter = new BinaryFormatter();
            formatter.Serialize(stream, directory);
            stream.Seek(0, SeekOrigin.Begin);
            var deserializedDirectory = formatter.Deserialize(stream) as ScopeDirectory;
            Assert.IsNotNull(deserializedDirectory, "Deserialized should not be null.");
        }
    }
}