﻿namespace NETing.Index.Documents.Parser
{
    using System.Collections.Generic;
    using DocumentModel;
    using DocumentModel.Scope;
    using NUnit.Framework;
    using Utils.Testing;

    [TestFixture]
    public class DocumentParserTest
    {
        private ScopeDirectory ScopeDirectory { get; set; }

        private DocumentParser Parser { get; set; }

        [TestFixtureSetUp]
        public void SetupFixture()
        {
            ScopeDirectory = new ScopeDirectory();
            Parser = new DocumentParser(ScopeDirectory);
        }

        private void AssertNextToken(
            IEnumerator<Token> parsed, 
            string term, 
            int position, 
            string[] path,
            ushort[] dewey)
        {
            Assert.That(parsed.MoveNext());
            Assert.AreEqual(parsed.Current.Term, new Term(term));
            Assert.AreEqual(parsed.Current.Position, position);
            Assert.AreEqual(
                parsed.Current.Scope, 
                new Scope(ScopeDirectory.GetPath(path), new Dewey(dewey)));
        }

        [Test]
        public void TestDocumentParsing()
        {
            var parsed = Parser.ParseDocument("<?xml version=\"1.0\" encoding=\"utf-8\"?><a><b hello=\"this thus\">hello</b>hi<c n=\"1\" k=\"2\">you<d>all</d></c>nice</a>".AsStream());

            AssertNextToken(parsed, "this", 0, new[] { "a", "b", "hello" }, new ushort[] { 0, 0, 0 });
            AssertNextToken(parsed, "thus", 1, new[] { "a", "b", "hello" }, new ushort[] { 0, 0, 0 });
            AssertNextToken(parsed, "hello", 2, new[] { "a", "b" }, new ushort[] { 0, 0, 0 });
            AssertNextToken(parsed, "hi", 3, new[] { "a" }, new ushort[] { 0, 1 });
            AssertNextToken(parsed, "1", 4, new[] { "a", "c", "n" }, new ushort[] { 0, 2, 0 });
            AssertNextToken(parsed, "2", 5, new[] { "a", "c", "k" }, new ushort[] { 0, 2, 0 });
            AssertNextToken(parsed, "you", 6, new[] { "a", "c" }, new ushort[] { 0, 2, 0 });
            AssertNextToken(parsed, "all", 7, new[] { "a", "c", "d" }, new ushort[] { 0, 2, 1, 0 });
            AssertNextToken(parsed, "nice", 8, new[] { "a" }, new ushort[] { 0, 3 });

            parsed = Parser.ParseDocument(@"<a><b><c><d>this<e>is</e></d>a<c>test</c></c></b>that<c>this</c>works</a>".AsStream());
            AssertNextToken(parsed, "this", 0, new[] { "a", "b", "c", "d" }, new ushort[] { 0, 0, 0, 0, 0 });
            AssertNextToken(parsed, "is", 1, new[] { "a", "b", "c", "d", "e" }, new ushort[] { 0, 0, 0, 0, 1, 0 });
            AssertNextToken(parsed, "a", 2, new[] { "a", "b", "c" }, new ushort[] { 0, 0, 0, 1 });
            AssertNextToken(parsed, "test", 3, new[] { "a", "b", "c", "c" }, new ushort[] { 0, 0, 0, 2, 0 });
            AssertNextToken(parsed, "that", 4, new[] { "a" }, new ushort[] { 0, 1 });
            AssertNextToken(parsed, "this", 5, new[] { "a", "c" }, new ushort[] { 0, 2, 0 });
            AssertNextToken(parsed, "works", 6, new[] { "a" }, new ushort[] { 0, 3 });
        }
    }
}


