﻿namespace NETing.Index.Utils.Collections
{
    using System;
    using System.Collections.Generic;
    using System.Linq;
    
    using NUnit.Framework;
    
    [TestFixture]
    public class BinaryHeapTest
    {
        [Test]
        public void TestBuildHeap()
        {
            int[] array = new[] { 1, 5, 9, 11, 2, 1, 1, 90, 2, 5 };
            var heap = new BinaryMaxHeap<int>(array);
            IEnumerator<int> sorted = array.OrderByDescending(i => i).ToList().GetEnumerator();
            for (int i = 0; i < array.Length; i++)
            {
                Assert.That(sorted.MoveNext());
                Assert.AreEqual(sorted.Current, heap.ExtractMax());
            }
        }

        private sealed class IntHolder : IComparable<IntHolder>
        {
            public int Value;

            public IntHolder(int value)
            {
                Value = value;
            }

            public int CompareTo(IntHolder other)
            {
                return Value.CompareTo(other.Value);
            }

            public override string ToString()
            {
                return Value.ToString();
            }

            public override bool Equals(object obj)
            {
                return Equals(obj as IntHolder);
            }

            public bool Equals(IntHolder other)
            {
                if (ReferenceEquals(null, other)) return false;
                if (ReferenceEquals(this, other)) return true;
                return other.Value == Value;
            }

            public override int GetHashCode()
            {
                return Value;
            }
        }

        [Test]
        public void TestIncreaseKey()
        {
            var array = new[] {new IntHolder(1), new IntHolder(2), new IntHolder(3), new IntHolder(4)};
            var heap = new BinaryMaxHeap<IntHolder>(array);

            Assert.AreEqual(heap.Max, new IntHolder(4));
            array[3].Value = 4;
            heap.IncreaseKey(3);
            Assert.AreEqual(heap.Max, new IntHolder(4));
            array[2].Value = 8;
            heap.IncreaseKey(2);
            Assert.AreEqual(heap.Max, new IntHolder(8));
        }
    }
}
