﻿namespace NETing.Index.ColumnAccess.Implementations.NPfor.ColumnWise
{
    using System;
    using Documents.DocumentModel.Scope;
 
    using Utils.Compression;
    
    using Writers;

    public sealed class CNPforScopeColumnWriter : AbstractScopeColumnWriter
    {
        private readonly NPforEncodingWrapper wrapper;
        private readonly ScopeDirectory scopeDirectory;

        private readonly uint[][] deweyColumnEncode = new uint[16][];

        private uint[] valuesWrapped = new uint[1024];
        private int deweyStartOffset;

        internal CNPforScopeColumnWriter(
            NPforEncodingWrapper wrapper, ScopeDirectory scopeDirectory)
        {
            this.wrapper = wrapper;
            this.scopeDirectory = scopeDirectory;

            for (int i = 0; i < deweyColumnEncode.Length; i++)
            {
                deweyColumnEncode[i] = new uint[128];
            }
        }

        public override int MaxHeaderBytes
        {
            get { return 4; }
        }

        public override int WriteHeader(byte[] buffer, int offset)
        {
            return PrefixedVByte.Encode(deweyStartOffset, buffer, offset);
        }

        public override int Write(byte[] buffer, int offset)
        {
            if (valuesWrapped.Length < count)
            {
                valuesWrapped = new uint[(count + 128) << 1];
            }

            for (int i = 0; i < count; i++)
            {
                valuesWrapped[i] = pathTypes[i];
            }

            int beginning = offset;
            offset = wrapper.EncodeMultiple(buffer, offset, valuesWrapped, count);
            deweyStartOffset = offset - beginning;

            foreach (uint[] t in deweyColumnEncode)
            {
                Array.Clear(t, 0, 128);
            }

            int occ = 0;
            int groupMax = 0;
            int deweyOffset = 0;
            while (occ < count)
            {
                var pos = new int[deweyColumnEncode.Length];
                int lastOcc = Utils.Min(occ + 128, count);
                for (; occ < lastOcc; occ++)
                {
                    int depth = scopeDirectory[pathTypes[occ]].DeweyDepth;
                    for (int i = 0; i < depth; i++)
                    {
                        deweyColumnEncode[i][pos[i]++] = deweyElements[deweyOffset++];
                    }

                    groupMax = Utils.Max(groupMax, depth);
                }

                offset = PrefixedVByte.Encode(groupMax, buffer, offset);
                int fullPrefix;
                for (fullPrefix = 0;
                     fullPrefix < groupMax && pos[fullPrefix] > wrapper.PadLimit; 
                     fullPrefix++)
                {
                }

                offset = PrefixedVByte.Encode(fullPrefix, buffer, offset);
                for (int i = fullPrefix; i < groupMax; i++)
                {
                    offset = PrefixedVByte.Encode(pos[i], buffer, offset);
                }

                for (int i = 0; i < groupMax; i++)
                {
                    offset = wrapper.EncodeSingle(
                        buffer, offset, deweyColumnEncode[i], 0, pos[i]);
                }
            }
            
            return offset;
        }
    }
}
