﻿namespace NETing.Index.ColumnAccess.Implementations.NPfor.ColumnWise
{
    using Documents.DocumentModel.Scope;
 
    using PostingFile;
    
    using Readers;
    using Utils.Collections;

    public class DynamicCNPForScopeColumnReader : AbstractScopeColumnReader
    {
        private readonly NPforScopeColumnReader rowwiseReader;
        private readonly CNPforScopeColumnReader columnwiseReader;

        private sealed class ChunkEnumerator : AbstractScopeChunkEnumerator
        {
            private readonly AbstractScopeChunkEnumerator rowwiseEnumerator;
            private readonly AbstractScopeChunkEnumerator columnwiseEnumerator;

            public ChunkEnumerator(
                ScopeDirectory scopeDirectory, 
                AbstractScopeColumnReader rowwiseReader,
                AbstractScopeColumnReader columnwiseReader) 
                : base(scopeDirectory)
            {
                rowwiseEnumerator = rowwiseReader.CreateEnumerator();
                columnwiseEnumerator = columnwiseReader.CreateEnumerator();
            }

            internal override MutableBitArray PathTypeMask
            {
                set
                {
                    rowwiseEnumerator.PathTypeMask = value;
                    columnwiseEnumerator.PathTypeMask = value;
                }
            }

            public override bool MoveNext()
            {
                return false;
            }

            public override void Reset()
            {
                base.Reset();
                rowwiseEnumerator.Reset();
                columnwiseEnumerator.Reset();
            }

            public override int ReadHeader(byte[] buffer, int offset, uint totalOccurrences)
            {
                if (totalOccurrences < NPforSettings.PadLimit)
                {
                    return rowwiseEnumerator.ReadHeader(buffer, offset, totalOccurrences);
                }

                return columnwiseEnumerator.ReadHeader(buffer, offset, totalOccurrences);
            }

            public override AbstractScopeChunkEnumerator LoadChunk(PostingFileDataReader dataReader, long filePosition, uint chunkLength, uint occurrences)
            {
                if (occurrences < NPforSettings.PadLimit)
                {
                    return rowwiseEnumerator.LoadChunk(dataReader, filePosition, chunkLength, occurrences);
                }

                return columnwiseEnumerator.LoadChunk(dataReader, filePosition, chunkLength, occurrences);
            }
        }

        public DynamicCNPForScopeColumnReader(int padLimit = NPforSettings.PadLimit)
        {
            rowwiseReader = new NPforScopeColumnReader(padLimit);
            columnwiseReader = new CNPforScopeColumnReader(padLimit);
        }
        
        public override AbstractScopeChunkEnumerator CreateEnumerator()
        {
            rowwiseReader.ScopeDirectory = ScopeDirectory;
            columnwiseReader.ScopeDirectory = ScopeDirectory;
            return new ChunkEnumerator(ScopeDirectory, rowwiseReader, columnwiseReader);
        }
    }
}
