﻿namespace NETing.Index.ColumnAccess.Implementations.NPfor.ColumnWise
{
    using Documents.DocumentModel.Scope;
    
    using Writers;

    public sealed class DynamicCNPForScopeColumnWriter : AbstractScopeColumnWriter
    {
        private readonly NPforScopeColumnWriter rowwiseWriter;
        private readonly CNPforScopeColumnWriter columnwiseWriter;

        internal DynamicCNPForScopeColumnWriter(NPforEncodingWrapper wrapper, ScopeDirectory scopeDirectory)
        {
            rowwiseWriter = new NPforScopeColumnWriter(wrapper, scopeDirectory);
            columnwiseWriter = new CNPforScopeColumnWriter(wrapper, scopeDirectory);
        }

        public override int MaxHeaderBytes
        {
            get { return count < NPforSettings.PadLimit ? rowwiseWriter.MaxHeaderBytes : columnwiseWriter.MaxHeaderBytes; }
        }

        public override int WriteHeader(byte[] buffer, int offset)
        {
            if (count < NPforSettings.PadLimit)
            {
                return rowwiseWriter.WriteHeader(buffer, offset);
            }

            return columnwiseWriter.WriteHeader(buffer, offset);
        }

        public override int Write(byte[] buffer, int offset)
        {
            if (count < NPforSettings.PadLimit)
            {
                rowwiseWriter.Reset();
                rowwiseWriter.Add(pathTypes, deweyElements, count, deweyCount);
                return rowwiseWriter.Write(buffer, offset);
            }

            columnwiseWriter.Reset();
            columnwiseWriter.Add(pathTypes, deweyElements, count, deweyCount);
            return columnwiseWriter.Write(buffer, offset);
        }

        public override void Reset()
        {
            base.Reset();
            if (count < NPforSettings.PadLimit)
            {
                rowwiseWriter.Reset();
            }
            else
            {
                columnwiseWriter.Reset();
            }
        }
    }
}
