﻿namespace NETing.Index.ColumnAccess.Implementations.NPfor.ColumnWise
{
    using System;
    using Documents.DocumentModel.Scope;
 
    using Utils.Compression;

    using Writers;

    public sealed class PfxCNPforScopeColumnWriter : AbstractScopeColumnWriter
    {
        private readonly NPforEncodingWrapper wrapper;
        private readonly ScopeDirectory scopeDirectory;

        private readonly uint[] prefixLengths = new uint[128];
        private readonly uint[][] deweyColumnEncode = new uint[16][];
        private readonly int[] offsets = new int[16];
        private readonly int[] columnCounts = new int[16];

        private uint[] valuesWrapped = new uint[1024];

        private int maxDepth;
        private int deweyStartOffset;
        private int columnsStartOffset;

        internal PfxCNPforScopeColumnWriter(ScopeDirectory scopeDirectory, NPforEncodingWrapper wrapper)
        {
            this.scopeDirectory = scopeDirectory;
            this.wrapper = wrapper;

            for (int i = 0; i < deweyColumnEncode.Length; i++)
            {
                deweyColumnEncode[i] = new uint[128];
            }
        }

        public override int MaxHeaderBytes
        {
            get { return 8 + (maxDepth << 3); }
        }

        public override int WriteHeader(byte[] buffer, int offset)
        {
            offset = PrefixedVByte.Encode(maxDepth, buffer, offset);
            offset = PrefixedVByte.Encode(deweyStartOffset, buffer, offset);
            offset = PrefixedVByte.Encode(columnsStartOffset, buffer, offset);
            for (int i = 0; i < maxDepth - 1; i++)
            {
                offset = PrefixedVByte.Encode(offsets[i], buffer, offset);
            }

            for (int i = 0; i < maxDepth; i++)
            {
                offset = PrefixedVByte.Encode(columnCounts[i], buffer, offset);
            }

            return offset;
        }

        public override int Write(byte[] buffer, int offset)
        {
            if (valuesWrapped.Length < count)
            {
                valuesWrapped = new uint[(count + 128) << 1];
            }

            for (int i = 0; i < count; i++)
            {
                valuesWrapped[i] = pathTypes[i];
            }

            int beginning = offset;
            offset = wrapper.EncodeMultiple(buffer, offset, valuesWrapped, count);
            deweyStartOffset = offset - beginning;
            int deweyStartPosition = offset;

            for (int i = 0; i < deweyColumnEncode.Length; i++)
            {
                if (deweyColumnEncode[i].Length < count + 128)
                {
                    deweyColumnEncode[i] = new uint[(count + 128) << 1];
                }
            }

            int occ = 0;
            
            maxDepth = 0;
            int deweyOffset = 0;

            Array.Clear(columnCounts, 0, 16);

            while (occ < count)
            {
                int previousCount = 16;
                var previous = new ushort[16];
                var chunkCounts = new uint[16];
                int entry = 0;
                int lastOcc = Utils.Min(occ + 128, count);
                int chunkMaxDepth = 0;
                for (; occ < lastOcc; occ++)
                {
                    int depth = scopeDirectory[pathTypes[occ]].DeweyDepth;
                    int limit = Utils.Min(depth, previousCount);
                    uint prefixLength;
                    for (prefixLength = 0;
                         prefixLength < limit && deweyElements[deweyOffset] == previous[prefixLength]; 
                         prefixLength++, deweyOffset++)
                    {
                        previous[prefixLength] = deweyElements[deweyOffset];
                    }

                    prefixLengths[entry++] = prefixLength;

                    for (uint i = prefixLength; i < depth; i++)
                    {
                        previous[i] = deweyElements[deweyOffset];
                        deweyColumnEncode[i][columnCounts[i]++] = deweyElements[deweyOffset++];
                        chunkCounts[i]++;
                    }

                    previousCount = depth;
                    maxDepth = Utils.Max(maxDepth, depth);
                    chunkMaxDepth = Utils.Max(chunkMaxDepth, depth);
                }

                offset = PrefixedVByte.Encode(chunkMaxDepth, buffer, offset);
                for (int i = 0; i < chunkMaxDepth; i++)
                {
                    offset = PrefixedVByte.Encode(chunkCounts[i], buffer, offset);
                }

                offset = wrapper.EncodeSingle(buffer, offset, prefixLengths, 0, entry);
            }

            columnsStartOffset = offset - deweyStartPosition;

            for (int i = 0; i < maxDepth; i++)
            {
                int begin = offset;
                offset = wrapper.EncodeMultiple(buffer, offset, deweyColumnEncode[i], columnCounts[i]);
                offsets[i] = offset - begin;
            }

            return offset;
        }
    }
}
