﻿namespace NETing.Index.ColumnAccess.Implementations.NPfor
{
    using PostingFile;
 
    using Readers;

    using Utils.Compression;

    public sealed class NPforDocumentColumnReader : AbstractDocumentColumnReader
    {
        private sealed class ChunkEnumerator : AbstractDocumentChunkEnumerator
        {
            private readonly byte[] dataBuffer = new byte[128 << 3];
            private readonly NPforDecodingWrapper wrapper;

            private int position;
            private int nextPosition;

            private bool loaded;
            private uint maxDocumentId;

            private readonly uint[] documentIdBuffer = new uint[128 + 28];
            private readonly uint[] frequencyBuffer = new uint[128];
            private readonly uint[] occurrenceIdBuffer = new uint[128];

            public ChunkEnumerator(NPforDecodingWrapper wrapper)
            {
                this.wrapper = wrapper;
            }

            public override int ReadHeader(byte[] buffer, int offset, uint totalOccurrences)
            {
                return PrefixedVByte.Decode(buffer, offset, out maxDocumentId);
            }

            public override bool MoveNext()
            {
                if (!loaded)
                {
                    DecodeData();
                }

                position = nextPosition++;
                CurrentOccurrenceId = (int)occurrenceIdBuffer[position];

                return position < TotalDocuments;
            }

            public override int SkipTo(uint documentId)
            {
                if (documentId > maxDocumentId)
                {
                    return TotalOccurrences + 1;
                }

                if (!loaded)
                {
                    DecodeData();
                }

                int index;
                for (index = nextPosition; 
                     index < TotalDocuments & documentIdBuffer[index] < documentId; 
                     index++)
                {
                }
                
                position = index;
                CurrentOccurrenceId = (int)occurrenceIdBuffer[position];
                nextPosition = position + 1;
                return CurrentOccurrenceId;
            }

            public override ushort Frequency
            {
                get { return (ushort)frequencyBuffer[position]; }
            }

            public override uint Current
            {
                get { return documentIdBuffer[position]; }
            }

            public override int RemainingDocuments
            {
                get { return TotalDocuments - position - 1; }
            }

            public void Reset()
            {
                CurrentOccurrenceId = -1;
                position = -1;
                nextPosition = 0;
            }

            public override void LoadChunk(
                PostingFileDataReader dataReader, 
                long documentFilePosition, 
                uint documentChunkLength, 
                long frequencyFilePosition, 
                uint frequencyChunkLength, 
                int chunkCount, 
                uint occurrences)
            {
                uint totalBytes = documentChunkLength + frequencyChunkLength;
                if (frequencyFilePosition == documentFilePosition + documentChunkLength)
                {
                    dataReader.LoadData(
                        dataBuffer,
                        0,
                        (int)totalBytes,
                        documentFilePosition);
                }
                else
                {
                    dataReader.LoadData(
                        dataBuffer, 0, (int)documentChunkLength, documentFilePosition);
                    dataReader.LoadData(
                        dataBuffer,
                        (int)documentChunkLength,
                        (int)frequencyChunkLength,
                        frequencyFilePosition);
                }

                loaded = false;
                SetupChunk(chunkCount, occurrences);
                Reset();
            }

            private void DecodeData()
            {
                int offset = wrapper.DecodeSingle(dataBuffer, 0, TotalDocuments, documentIdBuffer, 0);
                wrapper.DecodeSingle(dataBuffer, offset, TotalDocuments, frequencyBuffer, 0);

                frequencyBuffer[0]++;
                occurrenceIdBuffer[0] = 0;
                for (int i = 1; i < TotalDocuments; i++)
                {
                    documentIdBuffer[i] += documentIdBuffer[i - 1] + 1;
                    occurrenceIdBuffer[i] = occurrenceIdBuffer[i - 1] + frequencyBuffer[i - 1];
                    frequencyBuffer[i]++;
                }

                loaded = true;
            }
        }

        public override AbstractDocumentChunkEnumerator CreateEnumerator()
        {
            return new ChunkEnumerator(new NPforDecodingWrapper());
        }
    }
}
