﻿namespace NETing.Index.ColumnAccess.Implementations.NPfor
{
    using System;
    using Utils.ByteArrays;
    using Utils.Compression;
    using Utils.Compression.Patched;

    internal class NPforEncodingWrapper
    {
        private readonly uint[] intermediateBuffer = new uint[128];
        private readonly uint[] outputBuffer = new uint[128];
        private readonly uint[] exceptions = new uint[128];
        private readonly uint[] exceptionIndexes = new uint[128];

        private readonly uint[] simple9Buffer = new uint[256];

        private readonly int padLimit;

        public int PadLimit
        {
            get { return padLimit; }
        }

        public NPforEncodingWrapper(int padLimit = NPforSettings.PadLimit)
        {
            this.padLimit = padLimit;
        }

        public int EncodeMultiple(
            byte[] buffer, int offset, uint[] values, int count)
        {
            for (int i = 0; i < count; i += 128)
            {
                offset = EncodeSingle(buffer, offset, values, i, Utils.Min(128, count - i));    
            }

            return offset;
        }

        public int EncodeSingle(
            byte[] buffer, int offset, uint[] values, int valuesOffset, int count)
        {
            if (count < padLimit)
            {
                return EncodeWithFallbackMechanism(buffer, offset, values, valuesOffset, count);
            }

            uint frame = PforUtils.MinimumValueFrameOfReference(count, values, valuesOffset);

            PadRemainingSlots(frame, count, valuesOffset, values);

            int width = PforUtils.Log2HistogramAnalyseBits(NPforSettings.InRangeRequirement, frame, values, valuesOffset);

            Array.Clear(outputBuffer, 0, 128);
            int numException = NewPFor.Encode(
                width, frame, values, intermediateBuffer, valuesOffset, outputBuffer, exceptionIndexes, exceptions);

            int outputWord = 0;
            for (int i = 0;
                 i < numException;
                 i += Simple9.Encode(exceptions, simple9Buffer, i, outputWord), outputWord++)
            {
            }

            for (int i = 0;
                 i < numException;
                 i += Simple9.Encode(exceptionIndexes, simple9Buffer, i, outputWord), outputWord++)
            {
            }

            int codeBytes = width << 4;
            buffer.SetValueAt(offset, PforUtils.CreateControlWord((uint)width, (uint)numException, frame));
            offset += 4;
            BlockCopy.Copy(outputBuffer, 0, buffer, offset, codeBytes);
            offset += codeBytes;
            BlockCopy.Copy(simple9Buffer, 0, buffer, offset, outputWord << 2);

            return offset + (outputWord << 2);
        }

        private static int EncodeWithFallbackMechanism(byte[] buffer, int offset, uint[] values, int valuesOffset, int count)
        {
            for (int i = 0; i < count; i++)
            {
                offset = PrefixedVByte.Encode(values[valuesOffset + i], buffer, offset);
            }

            return offset;
        }

        private static void PadRemainingSlots(uint frameOfReference, int startAt, int valuesOffset, uint[] values)
        {
            for (int i = startAt; i < 128; i++)
            {
                values[valuesOffset + i] = frameOfReference;
            }
        }
    }
}
