﻿namespace NETing.Index.ColumnAccess.Implementations.NPfor
{
    using PostingFile;
 
    using Readers;
    
    using Utils.Compression;

    public sealed class NPforPositionColumnReader : AbstractPositionColumnReader
    {
        private sealed class ChunkEnumerator : AbstractPositionChunkEnumerator
        {
            private byte[] dataBuffer = new byte[1024];
            private int dataBufferOffset;

            private readonly NPforDecodingWrapper wrapper;
            private readonly uint[] decodedBuffer = new uint[128 + 28];

            private int previousPosition = -1;
            private int inBufferPosition;
            private int nextInBufferPosition;
            
            public ChunkEnumerator(NPforDecodingWrapper wrapper)
            {
                this.wrapper = wrapper;
            }

            public override bool MoveNext()
            {
                if (nextInBufferPosition < 0)
                {
                    ReadOneBuffer();
                    nextInBufferPosition = 0;
                }

                CurrentOccurrenceId++;
                inBufferPosition = nextInBufferPosition++;
                decodedBuffer[inBufferPosition] += (uint)(previousPosition + 1);
                previousPosition = (int)decodedBuffer[inBufferPosition];

                if (inBufferPosition >= 127)
                {
                    nextInBufferPosition = -1;
                }

                return CurrentOccurrenceId < TotalOccurrences;
            }

            public override void MoveTo(int documentStartPosition, int withinDocumentPosition, uint documentFrequency)
            {
                if (documentStartPosition > CurrentOccurrenceId)
                {
                    int remainingInBuffer = 128 - inBufferPosition - 1;
                    int documentDistance = documentStartPosition - CurrentOccurrenceId;

                    if (documentDistance <= remainingInBuffer)
                    {
                        if (nextInBufferPosition < 0)
                        {
                            ReadOneBuffer();
                            nextInBufferPosition = 0;
                        }

                        CurrentOccurrenceId += documentDistance;
                    }
                    else
                    {
                        CurrentOccurrenceId += remainingInBuffer % 128;
                        nextInBufferPosition = -1;
                        documentDistance -= remainingInBuffer % 128;

                        while (documentDistance > 128)
                        {
                            SkipOneBuffer();
                            CurrentOccurrenceId += 128;
                            documentDistance -= 128;
                        }

                        if (nextInBufferPosition < 0)
                        {
                            ReadOneBuffer();
                            nextInBufferPosition = 0;
                        }

                        inBufferPosition = -1;
                        CurrentOccurrenceId = documentStartPosition;
                    }

                    inBufferPosition += documentDistance;
                    nextInBufferPosition = inBufferPosition + 1;
                    previousPosition = Current;

                    if (nextInBufferPosition > 127)
                    {
                        nextInBufferPosition = -1;
                    }
                }

                while (CurrentOccurrenceId < withinDocumentPosition)
                {
                    MoveNext();
                }
            }

            public override ushort Current
            {
                get { return (ushort)decodedBuffer[inBufferPosition]; }
            }

            public override void NewDocument(uint frequency)
            {
                previousPosition = -1;
            }

            public void Reset()
            {
                dataBufferOffset = 0;
                CurrentOccurrenceId = -1;
                inBufferPosition = -1;
                previousPosition = -1;
                nextInBufferPosition = -1;
            }

            public override void LoadChunk(
                PostingFileDataReader dataReader, 
                long filePosition, 
                uint chunkLength, 
                uint occurrences)
            {
                if (dataBuffer.Length < chunkLength)
                {
                    dataBuffer = new byte[chunkLength << 1];
                }

                dataReader.LoadData(dataBuffer, 0, (int)chunkLength, filePosition);
                SetupChunk(0, occurrences);

                Reset();
            }

            private void ReadOneBuffer()
            {
                dataBufferOffset = wrapper.DecodeSingle(
                    dataBuffer, 
                    dataBufferOffset,
                    Utils.Min(128, TotalOccurrences - CurrentOccurrenceId - 1), 
                    decodedBuffer, 
                    0);
            }

            private void SkipOneBuffer()
            {
                dataBufferOffset = wrapper.SkipSingle(
                    dataBuffer,
                    dataBufferOffset,
                    Utils.Min(128, TotalOccurrences - CurrentOccurrenceId - 1));
            }
        }

        public override AbstractPositionChunkEnumerator CreateEnumerator()
        {
            return new ChunkEnumerator(new NPforDecodingWrapper());
        }
    }
}
