﻿namespace NETing.Index.ColumnAccess.Implementations.NPfor
{
    using System;
    using Utils.ByteArrays;
    using Writers;

    public sealed class NPforPostionColumnWriter : AbstractPositionColumnWriter
    {
        private readonly NPforEncodingWrapper wrapper;
        private uint[] wrappedValues = new uint[1024];

        internal NPforPostionColumnWriter(NPforEncodingWrapper wrapper)
        {
            this.wrapper = wrapper;
        }

        public override void Add(ushort[] values, int count)
        {
            if (this.values.Length < this.count + count + 128)
            {
                Array.Resize(ref this.values, (this.count + count + 128) << 1);
            }

            for (int i = count - 1; i > 0; i--)
            {
                values[i] -= (ushort)(values[i - 1] + 1);
            }

            BlockCopy.Copy(values, 0, this.values, this.count << 1, count << 1);
            this.count += count;
        }

        public override int Write(byte[] buffer, int offset)
        {
            if (wrappedValues.Length < count + 128)
            {
                wrappedValues = new uint[(count << 1) + 128];    
            }

            for (int i = 0; i < count; i++)
            {
                wrappedValues[i] = values[i];
            }

            return wrapper.EncodeMultiple(buffer, offset, wrappedValues, count);
        }
    }
}
