﻿namespace NETing.Index.ColumnAccess.Implementations.NPfor
{
    using Documents.DocumentModel.Scope;

    using Utils.Compression;

    using Writers;

    public sealed class NPforScopeColumnWriter : AbstractScopeColumnWriter
    {
        private readonly ScopeDirectory scopeDirectory;
        private readonly NPforEncodingWrapper wrapper;

        private uint[] valuesWrapped = new uint[1024 << 2];
        
        private int deweyOffset;

        internal NPforScopeColumnWriter(
            NPforEncodingWrapper wrapper, ScopeDirectory scopeDirectory)
        {
            this.wrapper = wrapper;
            this.scopeDirectory = scopeDirectory;
        }

        public override int MaxHeaderBytes
        {
            get { return 8 + (count >> 6) + 2; }
        }

        public override int WriteHeader(byte[] buffer, int offset)
        {
            int chunks = (count >> 7) + ((count & 127) > 0 ? 1 : 0);
            int pos = 0;
            for (int i = 0; i < chunks; i++)
            {
                int chunkCount = 0;
                int limit = pos + Utils.Min(128, count - pos);
                for (; pos < limit; pos++)
                {
                    chunkCount += scopeDirectory[pathTypes[pos]].DeweyDepth;
                }

                offset = PrefixedVByte.Encode(chunkCount, buffer, offset);
            }

            return PrefixedVByte.Encode(deweyOffset, buffer, offset);
        }
        
        public override int Write(byte[] buffer, int offset)
        {
            if (valuesWrapped.Length < deweyCount + 128)
            {
                valuesWrapped = new uint[(deweyCount + 128) << 1];
            }

            for (int i = 0; i < count; i++)
            {
                valuesWrapped[i] = pathTypes[i];
            }

            int beginning = offset;
            offset = wrapper.EncodeMultiple(buffer, offset, valuesWrapped, count);
            deweyOffset = offset - beginning;

            for (int i = 0; i < deweyCount; i++)
            {
                valuesWrapped[i] = deweyElements[i];
            }

            return wrapper.EncodeMultiple(buffer, offset, valuesWrapped, deweyCount);
        }
    }
}
