﻿namespace NETing.Index.ColumnAccess.Implementations.Raw
{
    using System.Runtime.InteropServices;
    
    using PostingFile;
    using Readers;

    public sealed class RawPositionColumnReader : AbstractPositionColumnReader
    {
        private sealed class ChunkEnumerator : AbstractPositionChunkEnumerator
        {
            [StructLayout(LayoutKind.Explicit)]
            private struct BufferUnion
            {
                [FieldOffset(0)] public byte[] ByteBuffer;
                [FieldOffset(0)] public readonly ushort[] UshortBuffer;
            }

            private BufferUnion buffers = new BufferUnion { ByteBuffer = new byte[128 << 3] }; // four positions for each occ. 
            private ushort[] dataBuffer;
            private int nextOccurrence;

            public ChunkEnumerator()
            {
                dataBuffer = buffers.UshortBuffer;
            }

            public override bool MoveNext()
            {
                CurrentOccurrenceId = nextOccurrence++;
                return CurrentOccurrenceId < TotalOccurrences;
            }

            public override ushort Current
            {
                get { return dataBuffer[CurrentOccurrenceId]; }
            }

            public override void LoadChunk(
                PostingFileDataReader dataReader, 
                long filePosition, 
                uint chunkLength, 
                uint occurrences)
            {
                if (buffers.UshortBuffer.Length < occurrences << 1)
                {
                    buffers.ByteBuffer = new byte[occurrences << 2];
                    dataBuffer = buffers.UshortBuffer;
                }

                dataReader.LoadData(buffers.ByteBuffer, 0, (int)(occurrences << 1), filePosition);
                SetupChunk(occurrences);
            }

            public void Reset()
            {
                nextOccurrence = 0;
                CurrentOccurrenceId = -1;
            }

            private void SetupChunk(uint occurrences)
            {
                SetupChunk(0, occurrences);
                Reset();
            }
        }
        
        public override AbstractPositionChunkEnumerator CreateEnumerator()
        {
            return new ChunkEnumerator();
        }
    }
}
