﻿namespace NETing.Index.ColumnAccess.Implementations.Raw
{
    using System.Runtime.InteropServices;
 
    using Documents.DocumentModel.Scope;
    
    using PostingFile;
    
    using Readers;

    public sealed class RawScopeColumnReader : AbstractScopeColumnReader
    {
        private sealed class ChunkEnumerator : AbstractScopeChunkEnumerator
        {
            [StructLayout(LayoutKind.Explicit)]
            private struct BufferUnion
            {
                [FieldOffset(0)] public byte[] ByteBuffer;
                [FieldOffset(0)] public readonly ushort[] UshortBuffer;
            }

            private BufferUnion buffers = new BufferUnion { ByteBuffer = new byte[1024]};
            private ushort[] dataBuffer;

            private int nextPosition;

            private int nextDeweyStart;
            private readonly Dewey deweyHolder;
            
            public ChunkEnumerator(ScopeDirectory scopeDirectory) 
                : base(scopeDirectory)
            {
                deweyHolder = new Dewey(buffers.UshortBuffer);
                holder.Dewey = deweyHolder;
                dataBuffer = buffers.UshortBuffer;
            }

            public override bool MoveNext()
            {
                CurrentOccurrenceId = nextPosition++;
                PathType pathType = pathTypes[dataBuffer[CurrentOccurrenceId]];
                deweyHolder.Offset = nextDeweyStart;
                nextDeweyStart += pathType.DeweyDepth;

                holder.PathType = pathType;
                deweyHolder.Count = pathType.DeweyDepth;

                return CurrentOccurrenceId < TotalOccurrences;
            }
            
            public override AbstractScopeChunkEnumerator LoadChunk(
                PostingFileDataReader dataReader, 
                long filePosition, 
                uint chunkLength, 
                uint occurrences)
            {
                if (buffers.ByteBuffer.Length < chunkLength)
                {
                    buffers.ByteBuffer = new byte[chunkLength << 1];
                    dataBuffer = buffers.UshortBuffer;
                    deweyHolder.DeweyArray = dataBuffer;
                }

                dataReader.LoadData(buffers.ByteBuffer, 0, (int)chunkLength, filePosition);
                SetupChunk(occurrences);

                return this;
            }

            public override void Reset()
            {
                base.Reset();

                CurrentOccurrenceId = -1;
                nextPosition = 0;

                nextDeweyStart = TotalOccurrences;
            }
        }

        public override AbstractScopeChunkEnumerator CreateEnumerator()
        {
            return new ChunkEnumerator(ScopeDirectory);
        }
    }
}
