﻿namespace NETing.Index.ColumnAccess.Implementations.VByte
{
    using PostingFile;
    
    using Readers;

    using Utils.Compression;

    public sealed class VByteDocumentColumnReader : AbstractDocumentColumnReader
    {
        private sealed class ChunkEnumerator : AbstractDocumentChunkEnumerator
        {
            private uint lastDocumentIdInChunk;

            private int previousDocumentId;
            private int documentIdPosition;
            private int documentIdLength;
            private int frequencyPosition;

            private int position;
            private int nextOccurrenceId;

            private byte[] dataBuffer = new byte[512];

            public override int ReadHeader(byte[] buffer, int offset, uint totalOccurrences)
            {
                return VByte.Decode(buffer, offset, out lastDocumentIdInChunk);
            }

            public override void LoadChunk(
                PostingFileDataReader dataReader, 
                long documentFilePosition,
                uint documentChunkLength, 
                long frequencyFilePosition,
                uint frequencyChunkLength, 
                int chunkCount, 
                uint occurrences)
            {
                uint totalBytes = documentChunkLength + frequencyChunkLength;
                if (dataBuffer.Length < totalBytes)
                {
                    dataBuffer = new byte[totalBytes << 1];
                }

                if (frequencyFilePosition == documentFilePosition + documentChunkLength)
                {
                    dataReader.LoadData(
                        dataBuffer, 
                        0, 
                        (int)totalBytes, 
                        documentFilePosition);
                }
                else
                {
                    dataReader.LoadData(
                        dataBuffer, 0, (int)documentChunkLength, documentFilePosition);
                    dataReader.LoadData(
                        dataBuffer, 
                        (int)documentChunkLength, 
                        (int)frequencyChunkLength, 
                        frequencyFilePosition);
                }

                position = -1;
                CurrentOccurrenceId = -1;
                nextOccurrenceId = 0;
                previousDocumentId = -1;
                documentIdPosition = 0;
                frequencyPosition = (int)documentChunkLength;
                documentIdLength = frequencyPosition;

                frequency = 0;
                current = 0;
                SetupChunk(chunkCount, occurrences);
            }

            private ushort frequency;
            private uint current;

            public override uint Current
            {
                get { return current; }
            }

            public override int RemainingDocuments
            {
                get { return TotalDocuments - position - 1; }
            }

            public override ushort Frequency
            {
                get { return frequency; }
            }

            public override bool MoveNext()
            {
                documentIdPosition = VByte.Decode(dataBuffer, documentIdPosition, out current);
                frequencyPosition = VByte.Decode(dataBuffer, frequencyPosition, out frequency);
                ++position;
                ++frequency;
                current += (uint)(previousDocumentId + 1);
                previousDocumentId = (int)current;
                
                CurrentOccurrenceId = nextOccurrenceId;
                nextOccurrenceId += frequency;
                return documentIdPosition <= documentIdLength;
            }

            public override int SkipTo(uint documentId)
            {
                if (documentId > lastDocumentIdInChunk)
                {
                    return TotalOccurrences + 1;
                }

                return base.SkipTo(documentId);
            }
        }

        public override AbstractDocumentChunkEnumerator CreateEnumerator()
        {
            return new ChunkEnumerator();
        }
    }
}
