﻿namespace NETing.Index.ColumnAccess.Implementations.VByte
{
    using PostingFile;

    using Readers;

    using Utils.Compression;

    public sealed class VBytePositionColumnReader : AbstractPositionColumnReader
    {
        private sealed class ChunkEnumerator : AbstractPositionChunkEnumerator
        {
            private byte[] dataBuffer = new byte[128 * 10];

            private int position;
            private uint chunkLength;

            private ushort current;
            private short previousPosition;

            public override bool MoveNext()
            {
                if (position < chunkLength)
                {
                    CurrentOccurrenceId++;
                    position = VByte.Decode(dataBuffer, position, out current);
                    current += (ushort) (previousPosition + 1);
                    previousPosition = (short) current;
                    return true;
                }

                return false;
            }

            public override ushort Current
            {
                get { return current; }
            }

            public override void NewDocument(uint frequency)
            {
                previousPosition = -1;
            }

            public override void LoadChunk(
                PostingFileDataReader dataReader, 
                long filePosition, 
                uint chunkLength, 
                uint occurrences)
            {
                if (dataBuffer.Length < chunkLength)
                {
                    dataBuffer = new byte[chunkLength << 1];
                }

                dataReader.LoadData(dataBuffer, 0, (int)chunkLength, filePosition);

                position = 0;
                previousPosition = -1;
                this.chunkLength = chunkLength;
                CurrentOccurrenceId = -1;
                SetupChunk(0, occurrences);
            }
        }

        public override AbstractPositionChunkEnumerator CreateEnumerator()
        {
            return new ChunkEnumerator();
        }
    }
}
