﻿namespace NETing.Index.ColumnAccess.Implementations.VByte
{
    using Documents.DocumentModel.Scope;

    using PostingFile;

    using Readers;

    using Utils.Collections;
    using Utils.Compression;

    public sealed class VByteScopeColumnReader : AbstractScopeColumnReader
    {
        private sealed class ChunkEnumerator : AbstractScopeChunkEnumerator
        {
            private ushort extraBytes;
    
            private int pathTypePosition;
            private int deweyPosition;
            private readonly ushort[] deweyArray = new ushort[32];
            
            private byte[] dataBuffer = new byte[1024 * 4];
            
            public ChunkEnumerator(ScopeDirectory scopeDirectory) : base(scopeDirectory)
            {
                holder.Dewey = new Dewey(deweyArray);
                holder.Dewey.Offset = 0;
            }

            public override int ReadHeader(byte[] buffer, int offset, uint totalOccurrences)
            {
                extraBytes = 0;
                return VByte.Decode(buffer, offset, out extraBytes);
            }

            public override bool MoveNext()
            {
                ushort pathTypeId;
                pathTypePosition = VByte.Decode(dataBuffer, pathTypePosition, out pathTypeId);
                
                PathType pt = pathTypes[pathTypeId];
                holder.PathType = pt;
                holder.Dewey.Count = pt.DeweyDepth;

                int depth = pt.DeweyDepth;
                for (int i = 0; i < depth; i++)
                {
                    deweyPosition = VByte.Decode(dataBuffer, deweyPosition, out deweyArray[i]);
                }

                return ++CurrentOccurrenceId < TotalOccurrences;
            }

            public override void MoveTo(int targetPosition)
            {
                if (targetPosition > TotalOccurrences)
                {
                    return;
                }

                int skipOccs = targetPosition - CurrentOccurrenceId - 1;
                int deweySkip = 0;
                for (int i = 0; i < skipOccs; i++)
                {
                    ushort pathTypeId;
                    pathTypePosition = VByte.Decode(dataBuffer, pathTypePosition, out pathTypeId);
                    deweySkip += pathTypes[pathTypeId].DeweyDepth;
                }

                CurrentOccurrenceId += skipOccs;
                deweyPosition = VByte.SkipMany(dataBuffer, deweyPosition, deweySkip);
                MoveNext();
            }

            public override int SkipToPathType(MutableBitArray candidatePathTypes)
            {
                bool hasMore = (holder.PathType == null || !candidatePathTypes[holder.PathType.PathId]) && MoveNext();
                if (hasMore && !candidatePathTypes[holder.PathType.PathId])
                {
                    PathType pathType = holder.PathType ?? pathTypes[0];
                    int deweySkip = 0;
                    while (!candidatePathTypes[pathType.PathId] && CurrentOccurrenceId < TotalOccurrences)
                    {
                        ushort pathTypeId;
                        pathTypePosition = VByte.Decode(dataBuffer, pathTypePosition, out pathTypeId);
                        pathType = pathTypes[pathTypeId];
                        deweySkip += pathType.DeweyDepth;
                        CurrentOccurrenceId++;
                    }

                    int depth = pathType.DeweyDepth;
                    holder.PathType = pathType;
                    holder.Dewey.Count = pathType.DeweyDepth;

                    deweySkip -= depth;
                    deweyPosition = VByte.SkipMany(dataBuffer, deweyPosition, deweySkip);

                    for (int i = 0; i < depth; i++)
                    {
                        deweyPosition = VByte.Decode(dataBuffer, deweyPosition, out deweyArray[i]);
                    }
                }

                return CurrentOccurrenceId;
            }

            public override AbstractScopeChunkEnumerator LoadChunk(
                PostingFileDataReader dataReader, 
                long filePosition, 
                uint chunkLength, 
                uint occurrences)
            {
                if (dataBuffer.Length < chunkLength)
                {
                    dataBuffer = new byte[chunkLength << 1];
                }

                pathTypePosition = 0;
                dataReader.LoadData(dataBuffer, 0, (int) chunkLength, filePosition);
                deweyPosition = (int)(pathTypePosition + occurrences + extraBytes);
                CurrentOccurrenceId = -1;
                SetupChunk(occurrences);

                return this;
            }
        }

        public override AbstractScopeChunkEnumerator CreateEnumerator()
        {
            return new ChunkEnumerator(ScopeDirectory);
        }
    }
}
