﻿namespace NETing.Index.ColumnAccess.Readers
{
    using Documents.DocumentModel.Scope;
    
    using PostingFile;
    
    using Utils.Collections;

    public abstract class AbstractScopeChunkEnumerator : AbstractChunkEnumerator
    {
        protected readonly ScopeDirectory scopeDirectory;
        protected PathType[] pathTypes;

        protected readonly Scope holder = new Scope(null, null);

        protected AbstractScopeChunkEnumerator(ScopeDirectory scopeDirectory)
        {
            this.scopeDirectory = scopeDirectory;
        }

        public Scope Holder { get { return holder; } }

        internal virtual MutableBitArray PathTypeMask { get; set; }

        public abstract bool MoveNext();

        public virtual int SkipToPathType(MutableBitArray candidatePathTypes)
        {
            while ((holder.PathType == null 
                || !candidatePathTypes[holder.PathType.PathId]) 
                && MoveNext())
            {
            }

            return CurrentOccurrenceId;
        }

        public int SkipToScope(
            Scope scope, int matchingPrefix, int lastOccurrence, out bool foundMatch)
        {
            int cmp = holder.Dewey.PrefixCompare(scope.Dewey, matchingPrefix);
            while (cmp < 0 && CurrentOccurrenceId < lastOccurrence)
            {
                if (!MoveNext())
                {
                    foundMatch = false;
                    return CurrentOccurrenceId;
                }

                cmp = holder.Dewey.PrefixCompare(scope.Dewey, matchingPrefix);
            }

            PathType lcp = PathTypeLcp.Lcp(scope.PathType, holder.PathType);
            while (cmp <= 0 && CurrentOccurrenceId < lastOccurrence)
            {
                if (cmp == 0 && lcp.Depth >= matchingPrefix)
                {
                    break;
                }

                if (!MoveNext())
                {
                    foundMatch = false;
                    return CurrentOccurrenceId;
                }

                cmp = holder.Dewey.PrefixCompare(scope.Dewey, matchingPrefix);
                lcp = PathTypeLcp.Lcp(scope.PathType, holder.PathType);
            }

            foundMatch = cmp == 0 & lcp.Depth >= matchingPrefix;
            return CurrentOccurrenceId;
        }

        public virtual void MoveTo(int targetPosition)
        {
            if (targetPosition > TotalOccurrences)
            {
                return;
            }

            while (CurrentOccurrenceId < targetPosition)
            {
                MoveNext();
            }
        }

        public virtual void Reset()
        {
            CurrentOccurrenceId = -1;
            holder.PathType = null;
        }

        public abstract AbstractScopeChunkEnumerator LoadChunk(
            PostingFileDataReader dataReader,
            long filePosition,
            uint chunkLength,
            uint occurrences);

        protected void SetupChunk(uint occurrences)
        {
            SetupChunk(0, occurrences);

            pathTypes = scopeDirectory.PathTypes;
            Reset();
        }
    }

}