﻿namespace NETing.Index.ColumnAccess.Writers
{
    using System;
    using Utils.ByteArrays;

    public abstract class AbstractScopeColumnWriter : AbstractColumnWriter
    {
        protected int deweyCount;

        protected ushort[] pathTypes = new ushort[5 * 128];
        protected ushort[] deweyElements = new ushort[20 * 128];
        
        public int MaxBytes
        {
            get { return 4 + ((count + deweyCount) << 1); }
        }

        public override void Reset()
        {
            base.Reset();
            deweyCount = 0;
        }

        public virtual void Add(ushort[] pathTypes, ushort[] deweys, int frequency, int totalDewey)
        {
            if (this.pathTypes.Length < count + frequency)
            {
                Array.Resize(ref this.pathTypes, (frequency + count) << 1);
            }

            if (deweyElements.Length < deweyCount + totalDewey)
            {
                Array.Resize(ref deweyElements, (deweyCount + totalDewey) << 1);
            }

            BlockCopy.Copy(pathTypes, 0, this.pathTypes, count << 1, frequency << 1);
            BlockCopy.Copy(deweys, 0, deweyElements, deweyCount << 1, totalDewey << 1);

            count += frequency;
            deweyCount += totalDewey;
        }
    }
}