﻿namespace NETing.Index.Controller.Configurations.ConcreteConfigurations
{
    using ColumnAccess.Implementations.NPfor;
    using ColumnAccess.Implementations.NPfor.ColumnWise;
    using ColumnAccess.Readers;
    using ColumnAccess.Writers;

    using Documents.DocumentModel.Scope;

    public class DCNPforPostingFileConfig : IPostingFileConfig
    {
        private readonly NPforEncodingWrapper otherEncodingWrapper = new NPforEncodingWrapper();
        private readonly NPforEncodingWrapper scopeEncodingWrapper;
        private readonly int padLimit;

        public DCNPforPostingFileConfig(int padLimit = NPforSettings.PadLimit)
        {
            this.padLimit = padLimit;
            scopeEncodingWrapper = new NPforEncodingWrapper(padLimit);
        }

        public string Name
        {
            get { return "DCNPFor" + padLimit; }
        }

        public ScopeDirectory ScopeDirectory { get; set; }

        public AbstractDocumentIdColumnWriter DocumentIdWriter
        {
            get { return new NPforDocumentIdColumnWriter(otherEncodingWrapper); }
        }

        public AbstractFrequencyColumnWriter FrequencyWriter
        {
            get { return new NPforFrequencyColumnWriter(otherEncodingWrapper); }
        }

        public AbstractPositionColumnWriter PositionWriter
        {
            get { return new NPforPostionColumnWriter(otherEncodingWrapper); }
        }

        public AbstractScopeColumnWriter ScopeWriter
        {
            get { return new DynamicCNPForScopeColumnWriter(scopeEncodingWrapper, ScopeDirectory); }
        }

        public AbstractDocumentColumnReader DocumentReader
        {
            get { return new NPforDocumentColumnReader(); }
        }

        public AbstractPositionColumnReader PositionReader
        {
            get { return new NPforPositionColumnReader(); }
        }

        public AbstractScopeColumnReader ScopeReader
        {
            get { return new DynamicCNPForScopeColumnReader(padLimit); }
        }
    }
}
