﻿namespace NETing.Index.Controller.Services
{
    using System;
    using System.Collections.Generic;
    using System.IO;

    using Documents.Parser;

    public class DictionaryService
    {
        private StorageProfile StorageProfile { get; set; }
        private bool read;
        
        private readonly IDictionary<Term, Tuple<long, int>> dictionary 
            = new Dictionary<Term, Tuple<long, int>>();

        internal DictionaryService (StorageProfile storageProfile)
        {
            StorageProfile = storageProfile;
        }

        internal void Register(Term term, long position, int documentFrequency)
        {
            dictionary[term] = new Tuple<long, int>(position, documentFrequency);
        }

        internal Tuple<long, int> Locate(Term term)
        {
            Tuple<long, int> entry;
            if (dictionary.TryGetValue(term, out entry))
            {
                return entry;
            }

            return null;
        }

        public IEnumerable<Term> GetAllTerms()
        {
            return dictionary.Keys;
        }

        public void Write()
        {
            using (Stream stream = StorageProfile.DictionaryFile.OpenWrite())
            {
                using (BinaryWriter writer = new BinaryWriter(new BufferedStream(stream)))
                {
                    writer.Write(dictionary.Count);
                    foreach (var entry in dictionary)
                    {
                        writer.Write(entry.Key.ToString());
                        writer.Write(entry.Value.Item1);
                        writer.Write(entry.Value.Item2);
                    }
                }
            }
        }

        public void Read()
        {
            if (read)
            {
                return;
            }

            using (Stream stream = StorageProfile.DictionaryFile.OpenRead())
            {
                using (BinaryReader reader = new BinaryReader(new BufferedStream(stream)))
                {
                    int count = reader.ReadInt32();
                    dictionary.Clear();
                    for (int i = 0; i < count; i++)
                    {
                        Term term = new Term(reader.ReadString());
                        dictionary.Add(term, new Tuple<long, int>(
                            reader.ReadInt64(), reader.ReadInt32()));
                    }
                }
            }

            read = true;
        }
    }
}


