﻿namespace NETing.Index.Documents.DocumentModel.Scope
{
    using System;

    internal sealed class DeweyBuilder
    {
        private ushort[] dewey = new ushort[8];
        private int depth;

        /// <summary>
        /// The enter.
        /// </summary>
        public void Enter()
        {
            if (depth + 1 >= dewey.Length)
            {
                Array.Resize(ref dewey, dewey.Length * 2);
            }

            dewey[depth++]++;
        }

        /// <summary>
        /// Exit the current node.
        /// </summary>
        public void Exit()
        {
            dewey[depth--] = 0;
        }

        public void RevertingExit()
        {
            Exit();
            dewey[depth]--;
        }

        public void Build(Dewey holder)
        {
            if (holder.DeweyArray == null || holder.DeweyArray.Length < depth)
            {
                holder.DeweyArray = new ushort[depth];
            }

            for (int i = 0; i < depth; i++)
            {   
                holder.DeweyArray[i] = (ushort)(dewey[i] - 1);                
            }

            holder.Count = depth;
            holder.Offset = 0;
        }

        public void Clear()
        {
            Array.Clear(dewey, 0, dewey.Length);
            depth = 0;
        }
    }
}