﻿namespace NETing.Index.Documents.DocumentModel.Scope
{
    using System;
    using System.Collections;
    using System.Collections.Generic;
    using System.Text;

    [Serializable]
    public sealed class PathType : IEnumerable<Tag>
    {
        internal readonly Dictionary<int, PathType> Children 
            = new Dictionary<int, PathType>();

        public int PathId;

        private int depth;
        public int Depth
        {
            get
            {
                return depth;
            }

            internal set
            {
                depth = value;
                if (isAttribute)
                {
                    DeweyDepth = value;
                }
                else
                {
                    DeweyDepth = value + 1;
                }
            }
        }

        internal int DeweyDepth;

        private bool isAttribute;
        internal bool IsAttribute
        {
            get
            {
                return isAttribute;
            }
            set
            {
                if (!value)
                {
                    DeweyDepth = Depth + 1;
                }
                else
                {
                    DeweyDepth = Depth;
                }

                isAttribute = value;
            }
        }

        internal Tag LastTag;

        internal PathType Parent;

        #region IEnumerable<Tag> Members

        public IEnumerator<Tag> GetEnumerator()
        {
            if (Parent != null && Parent.PathId != 0)
            {
                foreach (Tag tag in Parent)
                {
                    yield return tag;
                }
            }

            yield return LastTag;
        }

        IEnumerator IEnumerable.GetEnumerator()
        {
            return GetEnumerator();
        }

        #endregion

        public PathType Enter(Tag tag)
        {
            PathType child;
            if (!Children.TryGetValue(tag.TagId, out child))
            {
                child = new PathType { PathId = -1, LastTag = tag, Parent = this, Depth = Depth + 1 };
                child.IsAttribute = tag.IsAttribute;
                Children.Add(tag.TagId, child);
            }

            return child;
        }

        public override string ToString()
        {
            var sb = new StringBuilder();
            foreach (Tag tag in this)
            {
                sb.Append(tag.ToString()).Append("/");
            }

            return sb.ToString();
        }

        public override bool Equals(object obj)
        {
            if (ReferenceEquals(null, obj))
            {
                return false;
            }

            if (ReferenceEquals(this, obj))
            {
                return true;
            }

            if (obj.GetType() != typeof(PathType))
            {
                return false;
            }

            return Equals((PathType) obj);
        }

        public bool Equals(PathType obj)
        {
            if (ReferenceEquals(null, obj))
            {
                return false;
            }

            if (ReferenceEquals(this, obj))
            {
                return true;
            }

            return obj.PathId == PathId;
        }

        public override int GetHashCode()
        {
            return PathId;
        }

        public PathType Prefix(int prefixLength)
        {
            PathType type = this;
            while (type.Depth > prefixLength)
            {
                type = type.Parent;
            }

            return type;
        }
    }
}