﻿namespace NETing.Index.Documents.DocumentModel.Scope
{
    using System.Text;

    public sealed class Scope
    {
        public PathType PathType { get; set; }

        public Dewey Dewey { get; set; }

        public Scope(PathType pathType, Dewey dewey)
        {
            PathType = pathType;
            Dewey = dewey;
        }

        public override bool Equals(object obj)
        {
            return Equals(obj as Scope);
        }

        public bool Equals(Scope obj)
        {
            if (obj == null)
            {
                return false;
            }

            return PathType.Equals(obj.PathType) && Dewey.Equals(obj.Dewey);
        }

        public override int GetHashCode()
        {
            return PathType.PathId + 7 * Dewey.GetHashCode();
        }

        public override string ToString()
        {
            var sb = new StringBuilder(20);
            sb.Append("{");
            var pathEnum = PathType.GetEnumerator();
            var deweyEnum = Dewey.GetEnumerator();

            bool hasMoreDewey = deweyEnum.MoveNext();
            while (pathEnum.MoveNext() & hasMoreDewey) 
            {                
                sb.Append(pathEnum.Current);
                sb.Append("[").Append(deweyEnum.Current).Append("]");
                sb.Append("/");

                hasMoreDewey = deweyEnum.MoveNext();
            }

            if (hasMoreDewey)
            {
                sb.Append("[").Append(deweyEnum.Current).Append("]");
            }

            return sb.Append("}").ToString();
        }

        public Scope Duplicate()
        {
            return new Scope(PathType, Dewey.Duplicate());
        }

        public Scope Prefix(int prefixLength)
        {
            return new Scope(PathType.Prefix(prefixLength), Dewey.Prefix(prefixLength));
        }

        public int Lcp(Scope other)
        {
            int typeLcp = PathTypeLcp.Lcp(PathType, other.PathType).Depth;
            for (int i = 0; i < typeLcp; i++)
            {
                if (Dewey[i] != other.Dewey[i])
                {
                    return i;
                }
            }

            return typeLcp;
        }
    }
}


