﻿namespace NETing.Index.Documents.FileSource
{
    using System;
    using System.Collections;
    using System.Collections.Generic;
    using System.IO;

    public class DirectoryCrawler : IEnumerable<FileInfo>
    {
        public DirectoryInfo RootDirectory { get; private set; }

        public bool Recurse { get; set; }

        public string SearchPattern { get; set; }

        public DirectoryCrawler(DirectoryInfo rootDirectory, bool recurse)
        {
            if (!rootDirectory.Exists)
            {
                throw new ArgumentException("The supplied directory should exist", "rootDirectory");
            }

            RootDirectory = rootDirectory;
            Recurse = recurse;
            SearchPattern = SearchPattern ?? "*.xml";
        }

        IEnumerator IEnumerable.GetEnumerator() 
        {
            return GetEnumerator();
        }

        public IEnumerator<FileInfo> GetEnumerator()
        {
            var queue = new Queue<DirectoryInfo>();
            queue.Enqueue(RootDirectory);
            while (queue.Count > 0)
            {
                DirectoryInfo dir = queue.Dequeue();
                foreach (FileInfo file in dir.GetFiles(SearchPattern))
                {
                    yield return file;
                }

                foreach (DirectoryInfo childDir in dir.GetDirectories()) 
                {
                    queue.Enqueue(childDir);
                }
            }
        }
    }
}


