﻿namespace NETing.Index.Documents.FileSource
{
    using System.Collections;
    using System.Collections.Generic;
    using System.IO;

    public class RoundRobinDirectoryCrawler : IEnumerable<FileInfo>
    {
        private IEnumerable<FileInfo>[] crawlers;
        private int[] chunkSizes;

        public RoundRobinDirectoryCrawler(int[] chunkSizes, IEnumerable<FileInfo>[] crawlers)
        {
            this.crawlers = crawlers;
            this.chunkSizes = chunkSizes;
        }

        public IEnumerator<FileInfo> GetEnumerator()
        {
            int activePrefix = crawlers.Length;
            var enumerators = new IEnumerator<FileInfo>[crawlers.Length];
            var chunks = new int[crawlers.Length];
            for (int i = 0; i < crawlers.Length; i++)
            {
                enumerators[i] = crawlers[i].GetEnumerator();
                chunks[i] = chunkSizes[i];
            }           

            int position = 0;
            while (activePrefix > 0)
            {
                position %= activePrefix;

                for (int i = 0; i < chunks[position]; i++)
                {
                    if (!enumerators[position].MoveNext())
                    {   // replace with last enumerator within active prefix
                        activePrefix--;
                        enumerators[position].Dispose();
                        enumerators[position] = enumerators[activePrefix];
                        chunks[position] = chunks[activePrefix];
                        break;
                    }

                    yield return enumerators[position].Current;
                }

                position++;
            }
        }

        IEnumerator IEnumerable.GetEnumerator()
        {
            return GetEnumerator();
        }
    }
}


