﻿namespace NETing.Index.Documents.Parser
{
    using System;
    using Utils.Compression;

    public sealed class Term : IComparable<Term>
    {
        internal CharArrayHolder ArrayHolder { get; set; }

        internal int Start;

        public int Length { get; set; }

        internal bool IsInterned;

        internal int InvertedPos;

        internal int hashCodeCache;

        internal bool hashCodeCached;

        internal Term(CharArrayHolder array)
        {
            ArrayHolder = array;
        }

        public Term(string term)
        {
            ArrayHolder = new CharArrayHolder
                              {
                                  Array = term.ToLowerInvariant().ToCharArray()
                              };
            Start = 0;
            Length = term.Length;
        }

        public Term(char[] chars)
        {
            for (int i = 0; i < chars.Length; i++)
            {
                chars[i] = Char.ToLowerInvariant(chars[i]);
            }

            ArrayHolder = new CharArrayHolder { Array = chars };
            Start = 0;
            Length = chars.Length;
        }

        public override int GetHashCode()
        {
            if (IsInterned & hashCodeCached)
            {
                return hashCodeCache;
            }

            int hashCode = 0;
            int limit = Start + Length;
            char[] array = ArrayHolder.Array;
            for (int i = Start; i < limit; i++)
            {
                hashCode = hashCode * 31 + array[i];
            }

            hashCodeCache = hashCode;
            hashCodeCached = true;

            return hashCode;
        }

        public override bool Equals(object obj)
        {
            if (obj == null)
            {
                return false;
            }

            if (ReferenceEquals(this, obj))
            {
                return true;
            }

            if (obj.GetType() == GetType())
            {
                return Equals((Term)obj);    
            }

            return false;
        }

        private bool Equals(Term other)
        {
            int length = Length;
            if (other.Length != length)
            {
                return false;
            }

            char[] otherArray = other.ArrayHolder.Array;
            char[] thisArray = ArrayHolder.Array;
            int otherStart = other.Start;
            int thisStart = Start;
            
            for (int i = 0; i < length; i++, otherStart++, thisStart++)
            {
                if (otherArray[otherStart] != thisArray[thisStart])
                {
                    return false;
                }
            }

            return true;
        }

        public int CompareTo(Term other)
        {
            int prefix = Utils.Min(Length, other.Length);
            int cmp = 0;
            int start = Start;
            int otherStart = other.Start;
            char[] array = ArrayHolder.Array;
            char[] otherArray = other.ArrayHolder.Array;
            for (int i = 0; i < prefix && cmp == 0; i++)
            {
                cmp = array[start + i] - otherArray[otherStart + i];
            }

            if (cmp == 0)
            {
                return Length - other.Length;
            }

            return cmp;
        }

        public override string ToString()
        {
            return new string(ArrayHolder.Array, Start, Length);
        }
    }
}