﻿namespace NETing.Index.Documents.Parser
{
    using System;
    using System.Collections.Generic;

    internal sealed class TermStorage
    {
        private readonly CharArrayHolder characterStore 
            = new CharArrayHolder { Array = new char[1024 * 1024 * 10] };
            
        private int storeHighWaterMark;
            
        private readonly Dictionary<Term, Term> singletonTerms 
            = new Dictionary<Term, Term>(1024 * 50);
        
        public Term Intern(Term term)
        {
            Term interned;
            if (!singletonTerms.TryGetValue(term, out interned))
            {
                interned = CreateInternedTerm(term);
                singletonTerms.Add(interned, interned);
            }

            return interned;
        }

        private Term CreateInternedTerm(Term term)
        {
            if (characterStore.Array.Length - storeHighWaterMark < term.Length)
            {
                var newStore = new char[(storeHighWaterMark + term.Length) << 1];
                Buffer.BlockCopy(characterStore.Array, 0, newStore, 0, storeHighWaterMark * 2);
                characterStore.Array = newStore;
            }

            Buffer.BlockCopy(
                term.ArrayHolder.Array, 
                term.Start * 2, 
                characterStore.Array, 
                storeHighWaterMark * 2, 
                term.Length * 2);
            var interned = new Term(characterStore)
                               {
                                   Length = term.Length,
                                   Start = storeHighWaterMark,
                                   IsInterned = true,
                                   hashCodeCache = term.hashCodeCache,
                                   hashCodeCached = true
                               };
            storeHighWaterMark += term.Length;

            return interned;
        }
    }
}