﻿namespace NETing.Index.Documents.Parser
{
    using System.Collections;
    using System.Collections.Generic;

    internal sealed class Tokenizer : IEnumerator<Term>
    {
        private static readonly char[] charTransferTable = new char[char.MaxValue - char.MinValue];
        private static readonly bool[] charIsLetterOrDigit = new bool[char.MaxValue - char.MinValue];

        private readonly Term holder;
        private readonly CharArrayHolder arrayHolder 
            = new CharArrayHolder { Array = new char[1024 * 64] };

        private char[] source;
        private char[] array;
        private int tokenStart;

        private int limit;

        static Tokenizer()
        {
            for (char i = char.MinValue; i < char.MaxValue; i++)
            {
                charTransferTable[i - char.MinValue] = char.ToLowerInvariant(i);
                charIsLetterOrDigit[i - char.MinValue] = char.IsLetterOrDigit(i);
            }
        }

        public Tokenizer()
        {
            holder = new Term(arrayHolder) { Start = 0 };
            array = arrayHolder.Array;
        }

        public void Initialize(char[] source, int length)
        {
            this.source = source;
            limit = length;
            tokenStart = 0;
            while (tokenStart < limit & !charIsLetterOrDigit[source[tokenStart] - char.MinValue])
            {
                tokenStart++;
            }
        }

        public void Dispose()
        {
        }

        public bool MoveNext()
        {
            int pos = tokenStart;
            int i = 0;
            while (pos < limit & charIsLetterOrDigit[source[pos] - char.MinValue])
            {   // case folding while tokenizing
                array[i++] = charTransferTable[source[pos++] - char.MinValue];
            }

            holder.Length = i;

            while (pos < limit & !charIsLetterOrDigit[source[pos] - char.MinValue])
            {
                pos++;
            }

            tokenStart = pos;
            return holder.Length > 0;
        }

        public void Reset()
        {
            Initialize(source, limit);
        }

        public Term Current
        {
            get { return holder; }
        }

        object IEnumerator.Current
        {
            get { return Current; }
        }
    }
}