﻿namespace NETing.Index.Inversion
{
    using System.Linq;
 
    using Documents.DocumentModel.Scope;

    internal sealed class InvertedDocumentOccurrence
    {
        internal uint DocumentId;
        internal int Frequency;
        internal ushort[] Positions = new ushort[128];
        internal ushort[] PathTypes = new ushort[128];
        internal int TotalDeweys;
        internal ushort[] Deweys = new ushort[128 << 2];

        public InvertedDocumentOccurrence Duplicate()
        {
            return new InvertedDocumentOccurrence
                       {
                           Frequency = Frequency,
                           TotalDeweys = TotalDeweys,
                           Positions = Positions.ToArray(),
                           PathTypes = PathTypes.ToArray(),
                           Deweys = PathTypes.ToArray(),
                       };
        }

        public Scope ScopeAt(int i, ScopeDirectory scopeDirectory)
        {
            int offset = 0;
            for (int j = 0; j < i; j++)
            {
                offset += PathTypeAt(j, scopeDirectory).DeweyDepth;
            }

            PathType type = PathTypeAt(i, scopeDirectory);
            return new Scope(type, new Dewey(Deweys, offset, type.DeweyDepth));
        }

        public PathType PathTypeAt(int i, ScopeDirectory scopeDirectory)
        {
            return scopeDirectory[PathTypes[i]];
        }
    }
}
