﻿namespace NETing.Index.Inversion
{
    using Utils.ByteArrays;

    internal unsafe static class InvertedDocumentOccurrencesUtils
    {
        public static int SerializeTo(
            this InvertedDocumentOccurrence occurrences,
            uint documentId,
            ref byte[] buffer)
        {
            int exactSize = occurrences.ExactSize();
            EnsureCapacity(ref buffer, exactSize + 4);
            fixed (byte* buff = &buffer[0])
            {
                ByteArrayUtils.SetValueAt(buff, exactSize);
                ByteArrayUtils.SetValueAt(&buff[4], documentId);
                ByteArrayUtils.SetValueAt(&buff[8], (ushort)occurrences.Frequency);
                ByteArrayUtils.SetValueAt(&buff[10], (ushort)occurrences.TotalDeweys);
            }

            int bytes = occurrences.Frequency << 1;
            BlockCopy.Copy(occurrences.Positions, 0, buffer, 12, bytes);
            BlockCopy.Copy(occurrences.PathTypes, 0, buffer, 12 + bytes, bytes);
            BlockCopy.Copy(occurrences.Deweys, 0, buffer, 12 + bytes + bytes, occurrences.TotalDeweys << 1);

            return 4 + exactSize;
        }

        public static void DeserializeFrom(
            this InvertedDocumentOccurrence occurrences,
            byte[] buffer,
            int offset)
        {
            fixed (byte* buff = &buffer[offset])
            {
                occurrences.DocumentId = ByteArrayUtils.GetUInt32At(buff);
                occurrences.Frequency = ByteArrayUtils.GetUInt16At(&buff[4]);
                occurrences.TotalDeweys = ByteArrayUtils.GetUInt16At(&buff[6]);
            }

            if (occurrences.Frequency > occurrences.Positions.Length)
            {
                occurrences.Positions = new ushort[occurrences.Frequency << 1];
                occurrences.PathTypes = new ushort[occurrences.Frequency << 1];
            }

            if (occurrences.TotalDeweys > occurrences.Deweys.Length)
            {
                occurrences.Deweys = new ushort[occurrences.TotalDeweys << 1];
            }

            int bytes = occurrences.Frequency << 1;
            BlockCopy.Copy(buffer, offset + 8, occurrences.Positions, 0, bytes);
            BlockCopy.Copy(buffer, offset + 8 + bytes, occurrences.PathTypes, 0, bytes);
            BlockCopy.Copy(buffer, offset + 8 + (bytes << 1), occurrences.Deweys, 0, occurrences.TotalDeweys << 1);
        }

        private static void EnsureCapacity(ref byte[] buffer, int requiredSize)
        {
            if (requiredSize > buffer.Length)
            {
                buffer = new byte[requiredSize << 1];
            }
        }

        private static int ExactSize(this InvertedDocumentOccurrence occurrences)
        {   // 2 bytes per dewey, 4 bytes per occurrence, 8 bytes docId, freq, dewey cnt

            return (occurrences.TotalDeweys << 1) + (occurrences.Frequency << 2) + 8;
        }
    }
}
