﻿namespace NETing.Index.Inversion
{
    using System;
    using System.Diagnostics;
    using Utils.ByteArrays;

    internal sealed class InvertedListRef
    {
        public InversionStore Store { get; set; }

        public InvertedListRef Overflow { get; set; }

        public int FirstPosition { get; set; }

        public int SegmentStart { get; set; }

        public int TotalSize { get; set; }

        public int Append { get; set; }

        public int Limit { get; set; }

        public int Remaining
        {
            get { return Limit - Append; }
        }

        public InvertedListRef(InversionStore store, int size)
        {
            Store = store;
            SegmentStart = store.AllocateSegment(size);
            FirstPosition = SegmentStart;
            Append = SegmentStart + 4;
            Limit = SegmentStart + size - 4;
            TotalSize = size;
            WriteLimitPointer();
        }

        public void Extend(int minimumSize)
        {
            int size = Math.Min(
                Math.Max(TotalSize, minimumSize) + ScalingInMemoryInverter.SegmentOverhead, 
                Store.Remaining);
            SegmentStart = Store.AllocateSegment(size);
            WriteNextSegmentPointer(SegmentStart);
            Append = SegmentStart + 4;       // leave room for pointer at start
            Limit = SegmentStart + size - 4; // leave room for pointer at end
            WriteLimitPointer();
            TotalSize += size;
        }

        private void WriteNextSegmentPointer(int segmentStart)
        {
            Store.Buffer.SetValueAt(Limit, segmentStart);
        }

        private void WriteLimitPointer()
        {
            Store.Buffer.SetValueAt(SegmentStart, Limit);
        }

        public void Write(byte[] buffer, int offset, int length)
        {
            Debug.Assert(
                Remaining >= length, 
                "Do not issue write order larger than remaining.");
            BlockCopy.Copy(buffer, offset, Store.Buffer, Append, length);
            Append += length;
        }
    }
}
