﻿namespace NETing.Index.PathMatching.Syntax
{
    using System.Collections.Generic;
    using System.Data;
    using Documents.DocumentModel.Scope;

    internal class SyntaxChecker
    {
        private readonly ScopeDirectory scopeDirectory;

        public SyntaxChecker(ScopeDirectory scopeDirectory)
        {
            this.scopeDirectory = scopeDirectory;
        }

        public IEnumerable<ExpressionToken> Validate(IEnumerable<ExpressionToken> tokens)
        {
            ExpressionToken previous = null;
            foreach (var token in tokens)
            {
                if (previous != null && previous.Type.IsConnector() && token.Type.IsConnector())
                {
                    throw new SyntaxErrorException("Two connectors can not follow each other.");
                }

                if (token.Type == TokenType.ElementTag && token.Value != "*" && scopeDirectory.GetTag(token.Value) == null)
                {
                    throw new SyntaxErrorException("Unknown tag: " + token.Value);
                }

                yield return token;
                previous = token;
            }
        }
    }
}


