﻿namespace NETing.Index.PostingFile
{
    using System;
    using Storage.Buffer;
    using Utils.ByteArrays;
    using Utils.Compression;

    public class PostingFileDataReader : IDisposable
    {
        private readonly IBufferManager bufferManager;

        private BlockBuffer currentBuffer;

        internal PostingFileDataReader(
            IBufferManager bufferManager)
        {
            this.bufferManager = bufferManager;
        }

        public void LoadData(
            byte[] data, 
            int offset,
            int count,
            long fileOffset)
        {
            int read = 0;
            if (currentBuffer == null)
            {
                currentBuffer = bufferManager.Pin(fileOffset, true);
            }
            else if (!currentBuffer.Contains(fileOffset))
            {
                currentBuffer = bufferManager.ExchangeBuffers(currentBuffer, fileOffset, true);
            }

            int bufferOffset = currentBuffer.BufferOffset(fileOffset);
            while (count - read > 0)
            {
                int bytes = Utils.Min(count - read, currentBuffer.Remaining(bufferOffset));
                BlockCopy.Copy(currentBuffer.Buffer, bufferOffset, data, offset + read, bytes);
                read += bytes;
                bufferOffset += bytes;

                if (currentBuffer.Remaining(bufferOffset) <= 0)
                {
                    // Swap buffers
                    fileOffset += bufferOffset;
                    bufferOffset = 0;
                    currentBuffer = bufferManager.ExchangeBuffers(currentBuffer, fileOffset, true);
                }
            }
        }

        public void Dispose()
        {
            if (currentBuffer != null)
            {
                bufferManager.UnPin(currentBuffer);
                currentBuffer = null;
            }
        }
    }
}
