﻿namespace NETing.Index.PostingFile
{
    using Storage.Buffer;
    using Utils.Compression;

    internal static class PostingFileUtils
    {
        #region Chunk Headers

        public static int WriteChunkHeader(
            this PostingFileWriter writer,
            BlockBuffer buffer, 
            int offset,
            int headerBytes,
            uint totalTokens,
            int documentIdLength,
            int frequencyLength,
            int positionLength,
            int scopeLength)
        {
            buffer.IsDirty = true;
            offset = PrefixedVByte.Encode(headerBytes, buffer.Buffer, offset);
            offset = PrefixedVByte.Encode(totalTokens, buffer.Buffer, offset);
            offset = PrefixedVByte.Encode((uint)documentIdLength, buffer.Buffer, offset);
            offset = PrefixedVByte.Encode((uint)frequencyLength, buffer.Buffer, offset);
            offset = PrefixedVByte.Encode((uint)positionLength, buffer.Buffer, offset);
            return PrefixedVByte.Encode((uint)scopeLength, buffer.Buffer, offset);
        }

        public static int ReadChunkHeader(
            this PostingFileEnumerator enumerator,
            BlockBuffer buffer, 
            int offset,
            out uint headerBytes,
            out uint totalTokens,
            out uint documentIdLength,
            out uint frequencyLength,
            out uint positionLength,
            out uint scopeLength)
        {
            offset = PrefixedVByte.Decode(buffer.Buffer, offset, out headerBytes);
            offset = PrefixedVByte.Decode(buffer.Buffer, offset, out totalTokens);
            offset = PrefixedVByte.Decode(buffer.Buffer, offset, out documentIdLength);
            offset = PrefixedVByte.Decode(buffer.Buffer, offset, out frequencyLength);
            offset = PrefixedVByte.Decode(buffer.Buffer, offset, out positionLength);
            return PrefixedVByte.Decode(buffer.Buffer, offset, out scopeLength);
        }

        #endregion
    }
}
