﻿namespace NETing.Index.QueryLanguage
{
    using System.IO;
    using Documents.Parser;
    using PathMatching;
    using System.Linq;

    public class QueryParser
    {
        private PathMatcher PathMatcher { get; set; }

        internal QueryParser(PathMatcher pathMatcher)
        {
            PathMatcher = pathMatcher;
        }

        public static void PrintSyntax(TextWriter output)
        {
            output.WriteLine("Simple lookup: <term>");
            output.WriteLine("Path filtered lookup: <term> # <pathExpression>");
            output.WriteLine("Intersection query (AND): (<term> ...)");
            output.WriteLine("Scope intersection query (AND): [<term> ...] # <pathExpression>");
            output.WriteLine("-----------------------------------------------");
            output.WriteLine("Path expressions:");
            PathMatcher.PrintSyntax(output);
        }

        internal Query Parse(string query)
        {
            query = query.Trim();
            if (query.Length > 2 && query[0] == '[')
            {
                if (query.LastIndexOf(']') < query.LastIndexOf('#'))
                {
                    string[] terms = query.Substring(1, query.LastIndexOf(']') - 1).Split(' ');
                    string pathExpression = query.Substring(query.LastIndexOf('#') + 1).Trim();

                    return new Query(
                        terms.Select(t => new Term(t.Trim())).ToArray(),
                        PathMatcher.MatchingPaths(pathExpression));
                }

                if (query.LastIndexOf(']') < query.LastIndexOf('*'))
                {
                    string[] terms = query.Substring(1, query.LastIndexOf(']') - 1).Split(' ');
                    string pathExpression = query.Substring(query.LastIndexOf('#') + 1).Trim();

                    return new Query(
                        terms.Select(t => new Term(t.Trim())).ToArray(),
                        PathMatcher.MatchingPaths(pathExpression),
                        true);
                }
            }

            if (query.Length > 2 && query[0] == '(' && query.Last() == ')')
            {
                string[] terms = query.Substring(1, query.Length - 2).Split(' ');
                var q = new Query(terms.Select(t => new Term(t)).ToArray());
                return q;
            }

            string[] queryParts = query.Split('#');
            if (queryParts.Length < 2)
            {
                return new Query(new[] { new Term(query) });    
            }
            
            return new Query(
                new[] { new Term(queryParts[0].Trim()) },
                PathMatcher.MatchingPaths(queryParts[1].Trim()));
        }
    }
}


