﻿namespace NETing.Index.Search
{
    using System;
    using System.Collections;
    using System.Collections.Generic;
    using System.Diagnostics;

    using PostingFile;

    using Utils.Collections;

    [DebuggerDisplay("Current = {Current}")]
    public sealed class IntersectionQueryItem : IEnumerator<QueryResult[]>
    {
        private bool exhausted;
        private readonly QueryResult[] holder;
        private readonly ListHolder[] listHolders;
        private readonly BinaryMaxHeap<ListHolder> heap;

        internal IntersectionQueryItem(PostingFileEnumerator[] lists)
        {
            listHolders = new ListHolder[lists.Length];
            for (int i = 0; i < lists.Length; i++)
            {
                if (!lists[i].MoveNext())
                {
                    exhausted = true;
                    return;
                }

                listHolders[i] = new ListHolder
                                     {
                                         Enumerator = lists[i],
                                         Index = i
                                     };
            }

            holder = new QueryResult[lists.Length];
            for (int i = 0; i < holder.Length; i++)
            {
                holder[i] = new QueryResult();    
            }

            heap = new BinaryMaxHeap<ListHolder>(listHolders);
        }
        
        public void Dispose()
        {
            foreach (var listHolder in listHolders)
            {
                listHolder.Enumerator.Dispose();
            }
        }

        #region Implementation of IEnumerator

        public bool MoveNext()
        {
            while (!exhausted)
            {
                uint docId = heap.Array[0].Enumerator.Current.DocumentId;
                int position;
                for (position = 1; position < listHolders.Length; position++)
                {
                    if (heap.Array[position].Enumerator.Current.DocumentId < docId)
                    {
                        if (!heap.Array[position].Enumerator.SkipToDocumentId(docId))
                        {
                            exhausted = true;
                            return false;
                        }

                        // heap.IncreaseKey(position);
                        if (heap.Array[position].Enumerator.Current.DocumentId > docId)
                        {
                            heap.Heapify();
                            break;
                        }
                    }
                }

                if (position == listHolders.Length)
                {
                    PrepareResult();
                    heap.Heapify();
                    return true;
                }
            }

            return false;
        }

        private void PrepareResult()
        {
            foreach (ListHolder listHolder in listHolders)
            {
                QueryResult result = holder[listHolder.Index];
                PostingFileEnumerator enumerator = listHolder.Enumerator;
                enumerator.SynchronizeWithDocumentColumn();
                exhausted |= !enumerator.LoadResults(result, 0);
            }
        }

        public void Reset()
        {
            throw new NotImplementedException();
        }

        public QueryResult[] Current
        {
            get { return holder; }
        }

        object IEnumerator.Current
        {
            get { return Current; }
        }

        #endregion
    }
}
