﻿namespace NETing.Index.Storage
{
    using System.IO;

    public class BlockAccessFile : IBlockReader, IBlockWriter
    {
        public int BlockSize { get; private set; }

        private readonly Stream stream;

        public BlockAccessFile(Stream stream, int blockSize)
        {
            BlockSize = blockSize;
            this.stream = stream;           
        }

        public void WriteBlock(long blockPosition, byte[] buffer, int bufferOffset)
        {
            stream.Seek(blockPosition * BlockSize, SeekOrigin.Begin);
            stream.Write(buffer, bufferOffset, BlockSize);  
        }

        public void ReadBlock(long blockPosition, byte[] buffer, int bufferOffset)
        {
            stream.Seek(blockPosition * BlockSize, SeekOrigin.Begin);
            int read = 0;
            while (read < BlockSize)
            {
                read += stream.Read(buffer, bufferOffset + read, BlockSize - read);    
            }
        }

        public void ReadBlocks(long blockPosition, int blocks, byte[] buffer, int bufferOffset)
        {
            stream.Seek(blockPosition * BlockSize, SeekOrigin.Begin);
            int read = 0;
            int bytes = BlockSize * blocks;
            while (read < bytes)
            {
                read += stream.Read(buffer, bufferOffset + read, bytes - read);
            }
        }

        public void Dispose()
        {
            stream.Dispose();
        }
    }
}


