﻿namespace NETing.Index.Storage.Buffer
{
    using System.Diagnostics;

    [DebuggerDisplay("Key = {Key}")]
    public sealed class BlockBuffer
    {
        public int BlockSize { get; private set; }

        internal BufferKey Key { get; set; }
        
        public long BlockNumber
        {
            get { return Key.BlockNumber; }
        }
        
        public long FileOffset
        {
            get { return BlockSize * BlockNumber;  }
        }
        
        public long LastPosition
        {
            get { return FileOffset + BlockSize; }
        }

        internal BlockBuffer(int blockSize)
        {
            BlockSize = blockSize;
            Buffer = new byte[blockSize];
        }

        public int BufferOffset(long fileOffset)
        {
            return (int) (fileOffset - FileOffset);
        }

        public bool Contains(long fileOffset)
        {
            return fileOffset >= FileOffset && fileOffset < FileOffset + BlockSize;
        }

        public int Remaining(long position)
        {
            return BlockSize - BufferOffset(position);
        }

        public int Remaining(int offset)
        {
            return BlockSize - offset;
        }

        public bool IsDirty { get; set; }

        internal int PinCount { get; set; }

        public byte[] Buffer { get; private set; }
    }
}


