﻿namespace NETing.Index.Utils.ByteArrays
{
    public static class BlockCopy
    {
        public static unsafe void Copy(uint[] src, int srcOffset, byte[] dst, int dstOffset, int count)
        {
            fixed (uint* srcPtr = &src[0])
            {
                fixed (byte* dstPtr = &dst[dstOffset])
                {
                    Copy(((byte*) srcPtr) + srcOffset, dstPtr, count);
                }
            }
        }

        public static unsafe void Copy(uint[] src, int srcOffset, uint[] dst, int dstOffset, int count)
        {
            fixed (uint* srcPtr = &src[srcOffset])
            {
                fixed (uint* dstPtr = &dst[dstOffset])
                {
                    Copy(((byte*)srcPtr) + srcOffset, (byte*)dstPtr, count);
                }
            }
        }

        public static unsafe void Copy(uint[] src, int srcOffset, int[] dst, int dstOffset, int count)
        {
            fixed (uint* srcPtr = &src[srcOffset])
            {
                fixed (int* dstPtr = &dst[dstOffset])
                {
                    Copy(((byte*)srcPtr) + srcOffset, (byte*)dstPtr, count);
                }
            }
        }

        public static unsafe void Copy(ushort[] src, int srcOffset, ushort[] dst, int dstOffset, int count)
        {
            fixed (ushort* srcPtr = &src[0])
            {
                fixed (ushort* dstPtr = &dst[0])
                {
                    Copy(((byte*)srcPtr) + srcOffset, ((byte*)dstPtr) + dstOffset, count);
                }
            }
        }

        public static unsafe void Copy(ushort[] src, int srcOffset, byte[] dst, int dstOffset, int count)
        {
            fixed (ushort* srcPtr = &src[0])
            {
                fixed (byte* dstPtr = &dst[dstOffset])
                {
                    Copy(((byte*)srcPtr) + srcOffset, dstPtr, count);
                }
            }
        }

        public static unsafe void Copy(byte[] src, int srcOffset, ushort[] dst, int dstOffset, int count)
        {
            fixed (byte* srcPtr = &src[srcOffset])
            {
                fixed (ushort* dstPtr = &dst[0])
                {
                    Copy(srcPtr, ((byte*) dstPtr) + dstOffset, count);
                }
            }
        }

        public static unsafe void Copy(byte[] src, int srcOffset, byte[] dst, int dstOffset, int count)
        {
            fixed (byte* srcPtr = &src[srcOffset])
            {
                fixed (byte* dstPtr = &dst[dstOffset])
                {
                    Copy(srcPtr, dstPtr, count);
                }
            }
        }

        private static unsafe void Copy(byte* src, byte* dst, int count)
        {
            while (count > 31)
            {
                ((ulong*)dst)[0] = ((ulong*)src)[0];
                ((ulong*)dst)[1] = ((ulong*)src)[1];
                ((ulong*)dst)[2] = ((ulong*)src)[2];
                ((ulong*)dst)[3] = ((ulong*)src)[3];
                dst += 32;
                src += 32;
                count -= 32;
            }

            while (count > 15)
            {
                ((ulong*)dst)[0] = ((ulong*)src)[0];
                ((ulong*)dst)[1] = ((ulong*)src)[1];
                dst += 16;
                src += 16;
                count -= 16;
            }

            while (count > 7)
            {
                ((ulong*) dst)[0] = ((ulong*) src)[0];
                dst += 8;
                src += 8;
                count -= 8;
            }

            if (count > 3) 
            {
                ((uint*)dst)[0] = ((uint*)src)[0];
                dst += 4;
                src += 4;
                count -= 4;
            }

            if (count > 1)
            {
                ((ushort*)dst)[0] = ((ushort*)src)[0];
                dst += 2;
                src += 2;
                count -= 2;
            }

            if (count > 0)
            {
                dst[0] = src[0];
            }
        }
    }
}
