﻿namespace NETing.Index.Utils.ByteArrays
{
    internal static unsafe class ByteArrayUtils
    {
        public static void SetValueAt(this byte[] array, int position, int value)
        {
            if ((position & 0x3) == 0)
            {
                fixed (byte* ptr = &array[position])
                {
                    *((int*) ptr) = value;
                }
            }
            else
            {
                uint uintValue = (uint) value;
                array[position] = (byte)uintValue;
                array[position + 1] = (byte)((uintValue >> 8) & 0xff);
                array[position + 2] = (byte)((uintValue >> 16) & 0xff);
                array[position + 3] = (byte)((uintValue >> 24) & 0xff);
            }
        }

        public static void SetValueAt(byte* array, int value)
        {
            *((int*)array) = value;
        }

        public static void SetValueAt(this byte[] array, int position, ushort value)
        {
            if ((position & 0x1) == 0)
            {
                fixed (byte* ptr = &array[position])
                {
                    *((ushort*)ptr) = value;
                }
            }
            else
            {
                array[position] = (byte) value;
                array[position + 1] = (byte) ((value >> 8) & 0xff);
            }

        }

        public static void SetValueAt(byte* array, ushort value)
        {
            *((ushort*)array) = value;
        }

        public static void SetValueAt(this byte[] array, int position, uint value)
        {
            if ((position & 0x3) == 0)
            {
                fixed (byte* ptr = &array[position])
                {
                    *((uint*)ptr) = value;
                }
            }
            else
            {
                array[position] = (byte)value;
                array[position + 1] = (byte)((value >> 8) & 0xff);
                array[position + 2] = (byte)((value >> 16) & 0xff);
                array[position + 3] = (byte)((value >> 24) & 0xff);
            }
        }

        public static void SetValueAt(byte* array, uint value)
        {
            *((uint*)array) = value;
        }

        public static int GetInt32At(this byte[] array, int position)
        {
            if ((position & 0x3) == 0)
            {
                fixed (byte* ptr = &array[position])
                {
                    return *((int*) ptr);
                }
            }

            int res = array[position];
            res |= (array[position + 1] << 8);
            res |= (array[position + 2] << 16);
            res |= (array[position + 3] << 24);
            return res;
        }

        public static int GetInt32At(byte* array)
        {
            return *((int*) array);
        }

        public static ushort GetUInt16At(this byte[] array, int position)
        {
            if ((position & 0x1) == 0)
            {
                fixed (byte* ptr = &array[position])
                {
                    return *((ushort*)ptr);
                }
            }

            return (ushort) (array[position] | (array[position + 1] << 8));
        }

        public static ushort GetUInt16At(byte* array)
        {
            return *((ushort*)array);
        }

        public static uint GetUInt32At(this byte[] array, int position)
        {
            if ((position & 0x3) == 0)
            {
                fixed (byte* ptr = &array[position])
                {
                    return *((uint*) ptr);
                }
            }

            uint res = array[position];
            res |= (uint)(array[position + 1] << 8);
            res |= (uint)(array[position + 2] << 16);
            res |= (uint)(array[position + 3] << 24);
            return res;
        }

        public static uint GetUInt32At(byte* array)
        {
            return *((uint*)array);
        }
    }
}


