﻿namespace NETing.Index.Utils.Collections
{
    using System;

    public sealed class BinaryMaxHeap<T> where T : IComparable<T>
    {
        private int heapSize;
        private T[] array;

        public BinaryMaxHeap()
        {
            array = new T[10];
            heapSize = 0;
        }

        public BinaryMaxHeap(T[] array)
        {
            this.array = array;
            heapSize = array.Length;
            BuildMaxHeap();
        }

        #region Heap interface

        public T Max
        {
            get { return array[0]; }
        }

        public T[] Array
        {
            get { return array; }
        }

        public int Count
        {
            get { return heapSize; }
        }

        public T ExtractMax()
        {
            T max = array[0];
            array[0] = array[heapSize - 1];
            heapSize--;
            MaxHeapify(0);
            return max;
        }

        public void Add(T item)
        {
            if (heapSize >= array.Length)
            {
                System.Array.Resize(ref array, heapSize << 1);
            }

            array[heapSize] = item;
            IncreaseKey(heapSize++);
        }

        public void IncreaseKey(int i)
        {
            int parent = Parent(i);
            while (i > 0 && array[parent].CompareTo(array[i]) < 0)
            {
                Exchange(i, parent);

                i = parent;
                parent = Parent(i);
            }
        }

        public void Heapify()
        {
            BuildMaxHeap();
        }

        #endregion

        #region Binary tree relationships

        private static int Parent(int i)
        {
            return ((i + 1) >> 1) - 1;
        }

        private static int Left(int i)
        {
            return ((i + 1) << 1) - 1;
        }

        private static int Right(int i)
        {
            return ((i + 1) << 1);
        }

        #endregion

        #region Internal maintanance functionality

        private void Exchange(int i, int j)
        {
            var tmp = array[i];
            array[i] = array[j];
            array[j] = tmp;
        }

        private void BuildMaxHeap()
        {
            for (int i = (heapSize >> 1) - 1; i >= 0; i--)
            {
                MaxHeapify(i);
            }
        }

        private void MaxHeapify(int i)
        {
            int largest = i;
            do
            {
                i = largest;
                int l = Left(i);
                int r = Right(i);
                if (l < heapSize && array[l].CompareTo(array[i]) > 0)
                {
                    largest = l;
                }

                if (r < heapSize && array[r].CompareTo(array[largest]) > 0)
                {
                    largest = r;
                }

                Exchange(i, largest);
            } while (largest != i);
        }

        #endregion
    }
}
