﻿namespace NETing.Index.Utils.Collections
{
    using System;

    using Compression;

    public sealed class MutableBitArray
    {
        private static readonly long[] masks = new long[64];

        static MutableBitArray()
        {
            for (int i = 0; i < 64; i++)
            {
                masks[i] = (1 << i);
            }
        }
        
        private long[] array = new long[1];

        private int Count { get; set; }

        public void Clear()
        {
            int limit = Utils.Min(array.Length, (Count >> 6) + 1);
            for (int i = 0; i < limit; i++)
            {
                array[limit] = 0;
            }
        }

        public bool this[uint position]
        {
            get
            {
                return this[(int) position];
            }
        }

        public bool this[int position]
        {
            get
            {
                int qword = position >> 6;
                int bit = position & 0x3f;
                return qword < array.Length && (array[qword] & masks[bit]) > 0;
            }

            set
            {
                Count = Utils.Max(Count, position);
                int qword = position >> 6;
                if (qword >= array.Length)
                {
                    Array.Resize(ref array, array.Length << 1);    
                }

                int bit = position & 0x3f;
                long mask = masks[bit];
                if (value)
                {
                    array[qword] |= mask;
                }
                else
                {
                    array[qword] &= ~mask;
                }
            }
        }
    }
}
