﻿namespace NETing.Index.Utils.Compression
{
    /// <summary>
    /// WARNING: UGLY CODE AHEAD
    /// </summary>
    public static class FastPack
    {
        public static unsafe void FastNPack(int width, int n, uint[] input, int inputOffset, uint[] output)
        {
            fixed (uint* inp = &input[0])
            {
                fixed (uint* outp = &output[0])
                {
                    FastNPack(width, n, inp + inputOffset, outp);
                }
            }
        }

        public static unsafe void FastNPack(int width, int n, uint* input, uint* output)
        {
            while (n >= 128)
            {
                Fast128Pack(width, input, output);
                input += 128;
                output += width << 2;
                n -= 128;
            }

            while (n >= 32)
            {
                Fast32Pack(width, input, output);
                input += 32;
                output += width;
                n -= 32;
            }

            if (n > 0)
            {
                PackUtils.Pack(input, width, n, output);
            }
        }

        public static unsafe void FastNUnPack(int width, int n, int inputOffset, int outputOffset, uint[] input, uint[] output)
        {
            fixed (uint* inPtr = &input[inputOffset])
            {
                fixed (uint* outPtr = &output[outputOffset])
                {
                    FastNUnPack(width, n, inPtr, outPtr);
                }
            }
        }

        public static unsafe void FastNUnPack(int width, int n, uint* input, uint* output)
        {
            while (n >= 128)
            {
                Fast128UnPack(width, input, output);
                output += 128;
                input += width << 2;
                n -= 128;
            }

            while (n >= 32)
            {
                Fast32UnPack(width, input, output);
                output += 32;
                input += width;
                n -= 32;
            }

            if (n > 0)
            {
                PackUtils.Unpack(input, width, n, 0, 0, output);
            }
        }

        public static unsafe void Fast128Pack(int width, uint[] input, uint[] output)
        {
            fixed (uint* inp = &input[0])
            {
                fixed (uint* outp = &output[0])
                {
                    Fast128Pack(width, inp, outp);
                }
            }
        }

        public static unsafe void Fast128Pack(int width, uint[] input, int inputOffset, uint[] output)
        {
            fixed (uint* inp = &input[0])
            {
                fixed (uint* outp = &output[0])
                {
                    Fast128Pack(width, inp + inputOffset, outp);
                }
            }
        }

        private static unsafe void Fast128Pack(int width, uint* input, uint* output)
        {
            switch (width)
            {
                case 0: 
                    return;
                case 1: 
                    Fast32Pack1(input, output);
                    Fast32Pack1(input + 32, output + 1);
                    Fast32Pack1(input + 64, output + 2);
                    Fast32Pack1(input + 96, output + 3);
                    return;
                case 2:
                    Fast32Pack2(input, output);
                    Fast32Pack2(input + 32, output + 2);
                    Fast32Pack2(input + 64, output + 4);
                    Fast32Pack2(input + 96, output + 6);
                    return;
                case 3:
                    Fast32Pack3(input, output);
                    Fast32Pack3(input + 32, output + 3);
                    Fast32Pack3(input + 64, output + 6);
                    Fast32Pack3(input + 96, output + 9);
                    return;
                case 4:
                    Fast32Pack4(input, output);
                    Fast32Pack4(input + 32, output + 4);
                    Fast32Pack4(input + 64, output + 8);
                    Fast32Pack4(input + 96, output + 12);
                    return;
                case 5:
                    Fast32Pack5(input, output);
                    Fast32Pack5(input + 32, output + 5);
                    Fast32Pack5(input + 64, output + 10);
                    Fast32Pack5(input + 96, output + 15);
                    return;
                case 6:
                    Fast32Pack6(input, output);
                    Fast32Pack6(input + 32, output + 6);
                    Fast32Pack6(input + 64, output + 12);
                    Fast32Pack6(input + 96, output + 18);
                    return;
                case 7:
                    Fast32Pack7(input, output);
                    Fast32Pack7(input + 32, output + 7);
                    Fast32Pack7(input + 64, output + 14);
                    Fast32Pack7(input + 96, output + 21);
                    return;
                case 8:
                    Fast32Pack8(input, output);
                    Fast32Pack8(input + 32, output + 8);
                    Fast32Pack8(input + 64, output + 16);
                    Fast32Pack8(input + 96, output + 24);
                    return;
                case 9:
                    Fast32Pack9(input, output);
                    Fast32Pack9(input + 32, output + 9);
                    Fast32Pack9(input + 64, output + 18);
                    Fast32Pack9(input + 96, output + 27);
                    return;
                default:
                    PackUtils.Pack(input, width, 128, output);
                    return;
            }
        }

        public static unsafe void Fast128UnPack(int width, uint* input, uint* output)
        {
            switch (width)
            {
                case 0:
                    return;
                case 1:
                    Fast32UnPack1(input, output);
                    Fast32UnPack1(input + 1, output + 32);
                    Fast32UnPack1(input + 2, output + 64);
                    Fast32UnPack1(input + 3, output + 96);
                    return;
                case 2:
                    Fast32UnPack2(input, output);
                    Fast32UnPack2(input + 2, output + 32);
                    Fast32UnPack2(input + 4, output + 64);
                    Fast32UnPack2(input + 6, output + 96);
                    return;
                case 3:
                    Fast32UnPack3(input, output);
                    Fast32UnPack3(input + 3, output + 32);
                    Fast32UnPack3(input + 6, output + 64);
                    Fast32UnPack3(input + 9, output + 96);
                    return;
                case 4:
                    Fast32UnPack4(input, output);
                    Fast32UnPack4(input + 4, output + 32);
                    Fast32UnPack4(input + 8, output + 64);
                    Fast32UnPack4(input + 12, output + 96);
                    return;
                case 5:
                    Fast32UnPack5(input, output);
                    Fast32UnPack5(input + 5, output + 32);
                    Fast32UnPack5(input + 10, output + 64);
                    Fast32UnPack5(input + 15, output + 96);
                    return;
                case 6:
                    Fast32UnPack6(input, output);
                    Fast32UnPack6(input + 6, output + 32);
                    Fast32UnPack6(input + 12, output + 64);
                    Fast32UnPack6(input + 18, output + 96);
                    return;
                case 7:
                    Fast32UnPack7(input, output);
                    Fast32UnPack7(input + 7, output + 32);
                    Fast32UnPack7(input + 14, output + 64);
                    Fast32UnPack7(input + 21, output + 96);
                    return;
                case 8:
                    Fast32UnPack8(input, output);
                    Fast32UnPack8(input + 8, output + 32);
                    Fast32UnPack8(input + 16, output + 64);
                    Fast32UnPack8(input + 24, output + 96);
                    return;
                case 9:
                    Fast32UnPack9(input, output);
                    Fast32UnPack9(input + 9, output + 32);
                    Fast32UnPack9(input + 18, output + 64);
                    Fast32UnPack9(input + 27, output + 96);
                    return;
                case 10:
                    Fast32UnPack10(input, output);
                    Fast32UnPack10(input + 10, output + 32);
                    Fast32UnPack10(input + 20, output + 64);
                    Fast32UnPack10(input + 30, output + 96);
                    return;
                case 11:
                    Fast32UnPack11(input, output);
                    Fast32UnPack11(input + 11, output + 32);
                    Fast32UnPack11(input + 22, output + 64);
                    Fast32UnPack11(input + 33, output + 96);
                    return;
                case 12:
                    Fast32UnPack12(input, output);
                    Fast32UnPack12(input + 12, output + 32);
                    Fast32UnPack12(input + 24, output + 64);
                    Fast32UnPack12(input + 36, output + 96);
                    return;
                default:
                    PackUtils.Unpack(input, width, 128, 0, 0, output);
                    return;
            }
        }

        private static unsafe void Fast32Pack(int width, uint* input, uint* output)
        {
            switch (width)
            {
                case 0:
                    return;
                case 1:
                    Fast32Pack1(input, output);
                    return;
                case 2:
                    Fast32Pack2(input, output);
                    return;
                case 3:
                    Fast32Pack3(input, output);
                    return;
                case 4:
                    Fast32Pack4(input, output);
                    return;
                case 5:
                    Fast32Pack5(input, output);
                    return;
                case 6:
                    Fast32Pack6(input, output);
                    return;
                case 7:
                    Fast32Pack7(input, output);
                    return;
                case 8:
                    Fast32Pack8(input, output);
                    return;
                default:
                    PackUtils.Pack(input, width, 32, output);
                    return;
            }
        }

        private static unsafe void Fast32UnPack(int width, uint* input, uint* output)
        {
            switch (width)
            {
                case 0:
                    Fast32UnPack0(input, output);
                    return;
                case 1:
                    Fast32UnPack1(input, output);
                    return;
                case 2:
                    Fast32UnPack2(input, output);
                    return;
                case 3:
                    Fast32UnPack3(input, output);
                    return;
                case 4:
                    Fast32UnPack4(input, output);
                    return;
                case 5:
                    Fast32UnPack5(input, output);
                    return;
                case 6:
                    Fast32UnPack6(input, output);
                    return;
                case 7:
                    Fast32UnPack7(input, output);
                    return;
                case 8:
                    Fast32UnPack8(input, output);
                    return;
                case 9:
                    Fast32UnPack9(input, output);
                    return;
                case 10:
                    Fast32UnPack10(input, output);
                    return;
                default:
                    PackUtils.Unpack(input, width, 32, 0, 0, output);
                    return;
            }
        }

        #region Packers

        private static unsafe void Fast32UnPack0(uint* input, uint* output)
        {
            for (int i = 0; i < 128; i++)
            {
                output[i] = 0;
            }
        }

        private static unsafe void Fast32Pack1(uint* input, uint* output)
        {
            for (int i = 0; i < 32; i++)
            { // will probably be unrolled
                output[0] |= input[i] << i;
            }
        }

        private static unsafe void Fast32UnPack1(uint* input, uint* output)
        {
            output[0] = (input[0]) & 1;
            output[1] = (input[0] >> 1) & 1;
            output[2] = (input[0] >> 2) & 1;
            output[3] = (input[0] >> 3) & 1;
            output[4] = (input[0] >> 4) & 1;
            output[5] = (input[0] >> 5) & 1;
            output[6] = (input[0] >> 6) & 1;
            output[7] = (input[0] >> 7) & 1;
            output[8] = (input[0] >> 8) & 1;
            output[9] = (input[0] >> 9) & 1;
            output[10] = (input[0] >> 10) & 1;
            output[11] = (input[0] >> 11) & 1;
            output[12] = (input[0] >> 12) & 1;
            output[13] = (input[0] >> 13) & 1;
            output[14] = (input[0] >> 14) & 1;
            output[15] = (input[0] >> 15) & 1;
            output[16] = (input[0] >> 16) & 1;
            output[17] = (input[0] >> 17) & 1;
            output[18] = (input[0] >> 18) & 1;
            output[19] = (input[0] >> 19) & 1;
            output[20] = (input[0] >> 20) & 1;
            output[21] = (input[0] >> 21) & 1;
            output[22] = (input[0] >> 22) & 1;
            output[23] = (input[0] >> 23) & 1;
            output[24] = (input[0] >> 24) & 1;
            output[25] = (input[0] >> 25) & 1;
            output[26] = (input[0] >> 26) & 1;
            output[27] = (input[0] >> 27) & 1;
            output[28] = (input[0] >> 28) & 1;
            output[29] = (input[0] >> 29) & 1;
            output[30] = (input[0] >> 30) & 1;
            output[31] = (input[0] >> 31) & 1;
        }

        private static unsafe void Fast32Pack2(uint* input, uint* output)
        {
            Fast16Pack2(input, output);
            Fast16Pack2(input + 16, output + 1);
        }

        private static unsafe void Fast32UnPack2(uint* input, uint* output)
        {
            output[0] = (input[0] >> 0) & 3;
            output[1] = (input[0] >> 2) & 3;
            output[2] = (input[0] >> 4) & 3;
            output[3] = (input[0] >> 6) & 3;
            output[4] = (input[0] >> 8) & 3;
            output[5] = (input[0] >> 10) & 3;
            output[6] = (input[0] >> 12) & 3;
            output[7] = (input[0] >> 14) & 3;
            output[8] = (input[0] >> 16) & 3;
            output[9] = (input[0] >> 18) & 3;
            output[10] = (input[0] >> 20) & 3;
            output[11] = (input[0] >> 22) & 3;
            output[12] = (input[0] >> 24) & 3;
            output[13] = (input[0] >> 26) & 3;
            output[14] = (input[0] >> 28) & 3;
            output[15] = (input[0] >> 30) & 3;
            output[16] = (input[1] >> 0) & 3;
            output[17] = (input[1] >> 2) & 3;
            output[18] = (input[1] >> 4) & 3;
            output[19] = (input[1] >> 6) & 3;
            output[20] = (input[1] >> 8) & 3;
            output[21] = (input[1] >> 10) & 3;
            output[22] = (input[1] >> 12) & 3;
            output[23] = (input[1] >> 14) & 3;
            output[24] = (input[1] >> 16) & 3;
            output[25] = (input[1] >> 18) & 3;
            output[26] = (input[1] >> 20) & 3;
            output[27] = (input[1] >> 22) & 3;
            output[28] = (input[1] >> 24) & 3;
            output[29] = (input[1] >> 26) & 3;
            output[30] = (input[1] >> 28) & 3;
            output[31] = (input[1] >> 30) & 3;
        }

        private static unsafe void Fast16Pack2(uint* input, uint* output)
        {
            for (int i = 0; i < 16; i++)
            { // will probably be unrolled
                output[0] |= input[i] << (i << 1);
            }
        }

        private static unsafe void Fast32Pack3(uint* input, uint* output)
        {
            output[0] = input[0];
            output[0] |= input[1] << 3;
            output[0] |= input[2] << 6;
            output[0] |= input[3] << 9;
            output[0] |= input[4] << 12;
            output[0] |= input[5] << 15;
            output[0] |= input[6] << 18;
            output[0] |= input[7] << 21;
            output[0] |= input[8] << 24;
            output[0] |= input[9] << 27;
            output[0] |= input[10] << 30;
            output[1] = input[10] >> 2;

            output[1] |= input[11] << 1;
            output[1] |= input[12] << 4;
            output[1] |= input[13] << 7;
            output[1] |= input[14] << 10;
            output[1] |= input[15] << 13;
            output[1] |= input[16] << 16;
            output[1] |= input[17] << 19;
            output[1] |= input[18] << 22;
            output[1] |= input[19] << 25;
            output[1] |= input[20] << 28;
            output[1] |= input[21] << 31;
            output[2] = input[21] >> 1;

            output[2] |= input[22] << 2;
            output[2] |= input[23] << 5;
            output[2] |= input[24] << 8;
            output[2] |= input[25] << 11;
            output[2] |= input[26] << 14;
            output[2] |= input[27] << 17;
            output[2] |= input[28] << 20;
            output[2] |= input[29] << 23;
            output[2] |= input[30] << 26;
            output[2] |= input[31] << 29;
        }

        private static unsafe void Fast32UnPack3(uint* input, uint* output)
        {
            output[0] = 7 & input[0];
            output[1] = 7 & (input[0] >> 3);
            output[2] = 7 & (input[0] >> 6);
            output[3] = 7 & (input[0] >> 9);
            output[4] = 7 & (input[0] >> 12);
            output[5] = 7 & (input[0] >> 15);
            output[6] = 7 & (input[0] >> 18);
            output[7] = 7 & (input[0] >> 21);
            output[8] = 7 & (input[0] >> 24);
            output[9] = 7 & (input[0] >> 27);
            output[10] = 3 & (input[0] >> 30);
            output[10] |= 4 & (input[1] << 2);

            output[11] = 7 & (input[1] >> 1);
            output[12] = 7 & (input[1] >> 4);
            output[13] = 7 & (input[1] >> 7);
            output[14] = 7 & (input[1] >> 10);
            output[15] = 7 & (input[1] >> 13);
            output[16] = 7 & (input[1] >> 16);
            output[17] = 7 & (input[1] >> 19);
            output[18] = 7 & (input[1] >> 22);
            output[19] = 7 & (input[1] >> 25);
            output[20] = 7 & (input[1] >> 28);
            output[21] = 1 & (input[1] >> 31);
            output[21] |= 6 & (input[2] << 1);

            output[22] = 7 & (input[2] >> 2);
            output[23] = 7 & (input[2] >> 5);
            output[24] = 7 & (input[2] >> 8);
            output[25] = 7 & (input[2] >> 11);
            output[26] = 7 & (input[2] >> 14);
            output[27] = 7 & (input[2] >> 17);
            output[28] = 7 & (input[2] >> 20);
            output[29] = 7 & (input[2] >> 23);
            output[30] = 7 & (input[2] >> 26);
            output[31] = 7 & (input[2] >> 29);
        }

        private static unsafe void Fast8Pack4(uint* input, uint* output)
        {
            output[0] = input[0];
            output[0] |= input[1] << 4;
            output[0] |= input[2] << 8;
            output[0] |= input[3] << 12;
            output[0] |= input[4] << 16;
            output[0] |= input[5] << 20;
            output[0] |= input[6] << 24;
            output[0] |= input[7] << 28;
        }

        private static unsafe void Fast32Pack4(uint* input, uint* output)
        {
            Fast8Pack4(input, output);
            Fast8Pack4(input + 8, output + 1);
            Fast8Pack4(input + 16, output + 2);
            Fast8Pack4(input + 24, output + 3);
        }

        private static unsafe void Fast32UnPack4(uint* input, uint* output)
        {
            output[0] = (input[0] >> 0) & 15;
            output[1] = (input[0] >> 4) & 15;
            output[2] = (input[0] >> 8) & 15;
            output[3] = (input[0] >> 12) & 15;
            output[4] = (input[0] >> 16) & 15;
            output[5] = (input[0] >> 20) & 15;
            output[6] = (input[0] >> 24) & 15;
            output[7] = (input[0] >> 28) & 15;
            output[8] = (input[1] >> 0) & 15;
            output[9] = (input[1] >> 4) & 15;
            output[10] = (input[1] >> 8) & 15;
            output[11] = (input[1] >> 12) & 15;
            output[12] = (input[1] >> 16) & 15;
            output[13] = (input[1] >> 20) & 15;
            output[14] = (input[1] >> 24) & 15;
            output[15] = (input[1] >> 28) & 15;
            output[16] = (input[2] >> 0) & 15;
            output[17] = (input[2] >> 4) & 15;
            output[18] = (input[2] >> 8) & 15;
            output[19] = (input[2] >> 12) & 15;
            output[20] = (input[2] >> 16) & 15;
            output[21] = (input[2] >> 20) & 15;
            output[22] = (input[2] >> 24) & 15;
            output[23] = (input[2] >> 28) & 15;
            output[24] = (input[3] >> 0) & 15;
            output[25] = (input[3] >> 4) & 15;
            output[26] = (input[3] >> 8) & 15;
            output[27] = (input[3] >> 12) & 15;
            output[28] = (input[3] >> 16) & 15;
            output[29] = (input[3] >> 20) & 15;
            output[30] = (input[3] >> 24) & 15;
            output[31] = (input[3] >> 28) & 15;
        }

        private static unsafe void Fast32Pack5(uint* input, uint* output)
        {
            output[0] |= 31 & input[0];
            output[0] |= input[1] << 5;
            output[0] |= input[2] << 10;
            output[0] |= input[3] << 15;
            output[0] |= input[4] << 20;
            output[0] |= input[5] << 25;
            output[0] |= input[6] << 30;
            output[1] |= input[6] >> 2;

            output[1] |= input[7] << 3;
            output[1] |= input[8] << 8;
            output[1] |= input[9] << 13;
            output[1] |= input[10] << 18;
            output[1] |= input[11] << 23;
            output[1] |= input[12] << 28;
            output[2] |= input[12] >> 4;

            output[2] |= input[13] << 1;
            output[2] |= input[14] << 6;
            output[2] |= input[15] << 11;
            output[2] |= input[16] << 16;
            output[2] |= input[17] << 21;
            output[2] |= input[18] << 26;
            output[2] |= input[19] << 31;
            output[3] |= input[19] >> 1;

            output[3] |= input[20] << 4;
            output[3] |= input[21] << 9;
            output[3] |= input[22] << 14;
            output[3] |= input[23] << 19;
            output[3] |= input[24] << 24;
            output[3] |= input[25] << 29;
            output[4] |= input[25] >> 3;

            output[4] |= input[26] << 2;
            output[4] |= input[27] << 7;
            output[4] |= input[28] << 12;
            output[4] |= input[29] << 17;
            output[4] |= input[30] << 22;
            output[4] |= input[31] << 27;
        }

        private static unsafe void Fast32UnPack5(uint* input, uint* output)
        {
            output[0] = 31 & input[0];
            output[1] = 31 & (input[0] >> 5);
            output[2] = 31 & (input[0] >> 10);
            output[3] = 31 & (input[0] >> 15);
            output[4] = 31 & (input[0] >> 20);
            output[5] = 31 & (input[0] >> 25);
            output[6] = 3 & (input[0] >> 30);
            output[6] |= 28 & (input[1] << 2);

            output[7] = 31 & (input[1] >> 3);
            output[8] = 31 & (input[1] >> 8);
            output[9] = 31 & (input[1] >> 13);
            output[10] = 31 & (input[1] >> 18);
            output[11] = 31 & (input[1] >> 23);
            output[12] = 15 & (input[1] >> 28);
            output[12] |= 16 & (input[2] << 4);

            output[13] = 31 & (input[2] >> 1);
            output[14] = 31 & (input[2] >> 6);
            output[15] = 31 & (input[2] >> 11);
            output[16] = 31 & (input[2] >> 16);
            output[17] = 31 & (input[2] >> 21);
            output[18] = 31 & (input[2] >> 26);
            output[19] = 1 & (input[2] >> 31);
            output[19] |= 30 & (input[3] << 1);

            output[20] = 31 & (input[3] >> 4);
            output[21] = 31 & (input[3] >> 9);
            output[22] = 31 & (input[3] >> 14);
            output[23] = 31 & (input[3] >> 19);
            output[24] = 31 & (input[3] >> 24);
            output[25] = 7 & (input[3] >> 29);
            output[25] |= 24 & (input[4] << 3);

            output[26] = 31 & (input[4] >> 2);
            output[27] = 31 & (input[4] >> 7);
            output[28] = 31 & (input[4] >> 12);
            output[29] = 31 & (input[4] >> 17);
            output[30] = 31 & (input[4] >> 22);
            output[31] = 31 & (input[4] >> 27);
        }

        private static unsafe void Fast32Pack6(uint* input, uint* output)
        {
            Fast16Pack6(input, output);
            Fast16Pack6(input + 16, output + 3);
        }

        private static unsafe void Fast32UnPack6(uint* input, uint* output)
        {
            output[0] = (input[0] >> 0) & 63;
            output[1] = (input[0] >> 6) & 63;
            output[2] = (input[0] >> 12) & 63;
            output[3] = (input[0] >> 18) & 63;
            output[4] = (input[0] >> 24) & 63;
            output[5] = (input[0] >> 30) & 3;
            output[5] |= (input[1] << 2) & 60;

            output[6] = (input[1] >> 4) & 63;
            output[7] = (input[1] >> 10) & 63;
            output[8] = (input[1] >> 16) & 63;
            output[9] = (input[1] >> 22) & 63;
            output[10] = (input[1] >> 28) & 15;
            output[10] |= (input[2] << 4) & 48;

            output[11] = (input[2] >> 2) & 63;
            output[12] = (input[2] >> 8) & 63;
            output[13] = (input[2] >> 14) & 63;
            output[14] = (input[2] >> 20) & 63;
            output[15] = (input[2] >> 26) & 63;
            output[16] = (input[3] >> 0) & 63;
            output[17] = (input[3] >> 6) & 63;
            output[18] = (input[3] >> 12) & 63;
            output[19] = (input[3] >> 18) & 63;
            output[20] = (input[3] >> 24) & 63;
            output[21] = (input[3] >> 30) & 3;
            output[21] |= (input[4] << 2) & 60;

            output[22] = (input[4] >> 4) & 63;
            output[23] = (input[4] >> 10) & 63;
            output[24] = (input[4] >> 16) & 63;
            output[25] = (input[4] >> 22) & 63;
            output[26] = (input[4] >> 28) & 15;
            output[26] |= (input[5] << 4) & 48;

            output[27] = (input[5] >> 2) & 63;
            output[28] = (input[5] >> 8) & 63;
            output[29] = (input[5] >> 14) & 63;
            output[30] = (input[5] >> 20) & 63;
            output[31] = (input[5] >> 26) & 63;
        }

        private static unsafe void Fast16Pack6(uint* input, uint* output)
        {
            output[0] |= input[0];
            output[0] |= input[1] << 6;
            output[0] |= input[2] << 12;
            output[0] |= input[3] << 18;
            output[0] |= input[4] << 24;
            output[0] |= input[5] << 30;
            output[1] |= input[5] >> 2;

            output[1] |= input[6] << 4;
            output[1] |= input[7] << 10;
            output[1] |= input[8] << 16;
            output[1] |= input[9] << 22;
            output[1] |= input[10] << 28;
            output[2] |= input[10] >> 4;

            output[2] |= input[11] << 2;
            output[2] |= input[12] << 8;
            output[2] |= input[13] << 14;
            output[2] |= input[14] << 20;
            output[2] |= input[15] << 26;
        }

        private static unsafe void Fast32Pack7(uint* input, uint* output)
        {
            output[0] = input[0];
            output[0] |= input[1] << 7;
            output[0] |= input[2] << 14;
            output[0] |= input[3] << 21;
            output[0] |= input[4] << 28;
            output[1] = input[4] >> 4;

            output[1] |= input[5] << 3;
            output[1] |= input[6] << 10;
            output[1] |= input[7] << 17;
            output[1] |= input[8] << 24;
            output[1] |= input[9] << 31;
            output[2] = input[9] >> 1;

            output[2] |= input[10] << 6;
            output[2] |= input[11] << 13;
            output[2] |= input[12] << 20;
            output[2] |= input[13] << 27;
            output[3] = input[13] >> 5;
                
            output[3] |= input[14] << 2;
            output[3] |= input[15] << 9;
            output[3] |= input[16] << 16;
            output[3] |= input[17] << 23;
            output[3] |= input[18] << 30;
            output[4] = input[18] >> 2;

            output[4] |= input[19] << 5;
            output[4] |= input[20] << 12;
            output[4] |= input[21] << 19;
            output[4] |= input[22] << 26;
            output[5] = input[22] >> 6;
                
            output[5] |= input[23] << 1;
            output[5] |= input[24] << 8;
            output[5] |= input[25] << 15;
            output[5] |= input[26] << 22;
            output[5] |= input[27] << 29;
            output[6] = input[27] >> 3;

            output[6] |= input[28] << 4;
            output[6] |= input[29] << 11;
            output[6] |= input[30] << 18;
            output[6] |= input[31] << 25;
        }

        private static unsafe void Fast32UnPack7(uint* input, uint* output)
        {
            output[0] = 127 & input[0];
            output[1] = 127 & (input[0] >> 7);
            output[2] = 127 & (input[0] >> 14);
            output[3] = 127 & (input[0] >> 21);
            output[4] = 15 & (input[0] >> 28);
            output[4] |= 112 & (input[1] << 4);

            output[5] = 127 & (input[1] >> 3);
            output[6] = 127 & (input[1] >> 10);
            output[7] = 127 & (input[1] >> 17);
            output[8] = 127 & (input[1] >> 24);
            output[9] = 1 & (input[1] >> 31);
            output[9] |= 126 & (input[2] << 1);

            output[10] = 127 & (input[2] >> 6);
            output[11] = 127 & (input[2] >> 13);
            output[12] = 127 & (input[2] >> 20);
            output[13] = 31 & (input[2] >> 27);
            output[13] |= 96 & (input[3] << 5);

            output[14] = 127 & (input[3] >> 2);
            output[15] = 127 & (input[3] >> 9);
            output[16] = 127 & (input[3] >> 16);
            output[17] = 127 & (input[3] >> 23);
            output[18] = 3 & (input[3] >> 30);
            output[18] |= 124 & (input[4] << 2);

            output[19] = 127 & (input[4] >> 5);
            output[20] = 127 & (input[4] >> 12);
            output[21] = 127 & (input[4] >> 19);
            output[22] = 63 & (input[4] >> 26);
            output[22] |= 64 & (input[5] << 6);

            output[23] = 127 & (input[5] >> 1);
            output[24] = 127 & (input[5] >> 8);
            output[25] = 127 & (input[5] >> 15);
            output[26] = 127 & (input[5] >> 22);
            output[27] = 7 & (input[5] >> 29);
            output[27] |= 120 & (input[6] << 3);

            output[28] = 127 & (input[6] >> 4);
            output[29] = 127 & (input[6] >> 11);
            output[30] = 127 & (input[6] >> 18);
            output[31] = 127 & (input[6] >> 25);
        }

        private static unsafe void Fast32Pack8(uint* input, uint* output)
        {
            Fast4Pack8(input, output);
            Fast4Pack8(input + 4, output + 1);
            Fast4Pack8(input + 8, output + 2);
            Fast4Pack8(input + 12, output + 3);
            Fast4Pack8(input + 16, output + 4);
            Fast4Pack8(input + 20, output + 5);
            Fast4Pack8(input + 24, output + 6);
            Fast4Pack8(input + 28, output + 7);
        }

        private static unsafe void Fast32UnPack8(uint* input, uint* output)
        {
            Fast4UnPack8(input, output);
            Fast4UnPack8(input + 1, output + 4);
            Fast4UnPack8(input + 2, output + 8);
            Fast4UnPack8(input + 3, output + 12);
            Fast4UnPack8(input + 4, output + 16);
            Fast4UnPack8(input + 5, output + 20);
            Fast4UnPack8(input + 6, output + 24);
            Fast4UnPack8(input + 7, output + 28);
        }

        private static unsafe void Fast4Pack8(uint* input, uint* output)
        {
            output[0] |= input[0];
            output[0] |= input[1] << 8;
            output[0] |= input[2] << 16;
            output[0] |= input[3] << 24;
        }

        private static unsafe void Fast4UnPack8(uint* input, uint* output)
        {
            byte* inp = (byte*) input;
            output[0] = inp[0];
            output[1] = inp[1];
            output[2] = inp[2];
            output[3] = inp[3];
        }

        private static unsafe void Fast32Pack9(uint* input, uint* output)
        {
            output[0] = input[0];
            output[0] |= input[1] << 9;
            output[0] |= input[2] << 18;
            output[0] |= input[3] << 27;
            output[1] = input[3] >> 5;

            output[1] |= input[4] << 4;
            output[1] |= input[5] << 13;
            output[1] |= input[6] << 22;
            output[1] |= input[7] << 31;
            output[2] = input[7] >> 1;

            output[2] |= input[8] << 8;
            output[2] |= input[9] << 17;
            output[2] |= input[10] << 26;
            output[3] = input[10] >> 6;

            output[3] |= input[11] << 3;
            output[3] |= input[12] << 12;
            output[3] |= input[13] << 21;
            output[3] |= input[14] << 30;
            output[4] = input[14] >> 2;

            output[4] |= input[15] << 7;
            output[4] |= input[16] << 16;
            output[4] |= input[17] << 25;
            output[5] = input[17] >> 7;

            output[5] |= input[18] << 2;
            output[5] |= input[19] << 11;
            output[5] |= input[20] << 20;
            output[5] |= input[21] << 29;
            output[6] = input[21] >> 3;

            output[6] |= input[22] << 6;
            output[6] |= input[23] << 15;
            output[6] |= input[24] << 24;
            output[7] = input[24] >> 8;

            output[7] |= input[25] << 1;
            output[7] |= input[26] << 10;
            output[7] |= input[27] << 19;
            output[7] |= input[28] << 28;
            output[8] = input[28] >> 4;

            output[8] |= input[29] << 5;
            output[8] |= input[30] << 14;
            output[8] |= input[31] << 23;
        }

        private static unsafe void Fast32UnPack9(uint* input, uint* output)
        {
            output[0] = (input[0] >> 0) & 511;
            output[1] = (input[0] >> 9) & 511;
            output[2] = (input[0] >> 18) & 511;
            output[3] = (input[0] >> 27) & 31;
            output[3] |= (input[1] << 5) & 480;

            output[4] = (input[1] >> 4) & 511;
            output[5] = (input[1] >> 13) & 511;
            output[6] = (input[1] >> 22) & 511;
            output[7] = (input[1] >> 31) & 1;
            output[7] |= (input[2] << 1) & 510;

            output[8] = (input[2] >> 8) & 511;
            output[9] = (input[2] >> 17) & 511;
            output[10] = (input[2] >> 26) & 63;
            output[10] |= (input[3] << 6) & 448;

            output[11] = (input[3] >> 3) & 511;
            output[12] = (input[3] >> 12) & 511;
            output[13] = (input[3] >> 21) & 511;
            output[14] = (input[3] >> 30) & 3;
            output[14] |= (input[4] << 2) & 508;

            output[15] = (input[4] >> 7) & 511;
            output[16] = (input[4] >> 16) & 511;
            output[17] = (input[4] >> 25) & 127;
            output[17] |= (input[5] << 7) & 384;

            output[18] = (input[5] >> 2) & 511;
            output[19] = (input[5] >> 11) & 511;
            output[20] = (input[5] >> 20) & 511;
            output[21] = (input[5] >> 29) & 7;
            output[21] |= (input[6] << 3) & 504;

            output[22] = (input[6] >> 6) & 511;
            output[23] = (input[6] >> 15) & 511;
            output[24] = (input[6] >> 24) & 255;
            output[24] |= (input[7] << 8) & 256;

            output[25] = (input[7] >> 1) & 511;
            output[26] = (input[7] >> 10) & 511;
            output[27] = (input[7] >> 19) & 511;
            output[28] = (input[7] >> 28) & 15;
            output[28] |= (input[8] << 4) & 496;

            output[29] = (input[8] >> 5) & 511;
            output[30] = (input[8] >> 14) & 511;
            output[31] = (input[8] >> 23) & 511;
        }

        private static unsafe void Fast32UnPack10(uint* input, uint* output)
        {
            output[0] = (input[0] >> 0) & 1023;
            output[1] = (input[0] >> 10) & 1023;
            output[2] = (input[0] >> 20) & 1023;
            output[3] = (input[0] >> 30) & 3;
            output[3] |= (input[1] << 2) & 1020;

            output[4] = (input[1] >> 8) & 1023;
            output[5] = (input[1] >> 18) & 1023;
            output[6] = (input[1] >> 28) & 15;
            output[6] |= (input[2] << 4) & 1008;

            output[7] = (input[2] >> 6) & 1023;
            output[8] = (input[2] >> 16) & 1023;
            output[9] = (input[2] >> 26) & 63;
            output[9] |= (input[3] << 6) & 960;

            output[10] = (input[3] >> 4) & 1023;
            output[11] = (input[3] >> 14) & 1023;
            output[12] = (input[3] >> 24) & 255;
            output[12] |= (input[4] << 8) & 768;

            output[13] = (input[4] >> 2) & 1023;
            output[14] = (input[4] >> 12) & 1023;
            output[15] = (input[4] >> 22) & 1023;
            output[16] = (input[5] >> 0) & 1023;
            output[17] = (input[5] >> 10) & 1023;
            output[18] = (input[5] >> 20) & 1023;
            output[19] = (input[5] >> 30) & 3;
            output[19] |= (input[6] << 2) & 1020;

            output[20] = (input[6] >> 8) & 1023;
            output[21] = (input[6] >> 18) & 1023;
            output[22] = (input[6] >> 28) & 15;
            output[22] |= (input[7] << 4) & 1008;

            output[23] = (input[7] >> 6) & 1023;
            output[24] = (input[7] >> 16) & 1023;
            output[25] = (input[7] >> 26) & 63;
            output[25] |= (input[8] << 6) & 960;

            output[26] = (input[8] >> 4) & 1023;
            output[27] = (input[8] >> 14) & 1023;
            output[28] = (input[8] >> 24) & 255;
            output[28] |= (input[9] << 8) & 768;

            output[29] = (input[9] >> 2) & 1023;
            output[30] = (input[9] >> 12) & 1023;
            output[31] = (input[9] >> 22) & 1023;
        }

        private static unsafe void Fast32UnPack11(uint* input, uint* output)
        {
            output[0] = input[0] & 2047;
            output[1] = (input[0] >> 11) & 2047;
            output[2] = (input[0] >> 22) & 1023;
            output[2] |= (input[1] << 10) & 1024;

            output[3] = (input[1] >> 1) & 2047;
            output[4] = (input[1] >> 12) & 2047;
            output[5] = (input[1] >> 23) & 511;
            output[5] |= (input[2] << 9) & 1536;

            output[6] = (input[2] >> 2) & 2047;
            output[7] = (input[2] >> 13) & 2047;
            output[8] = (input[2] >> 24) & 255;
            output[8] |= (input[3] << 8) & 1792;

            output[9] = (input[3] >> 3) & 2047;
            output[10] = (input[3] >> 14) & 2047;
            output[11] = (input[3] >> 25) & 127;
            output[11] |= (input[4] << 7) & 1920;

            output[12] = (input[4] >> 4) & 2047;
            output[13] = (input[4] >> 15) & 2047;
            output[14] = (input[4] >> 26) & 63;
            output[14] |= (input[5] << 6) & 1984;

            output[15] = (input[5] >> 5) & 2047;
            output[16] = (input[5] >> 16) & 2047;
            output[17] = (input[5] >> 27) & 31;
            output[17] |= (input[6] << 5) & 2016;

            output[18] = (input[6] >> 6) & 2047;
            output[19] = (input[6] >> 17) & 2047;
            output[20] = (input[6] >> 28) & 15;
            output[20] |= (input[7] << 4) & 2032;

            output[21] = (input[7] >> 7) & 2047;
            output[22] = (input[7] >> 18) & 2047;
            output[23] = (input[7] >> 29) & 7;
            output[23] |= (input[8] << 3) & 2040;

            output[24] = (input[8] >> 8) & 2047;
            output[25] = (input[8] >> 19) & 2047;
            output[26] = (input[8] >> 30) & 3;
            output[26] |= (input[9] << 2) & 2044;

            output[27] = (input[9] >> 9) & 2047;
            output[28] = (input[9] >> 20) & 2047;
            output[29] = (input[9] >> 31) & 1;
            output[29] |= (input[10] << 1) & 2046;

            output[30] = (input[10] >> 10) & 2047;
            output[31] = (input[10] >> 21) & 2047;
        }

        private static unsafe void Fast32UnPack12(uint* input, uint* output)
        {
            output[0] = input[0] & 4095;
            output[1] = (input[0] >> 12) & 4095;
            output[2] = (input[0] >> 24) & 255;
            output[2] |= (input[1] << 8) & 3840;

            output[3] = (input[1] >> 4) & 4095;
            output[4] = (input[1] >> 16) & 4095;
            output[5] = (input[1] >> 28) & 15;
            output[5] |= (input[2] << 4) & 4080;

            output[6] = (input[2] >> 8) & 4095;
            output[7] = (input[2] >> 20) & 4095;
            output[8] = (input[3] >> 0) & 4095;
            output[9] = (input[3] >> 12) & 4095;
            output[10] = (input[3] >> 24) & 255;
            output[10] |= (input[4] << 8) & 3840;

            output[11] = (input[4] >> 4) & 4095;
            output[12] = (input[4] >> 16) & 4095;
            output[13] = (input[4] >> 28) & 15;
            output[13] |= (input[5] << 4) & 4080;

            output[14] = (input[5] >> 8) & 4095;
            output[15] = (input[5] >> 20) & 4095;
            output[16] = (input[6] >> 0) & 4095;
            output[17] = (input[6] >> 12) & 4095;
            output[18] = (input[6] >> 24) & 255;
            output[18] |= (input[7] << 8) & 3840;

            output[19] = (input[7] >> 4) & 4095;
            output[20] = (input[7] >> 16) & 4095;
            output[21] = (input[7] >> 28) & 15;
            output[21] |= (input[8] << 4) & 4080;

            output[22] = (input[8] >> 8) & 4095;
            output[23] = (input[8] >> 20) & 4095;
            output[24] = (input[9] >> 0) & 4095;
            output[25] = (input[9] >> 12) & 4095;
            output[26] = (input[9] >> 24) & 255;
            output[26] |= (input[10] << 8) & 3840;

            output[27] = (input[10] >> 4) & 4095;
            output[28] = (input[10] >> 16) & 4095;
            output[29] = (input[10] >> 28) & 15;
            output[29] |= (input[11] << 4) & 4080;

            output[30] = (input[11] >> 8) & 4095;
            output[31] = (input[11] >> 20) & 4095;
        }

        #endregion
    }
}


