﻿namespace NETing.Index.Utils.Compression
{
    public static class PackUtils
    {
        internal static unsafe int Pack(uint* input, int width, int n, uint* output)
        {
            if (32 % width == 0)
            {
                int bitPosition = 0;
                for (int i = 0; i < n; i++, bitPosition += width)
                {
                    int wordPosition = bitPosition >> 5;
                    int intraWordPosition = bitPosition & 0x1f;
                    output[wordPosition] |= input[i] << intraWordPosition;
                }

                return bitPosition;
            }
            else
            {
                int bitPosition = 0;
                for (int i = 0; i < n; i++, bitPosition += width)
                {
                    int wordPosition = bitPosition >> 5;
                    int intraWordPosition = bitPosition & 0x1f;
                    int overflow = intraWordPosition + width - 32;
                    if (overflow <= 0)
                    {
                        output[wordPosition] |= input[i] << intraWordPosition;
                    }
                    else
                    {
                        output[wordPosition] |= input[i] << intraWordPosition;
                        output[wordPosition + 1] = input[i] >> (width - overflow);
                    }
                }

                return bitPosition;
            }
        }

        private static readonly uint[] masks = new uint[]
                                                   {
                                                       0, 1, 3, 7, 0xf, 
                                                       0x1f, 0x3f, 0x7f, 0xff, 
                                                       0x1ff, 0x3ff, 0x7ff, 0xfff, 
                                                       0x1fff, 0x3fff, 0x7fff, 0xffff, 
                                                       0x1ffff, 0x3ffff, 0x7ffff, 0xfffff,
                                                       0x1fffff, 0x3fffff, 0x7fffff, 0xffffff, 
                                                       0x1ffffff, 0x3ffffff, 0x7ffffff, 0xfffffff,
                                                       0x1fffffff, 0x3fffffff, 0x7fffffff, 0xffffffff,
                                                   };

        internal static unsafe void Unpack(uint* input, int width, int n, int inputOffset, int outputOffset, uint* output)
        {
            for (int i = 0, outputPos = outputOffset, bit = 0; i < n; i++, outputPos++, bit += width)
            {
                int word = inputOffset + (bit >> 5);
                int wordPos = bit - (word << 5);
                int remainder = 32 - width - wordPos;
                if (remainder >= 0)
                {   // inside
                    output[outputPos] = (input[word] >> wordPos) & masks[width];
                }
                else
                {   // spanning
                    remainder = -1 * remainder;
                    output[outputPos] = (input[word] >> wordPos) & masks[width];
                    output[outputPos] |= (input[word + 1] & (masks[width] >> (width - remainder))) << (width - remainder);
                }
            }
        }
    }
}