﻿namespace NETing.Index.Utils.Compression.Patched
{
    public static class PforUtils 
    {
        public static unsafe int Log2HistogramAnalyseBits(int requiredInRange, uint frameOfReference, uint[] data, int offset)
        {
            int* historgram = stackalloc int[33];
            for (int i = 0; i < 128; i++)
            {
                historgram[Utils.MostSignificantBit(data[offset + i] - frameOfReference)]++;
            }

            int cumSum = 0;
            for (int i = 0; i < 32; i++)
            {
                cumSum += historgram[i];
                if (cumSum >= requiredInRange)
                {
                    return i;
                }
            }

            return 32;
        }

        public static uint MinimumValueFrameOfReference(int count, uint[] buffer, int offset)
        {
            uint min = buffer[offset];
            for (int i = offset + 1; i < count + offset; i++)
            {
                min = Utils.Min(min, buffer[i]);
            }

            return min & 0x7ffff;
        }

        public static void UnpackControlWord(
            uint word,
            out uint bitWidth, 
            out uint chunkLength, 
            out uint numberOfExceptions, 
            out uint frameOfReference)
        {
            bitWidth = word >> 27;
            chunkLength = 128;
            numberOfExceptions = (word >> 19) & 0xff;
            frameOfReference = word & 0x7ffff;
        }

        public static uint CreateControlWord(uint bitWidth, uint numberOfExceptions, uint frameOfReference)
        {
            // 5 bit bit width, 7 bit exceptions, 20 bit frameOfReference. Chunk length is allways 128.
            return ((bitWidth & 0x1f) << 27) | ((numberOfExceptions & 0xff) << 19) | (frameOfReference & 0x7ffff);
        }
    }
}


