﻿namespace NETing.Index.Utils.Compression
{
    public static class Simple9
    {
        private static readonly uint[] firstPassLimits = new uint[]
                                                             {
                                                                 0xfffffff, // 1 28 bit
                                                                 0x3fff,    // 2 14 bit
                                                                 0x1ff,     // 3 9 bit
                                                                 0x7f,      // 4 7 bit
                                                                 0x1f,      // 5 5 bit
                                                                 0xf,
                                                                 0xf,       // 7 4 bit
                                                                 0x7,        
                                                                 0x7,       // 9 3 bit
                                                                 0x3,       
                                                                 0x3,       
                                                                 0x3,       
                                                                 0x3,       
                                                                 0x3,       // 14 2 bit
                                                                 0x1,       
                                                                 0x1,       
                                                                 0x1,       
                                                                 0x1,       
                                                                 0x1,       
                                                                 0x1,       
                                                                 0x1,       
                                                                 0x1,       
                                                                 0x1,       
                                                                 0x1,       
                                                                 0x1,       
                                                                 0x1,       
                                                                 0x1,   
                                                                 0x1,
                                                                 0x1,       // 28 1 bit
                                                             };

        private static readonly int[] firstPassNumbers = new[]
                                                             {
                                                                 1, 2, 3, 4, 5, 5, 6, 6, 7, 7, 7, 7, 7, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 9
                                                             };

        private static int EncodeValue(uint code, int numValues, int bitWidth, int inputOffset, int encodedPosition, uint[] input, uint[] encoded)
        {
            encoded[encodedPosition] = code << 28;
            int limit = Utils.Min(input.Length - inputOffset, numValues);
            int position = 28 - bitWidth;
            for (int i = 0; i < limit; i++)
            {
                encoded[encodedPosition] |= input[inputOffset + i] << position;
                position -= bitWidth;
            }

            return limit;
        }

        private static int EncodeValue(uint code, int numValues, int bitWidth, int inputOffset, int encodedPosition, int[] input, uint[] encoded)
        {
            encoded[encodedPosition] = code << 28;
            int limit = Utils.Min(input.Length - inputOffset, numValues);
            int position = 28 - bitWidth;
            for (int i = 0; i < limit; i++)
            {
                encoded[encodedPosition] |= (uint)input[inputOffset + i] << position;
                position -= bitWidth;
            }

            return limit;
        }

        private static int EncodeValue(uint code, int numValues, int bitWidth, int inputOffset, uint[] input, out uint encoded)
        {
            encoded = code << 28;
            int limit = Utils.Min(input.Length - inputOffset, numValues);
            int position = 28 - bitWidth;
            for (int i = 0; i < limit; i++)
            {
                encoded |= (uint)input[inputOffset + i] << position;
                position -= bitWidth;
            }

            return limit;
        }

        private static int DecodeValueA(uint encoded, int outputOffset, uint[] output)
        {
            output[outputOffset + 0] = (encoded >> 27) & 1;
            output[outputOffset + 1] = (encoded >> 26) & 1;
            output[outputOffset + 2] = (encoded >> 25) & 1;
            output[outputOffset + 3] = (encoded >> 24) & 1;
            output[outputOffset + 4] = (encoded >> 23) & 1;
            output[outputOffset + 5] = (encoded >> 22) & 1;
            output[outputOffset + 6] = (encoded >> 21) & 1;
            output[outputOffset + 7] = (encoded >> 20) & 1;
            output[outputOffset + 8] = (encoded >> 19) & 1;
            output[outputOffset + 9] = (encoded >> 18) & 1;
            output[outputOffset + 10] = (encoded >> 17) & 1;
            output[outputOffset + 11] = (encoded >> 16) & 1;
            output[outputOffset + 12] = (encoded >> 15) & 1;
            output[outputOffset + 13] = (encoded >> 14) & 1;
            output[outputOffset + 14] = (encoded >> 13) & 1;
            output[outputOffset + 15] = (encoded >> 12) & 1;
            output[outputOffset + 16] = (encoded >> 11) & 1;
            output[outputOffset + 17] = (encoded >> 10) & 1;
            output[outputOffset + 18] = (encoded >> 9) & 1;
            output[outputOffset + 19] = (encoded >> 8) & 1;
            output[outputOffset + 20] = (encoded >> 7) & 1;
            output[outputOffset + 21] = (encoded >> 6) & 1;
            output[outputOffset + 22] = (encoded >> 5) & 1;
            output[outputOffset + 23] = (encoded >> 4) & 1;
            output[outputOffset + 24] = (encoded >> 3) & 1;
            output[outputOffset + 25] = (encoded >> 2) & 1;
            output[outputOffset + 26] = (encoded >> 1) & 1;
            output[outputOffset + 27] = encoded & 1;

            return 28;
        }

        private static int DecodeValueB(uint encoded, int outputOffset, uint[] output)
        {
            output[outputOffset] = (encoded >> 26) & 3;
            output[outputOffset + 1] = (encoded >> 24) & 3;
            output[outputOffset + 2] = (encoded >> 22) & 3;
            output[outputOffset + 3] = (encoded >> 20) & 3;
            output[outputOffset + 4] = (encoded >> 18) & 3;
            output[outputOffset + 5] = (encoded >> 16) & 3;
            output[outputOffset + 6] = (encoded >> 14) & 3;
            output[outputOffset + 7] = (encoded >> 12) & 3;
            output[outputOffset + 8] = (encoded >> 10) & 3;
            output[outputOffset + 9] = (encoded >> 8) & 3;
            output[outputOffset + 10] = (encoded >> 6) & 3;
            output[outputOffset + 11] = (encoded >> 4) & 3;
            output[outputOffset + 12] = (encoded >> 2) & 3;
            output[outputOffset + 13] = encoded & 3;

            return 14;
        }

        private static int DecodeValueC(uint encoded, int outputOffset, uint[] output)
        {
            output[outputOffset] = (encoded >> 25) & 7;
            output[outputOffset + 1] = (encoded >> 22) & 7;
            output[outputOffset + 2] = (encoded >> 19) & 7;
            output[outputOffset + 3] = (encoded >> 16) & 7;
            output[outputOffset + 4] = (encoded >> 13) & 7;
            output[outputOffset + 5] = (encoded >> 10) & 7;
            output[outputOffset + 6] = (encoded >> 7) & 7;
            output[outputOffset + 7] = (encoded >> 4) & 7;
            output[outputOffset + 8] = (encoded >> 1) & 7;

            return 9;
        }

        private static int DecodeValueD(uint encoded, int outputOffset, uint[] output)
        {
            output[outputOffset] = (encoded >> 24) & 15;
            output[outputOffset + 1] = (encoded >> 20) & 15;
            output[outputOffset + 2] = (encoded >> 16) & 15;
            output[outputOffset + 3] = (encoded >> 12) & 15;
            output[outputOffset + 4] = (encoded >> 8) & 15;
            output[outputOffset + 5] = (encoded >> 4) & 15;
            output[outputOffset + 6] = encoded & 15;

            return 7;
        }

        private static int DecodeValueE(uint encoded, int outputOffset, uint[] output)
        {
            output[outputOffset] = (encoded >> 23) & 31;
            output[outputOffset + 1] = (encoded >> 18) & 31;
            output[outputOffset + 2] = (encoded >> 13) & 31;
            output[outputOffset + 3] = (encoded >> 8) & 31;
            output[outputOffset + 4] = (encoded >> 3) & 31;

            return 5;
        }

        private static int DecodeValueF(uint encoded, int outputOffset, uint[] output)
        {
            output[outputOffset] = (encoded >> 21) & 127;
            output[outputOffset + 1] = (encoded >> 14) & 127;
            output[outputOffset + 2] = (encoded >> 7) & 127;
            output[outputOffset + 3] = encoded & 127;

            return 4;
        }

        private static int DecodeValueG(uint encoded, int outputOffset, uint[] output)
        {
            output[outputOffset] = (encoded >> 19) & 511;
            output[outputOffset + 1] = (encoded >> 10) & 511;
            output[outputOffset + 2] = (encoded >> 1) & 511;

            return 3;
        }

        private static int DecodeValueH(uint encoded, int outputOffset, uint[] output)
        {
            for (int i = 0, pos = 14; i < 2; i++, pos -= 14)
            {
                output[outputOffset + i] = (encoded >> pos) & 16383;
            }

            return 2;
        }

        private static int DecodeValueI(uint encoded, int outputOffset, uint[] output)
        {
            output[outputOffset] = encoded & 268435455;
            return 1;
        }

        /// <summary>
        /// Writes a number of ints to the output using the Simple-16 compression method.
        /// </summary>
        /// <param name="input">The input data.</param>
        /// <param name="output">The output data.</param>
        /// <param name="inputOffset">The start offset in the input data.</param>
        /// <param name="outputPosition">The to write the compressed data to.</param>
        /// <returns>The number of values compressed.</returns>
        public static int Encode(
            uint[] input,
            uint[] output,
            int inputOffset,
            int outputPosition)
        {
            switch (SelectScheme(input, inputOffset))
            {
                case 1: return EncodeValue(8, 1, 28, inputOffset, outputPosition, input, output);
                case 2: return EncodeValue(7, 2, 14, inputOffset, outputPosition, input, output);
                case 3: return EncodeValue(6, 3, 9, inputOffset, outputPosition, input, output);
                case 4: return EncodeValue(5, 4, 7, inputOffset, outputPosition, input, output);
                case 5: return EncodeValue(4, 5, 5, inputOffset, outputPosition, input, output);
                case 6: return EncodeValue(3, 7, 4, inputOffset, outputPosition, input, output);
                case 7: return EncodeValue(2, 9, 3, inputOffset, outputPosition, input, output);
                case 8: return EncodeValue(1, 14, 2, inputOffset, outputPosition, input, output);
                default: return EncodeValue(0, 28, 1, inputOffset, outputPosition, input, output);
            }
        }

        public static int Encode(
            uint[] input,
            int inputOffset,
            out uint output)
        {
            switch (SelectScheme(input, inputOffset))
            {
                case 1: return EncodeValue(8, 1, 28, inputOffset, input, out output);
                case 2: return EncodeValue(7, 2, 14, inputOffset, input, out output);
                case 3: return EncodeValue(6, 3, 9, inputOffset, input, out output);
                case 4: return EncodeValue(5, 4, 7, inputOffset, input, out output);
                case 5: return EncodeValue(4, 5, 5, inputOffset, input, out output);
                case 6: return EncodeValue(3, 7, 4, inputOffset, input, out output);
                case 7: return EncodeValue(2, 9, 3, inputOffset, input, out output);
                case 8: return EncodeValue(1, 14, 2, inputOffset, input, out output);
                default: return EncodeValue(0, 28, 1, inputOffset, input, out output);
            }
        }

        public static int Encode(
            int[] input,
            uint[] output,
            int inputOffset,
            int outputPosition)
        {
            switch (SelectScheme(input, inputOffset))
            {
                case 1: return EncodeValue(8, 1, 28, inputOffset, outputPosition, input, output);
                case 2: return EncodeValue(7, 2, 14, inputOffset, outputPosition, input, output);
                case 3: return EncodeValue(6, 3, 9, inputOffset, outputPosition, input, output);
                case 4: return EncodeValue(5, 4, 7, inputOffset, outputPosition, input, output);
                case 5: return EncodeValue(4, 5, 5, inputOffset, outputPosition, input, output);
                case 6: return EncodeValue(3, 7, 4, inputOffset, outputPosition, input, output);
                case 7: return EncodeValue(2, 9, 3, inputOffset, outputPosition, input, output);
                case 8: return EncodeValue(1, 14, 2, inputOffset, outputPosition, input, output);
                default:return EncodeValue(0, 28, 1, inputOffset, outputPosition, input, output);
            }
        }

        public static int Decode(uint[] input, uint[] output, int inputPosition, int outputOffset)
        {
            return Decode(input[inputPosition], output, outputOffset);
        }

        public static int Decode(uint input, uint[] output, int outputOffset)
        {
            switch (input >> 28)
            {
                case 0: return DecodeValueA(input, outputOffset, output);
                case 1: return DecodeValueB(input, outputOffset, output);
                case 2: return DecodeValueC(input, outputOffset, output);
                case 3: return DecodeValueD(input, outputOffset, output);
                case 4: return DecodeValueE(input, outputOffset, output);
                case 5: return DecodeValueF(input, outputOffset, output);
                case 6: return DecodeValueG(input, outputOffset, output);
                case 7: return DecodeValueH(input, outputOffset, output);
                default:return DecodeValueI(input, outputOffset, output);
            }
        }

        private readonly static int[] numValues = new[] { 28, 14, 9, 7, 5, 4, 3, 2, 1 };

        public static int NumValues(uint input)
        {
            return numValues[input >> 28];
        }

        private static int SelectScheme(uint[] input, int inputOffset)
        {
            int limit = (int) Utils.Min((uint)(input.Length - inputOffset), 28);
            uint maxVal = input[inputOffset];
            int i;
            for (i = 1; i < limit; i++)
            {
                maxVal = Utils.Max(maxVal, input[inputOffset + i]);
                if (maxVal > firstPassLimits[i])
                {
                    break;
                }
            }

            return firstPassNumbers[i - 1];
        }

        private static int SelectScheme(int[] input, int inputOffset)
        {
            int limit = (int)Utils.Min((uint)(input.Length - inputOffset), 28);
            int maxVal = input[inputOffset];
            int i;
            for (i = 1; i < limit; i++)
            {
                maxVal = Utils.Max(maxVal, input[inputOffset + i]);
                if (maxVal > firstPassLimits[i])
                {
                    break;
                }
            }

            return firstPassNumbers[i - 1];
        }
    }
}


