﻿namespace NETing.Index.Utils.Compression
{
    public static class Utils
    {
        private static readonly uint[] multiplyDeBruijnBitPosition = new uint[]
                                                                         {
                                                                             0, 1, 28, 2, 29, 14, 24, 3, 30, 22, 20, 15, 25, 17, 4, 8, 
                                                                             31, 27, 13, 23, 21, 19, 16, 7, 26, 12, 18, 6, 11, 5, 10, 9
                                                                         };

        // http://graphics.stanford.edu/~seander/bithacks.html
        public static uint MostSignificantBit(uint number)
        {  
            number |= number >> 1; // first round down to power of 2 
            number |= number >> 2;
            number |= number >> 4;
            number |= number >> 8;
            number |= number >> 16;
            number = (number >> 1) + 1;

            return multiplyDeBruijnBitPosition[(number * 0x077CB531U) >> 27];
        }

        public static uint Abs(int a)
        {
            unchecked
            {
                var temp = a >> 31;
                return (uint)((a ^ temp) - temp);    
            }
        }

        public static uint Min(uint a, uint b)
        {
            return (a + b - Abs((int)a - (int)b)) >> 1;
        }

        public static int Min(int a, int b)
        {
            return (int)((a + b - Abs(a - b)) >> 1);
        }

        public static uint Max(uint a, uint b)
        {
            return (a + b + Abs((int)a - (int)b)) >> 1;
        }

        public static int Max(int a, int b)
        {
            return (int)((a + b + Abs(a - b)) >> 1);
        }
    }
}


