﻿namespace NETing.Index.Utils.Compression
{
    public static class VByte
    {
        public static int Encode(ushort value, byte[] buffer, int offset)
        {
            while (value >= 128)
            {
                buffer[offset++] = (byte)((value & 127) | 128);
                value = (ushort)(value >> 7);
            }

            buffer[offset] = (byte)value;
            return offset + 1;
        }

        public static int Encode(uint value, byte[] buffer, int offset)
        {
            while (value >= 128)
            {
                buffer[offset++] = (byte)((value & 127) | 128);
                value = (ushort)(value >> 7);
            }
            
            buffer[offset] = (byte)value;
            return offset + 1;
        }

        public static int Write(byte[] buffer, int offset, ulong value)
        {
            while (value >= 128)
            {
                buffer[offset++] = (byte)((value & 127) | 128);
                value = value >> 7;
            }

            buffer[offset] = (byte)value;
            return offset + 1;
        }

        public static int Decode(byte[] buffer, int offset, out ushort value)
        {
            value = 0;
            for (int shift = 0;; shift += 7)
            {
                byte temp = buffer[offset++];
                value += (ushort)((temp & 127) << shift);
                if (temp < 128)
                {
                    break;
                }
            }

            return offset;
        }

        public static int Decode(byte[] buffer, int offset, out uint value)
        {
            value = 0;
            for (int shift = 0;; shift += 7)
            {
                byte temp = buffer[offset++];                
                value += (uint)((temp & 127) << shift);
                if (temp < 128)
                {
                    break;
                }
            }

            return offset;
        }

        public static int Read(byte[] buffer, int offset, out ulong value)
        {
            value = 0;
            for (int shift = 0;; shift += 7)
            {
                byte temp = buffer[offset++];
                value += ((ulong)(temp & 127)) << shift;
                if (temp < 128)
                {
                    break;
                }
            }

            return offset;
        }

        public static int SkipOne(byte[] input, int position)
        {
            for (int shift = 0; ; shift += 7)
            {
                byte temp = input[position++];
                if (temp < 128)
                {
                    break;
                }
            }

            return position;
        }

        public static int SkipMany(byte[] input, int position, int n)
        {
            while (n > 31)
            {
                position = SkipOne(input, position);
                position = SkipOne(input, position);
                position = SkipOne(input, position);
                position = SkipOne(input, position);

                position = SkipOne(input, position);
                position = SkipOne(input, position);
                position = SkipOne(input, position);
                position = SkipOne(input, position);

                position = SkipOne(input, position);
                position = SkipOne(input, position);
                position = SkipOne(input, position);
                position = SkipOne(input, position);

                position = SkipOne(input, position);
                position = SkipOne(input, position);
                position = SkipOne(input, position);
                position = SkipOne(input, position);

                position = SkipOne(input, position);
                position = SkipOne(input, position);
                position = SkipOne(input, position);
                position = SkipOne(input, position);

                position = SkipOne(input, position);
                position = SkipOne(input, position);
                position = SkipOne(input, position);
                position = SkipOne(input, position);

                position = SkipOne(input, position);
                position = SkipOne(input, position);
                position = SkipOne(input, position);
                position = SkipOne(input, position);

                position = SkipOne(input, position);
                position = SkipOne(input, position);
                position = SkipOne(input, position);
                position = SkipOne(input, position);
                n -= 32;
            }

            while (n > 15)
            {
                position = SkipOne(input, position);
                position = SkipOne(input, position);
                position = SkipOne(input, position);
                position = SkipOne(input, position);

                position = SkipOne(input, position);
                position = SkipOne(input, position);
                position = SkipOne(input, position);
                position = SkipOne(input, position);

                position = SkipOne(input, position);
                position = SkipOne(input, position);
                position = SkipOne(input, position);
                position = SkipOne(input, position);

                position = SkipOne(input, position);
                position = SkipOne(input, position);
                position = SkipOne(input, position);
                position = SkipOne(input, position);
                n -= 16;
            }

            while (n > 7)
            {
                position = SkipOne(input, position);
                position = SkipOne(input, position);
                position = SkipOne(input, position);
                position = SkipOne(input, position);

                position = SkipOne(input, position);
                position = SkipOne(input, position);
                position = SkipOne(input, position);
                position = SkipOne(input, position);
                n -= 8;
            }

            while (n > 3)
            {
                position = SkipOne(input, position);
                position = SkipOne(input, position);
                position = SkipOne(input, position);
                position = SkipOne(input, position);
                n -= 4;
            }

            while (n > 0)
            {
                position = SkipOne(input, position);
                n--;
            }

            return position;
        }
    }
}