#include "StdAfx.h"
#include "NPfor.h"
#include "UnPackers.h"
#include "Simple9.h"

#pragma unmanaged
void inline unPack(int width, const unsigned int* input, unsigned int* output)
{
	unPackers.unpack[width](input, output);
	unPackers.unpack[width](input + width, output + 32);
	unPackers.unpack[width](input + width + width, output + 64);
	unPackers.unpack[width](input + width + width + width, output + 96);
}

#pragma unmanaged
void inline unpackControlWord(
	unsigned int word,
	unsigned int* bitWidth, 
    unsigned int* numberOfExceptions, 
    unsigned int* frameOfReference)
{
    *bitWidth = word >> 27;
    *numberOfExceptions = (word >> 19) & 0xff;
    *frameOfReference = word & 0x7ffff;
}

#pragma unmanaged
int decode(
	const unsigned int* input,
	unsigned int* output,
	unsigned int* exceptions,
	unsigned int* indexes)
{
	unsigned int bitWidth, numberOfExceptions, frameOfReference;
	unpackControlWord(input[0], &bitWidth, &numberOfExceptions, &frameOfReference);
	unPack(bitWidth, &input[1], output);
	
	int word = 1 + (bitWidth << 2);	
    for (unsigned int i = 0; i < numberOfExceptions;
		i += simple9.Decode(input[word], exceptions, i), word++)
    {
    }

    for (unsigned int i = 0; i < numberOfExceptions;
        i += simple9.Decode(input[word], indexes, i), word++)
    {
    }

	int exceptionIndex = -1;
    for (unsigned int i = 0; i < numberOfExceptions; i++)
    {
        exceptionIndex += indexes[i] + 1;
        output[exceptionIndex] |= exceptions[i] << bitWidth;
    }

    if (frameOfReference > 0)
    {
        for (int i = 0; i < 128; i++)
        {
            output[i] += frameOfReference;
        }
    }

	return word;
}

#pragma unmanaged
int decode(
	const unsigned int* input,
	unsigned int* output,
	unsigned int* exceptions,
	unsigned int* indexes,
	int count)
{
	int word = 0;
	for (int i = 0; i < count; i++)
	{
		word += decode(
			&input[word],
			&output[i << 7],
			exceptions,
			indexes);
	}

	return word;
}

#pragma unmanaged
static const int S9NumValues[9] = { 28, 14, 9, 7, 5, 4, 3, 2, 1 };

#pragma unmanaged
int skipBlocks(const unsigned int* input, const int skipCount)
{
	int skippedWords = 0;
	for (int j = 0; j < skipCount; j++)
	{
		unsigned int bitWidth, exceptions, frame;
		unpackControlWord(input[0], &bitWidth, &exceptions, &frame);

		int word = 1 + (bitWidth << 2);	
		for (unsigned int i = 0; i < exceptions; i += S9NumValues[input[word] >> 28], word += 1)
        {
        }

        for (unsigned int i = 0; i < exceptions; i += S9NumValues[input[word] >> 28], word += 1)
        {
        }

		skippedWords += word;
		input += word;
	}

	return skippedWords;
}

#pragma unmanaged
int skipBlocksRecord(const unsigned int* input, const int skipCount, const int reference, int* output)
{
	int skippedWords = 0;
	for (int j = 0; j < skipCount; j++)
	{
		unsigned int bitWidth, exceptions, frame;
		unpackControlWord(input[0], &bitWidth, &exceptions, &frame);

		int word = 1 + (bitWidth << 2);	
		for (unsigned int i = 0; i < exceptions; i += S9NumValues[input[word] >> 28], word += 1)
        {
        }

        for (unsigned int i = 0; i < exceptions; i += S9NumValues[input[word] >> 28], word += 1)
        {
        }

		skippedWords += word;
		input += word;
		output[j] = reference + (skippedWords << 2);
	}

	return skippedWords;
}

#pragma managed

namespace NETing { namespace Index { namespace NativeCode {

int NPfor::Decode(
		const unsigned int* input, 
		unsigned int* output,
		unsigned int* exceptions,
		unsigned int* exceptionIndexes)
{
	return decode(
		input,
		output,
		exceptions,
		exceptionIndexes);
}

int NPfor::Decode(
		const unsigned int* input, 
		unsigned int* output,
		unsigned int* exceptions,
		unsigned int* exceptionIndexes,
		int blocks)
{
	return decode(
		input,
		output,
		exceptions,
		exceptionIndexes,
		blocks);
}

int NPfor::SkipBlocks(
		const unsigned int* input,
		const int skipCount)
{
	return skipBlocks(input, skipCount);
}

int NPfor::SkipBlocks(
		const unsigned int* input,
		const int skipCount,
		const int reference,
		int* outputs)
{
	return skipBlocksRecord(input, skipCount, reference, outputs);
}
}}}